#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014-2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.util as util
    import copy
    import gc
    import json
    import nlmanager.nlmanager
    import nlmanager.nlpacket
    import os
    import platform
    import select
    import socket
    import struct
    import sys
    import threading
    import time
    import traceback
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

yappiAvail = True
try:
    import yappi
except ImportError, e:
    yappiAvail = False

class cmdsrv:
    """
    This class handles everything required to implement the command server
    interface for clagd. This interface allows other processes to communicate
    with clagd to get status and alter configuration. clagctl is an example
    of a process which connects to this interface.
    """

    _CmdSocketFile = "/var/run/clagd.socket"

    #   XXX: This is a hack. Python socket module doesn't define the SO_PEERCRED
    #   value. And this value is different on different CPU architectures. So on
    #   PowerPC platforms the value is 21 and on all other platforms the value is 17.
    _SOCKET_PEERCRED = 21 if platform.machine() == "ppc" else 17
    _JSON_SORT_KEYS = True
    _JSON_INDENT = 4


    def __init__(self, Daemon, Intf, Parser, Log, LacpSync, VxLanSync, FdbSync, MdbSync, NeighSync, VlanSync, HealthCheck):
        """
        Constructor: Initialize data structures, threads, etc.
        """

        #
        #   Save the parameters passed in
        #
        self.Daemon     = Daemon
        self.Intf       = Intf
        self.Parser     = Parser
        self.Log        = Log
        self.LacpSync   = LacpSync
        self.VxLanSync  = VxLanSync
        self.FdbSync    = FdbSync
        self.MdbSync    = MdbSync
        self.NeighSync  = NeighSync
        self.VlanSync   = VlanSync
        self.HealthCheck = HealthCheck


        #
        #   CLAGD command interface sockets. The cmdSrvSock is a socket to which clients
        #   can connect to open up communication with the clagd daemon. The cmdOutputBufs
        #   is a dictionary keyed by the client socket, with the string of data to write
        #   to the socket as the value. Only sockets which have data waiting to be output 
        #   are in this dictionary. The cmdGetAll set is a list of sockets which are 
        #   currently in the "getall" state, monitoring the status of CLAGD.
        #
        self.cmdSrvSock         = None
        self.cmdOutputLock      = threading.Lock()
        self.cmdOutputBufs      = {}
        self.cmdGetAll          = set()

        #
        #   Start a thread to handle requests from other tools
        #
        self.keep_going         = True
        self.commandThread = threading.Thread(None, self.CmdRecvT, "CmdRecv")
        self.commandThread.daemon = True
        self.commandThread.start()

    def CmdRecvT(self):
        try:
            self.Log.log("[%d]Init CmdRecvT" % (self.Parser.getTid()))
            self.CmdRecv()
        except:
            self.Daemon.DumpTraceback()


    def __del__(self):
        """
        Destructor:
        """
        #
        #   Signal the command thread to stop and wait for it to die
        #
        self.SignalThreads()
        self.JoinThreads()

    def SignalThreads(self):
        self.keep_going = False

    def JoinThreads(self):
        if self.cmdSrvSock is not None:
            cmdSrvSock = self.cmdSrvSock
            self.cmdSrvSock = None
            cmdSrvSock.shutdown(socket.SHUT_RDWR)
        self.commandThread.join()

    def CmdRecv(self):
        '''
        This function is an independent thread of the script. It creates a socket,
        cmdSrvSock, and accepts connections on that socket. It receives the commands
        sent through those connections and replies with the proper command responses.
        '''
        self.Log.log_verbose(self.Log._LOG_DEBUG_CMD_SERVER, "Beginning execution of the thread CmdRecv")

        # Create a socket upon which we will accept connections
        try:
            self.cmdSrvSock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.cmdSrvSock.setblocking(0)
        except socket.error as msg:
            self.cmdSrvSock = None
            self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Unable to create a Unix stream socket (%s)" % (msg,))
        else:
            try:
                os.remove(cmdsrv._CmdSocketFile)
            except OSError:
                pass

            try:
                self.cmdSrvSock.bind(cmdsrv._CmdSocketFile)
                os.chmod(cmdsrv._CmdSocketFile, 0777)
                self.cmdSrvSock.listen(5)
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Server socket to %s was created" % (cmdsrv._CmdSocketFile))
            except socket.error as msg:
                self.cmdSrvSock.close()
                os.remove(cmdsrv._CmdSocketFile)
                self.cmdSrvSock = None
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Unable to bind socket to %s (%s)" % (cmdsrv._CmdSocketFile, msg))

        inputBufs = {}
        inputs  = [ self.cmdSrvSock ]
        nlm = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_CMD_ID)
        while (inputs and self.cmdSrvSock and self.keep_going) or self.cmdOutputBufs:

            self.cmdOutputLock.acquire()
            cmdOutputSocks = self.cmdOutputBufs.keys()
            self.cmdOutputLock.release()

            # Wait until there is a socket to read, write, or an error
            try:
                (readable, writable, exceptional) = select.select(inputs, cmdOutputSocks, inputs + cmdOutputSocks, self.Parser.args.cmdConnect)
            except select.error as e:
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "A select.error occurred")
                (readable, writable, exceptional) = ([],[],[])
            except socket.error:
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "A socket.error occurred")
                (readable, writable, exceptional) = ([],[],[])

            if self.keep_going:
                for readSock in readable:
                    if readSock is self.cmdSrvSock:
                        # Accept a new socket connection
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Accepting a new cmd connection")
                        (newSock, clientAddr) = readSock.accept()
                        newSock.setblocking(0)
                        inputs.append(newSock)
                        inputBufs[newSock] = ""
                    else:
                        # Read data from this socket
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Reading data from the command socket")
                        try:
                            inputBuf = readSock.recv(2048)
                        except:
                            inputBuf = 0
                        if inputBuf:
                            # Add the data read to this socket's receive buffer
                            inputBufs[readSock] += inputBuf
                            while chr(0) in inputBufs[readSock]:
                                inEnd = inputBufs[readSock].find(chr(0))
                                inCmd = inputBufs[readSock][:inEnd]
                                inputBufs[readSock] = inputBufs[readSock][inEnd+1:]
                                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "    %s" % (inCmd,))
                                self.HandleCommand(nlm, inCmd, readSock)
                        else:
                            # If no data was available, the socket has been closed, end of message.
                            if readSock in inputs:
                                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Closing a socket because there's nothing to read")
                                del inputBufs[readSock]
                                inputs.remove(readSock)
                                readSock.close()

            for writeSock in writable:
                self.cmdOutputLock.acquire()
                writeStr = self.cmdOutputBufs.get(writeSock)
                self.cmdOutputLock.release()
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Sending data to a socket: %s" % (writeStr,))
                if writeStr is not None:
                    try:
                        numSent = writeSock.send(writeStr)
                    except:
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Error sending data to the socket. Closing socket.")
                        numSent = len(writeStr)
                        writeSock.close()
                    self.cmdOutputLock.acquire()
                    self.cmdOutputBufs[writeSock] = writeStr[numSent:]
                    if not self.cmdOutputBufs[writeSock]:
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "No more data to output right now on the socket")
                        del self.cmdOutputBufs[writeSock]
                    self.cmdOutputLock.release()

            for errSock in exceptional:
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Closing a socket because of an error")
                if errSock in inputs:
                    if errSock in inputBufs:
                        del inputBufs[errSock]
                    inputs.remove(errSock)
                self.cmdOutputLock.acquire()
                if errSock in self.cmdOutputBufs:
                    del self.cmdOutputBufs[errSock]
                self.cmdOutputLock.release()
                errSock.close()
                if errSock is self.cmdSrvSock:
                    self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "   The socket was the server socket")
                    self.cmdSrvSock = None

            # Check to see if any sockets have closed
            self.cmdOutputLock.acquire()
            allSocks = self.cmdOutputBufs.keys() + inputs
            self.cmdOutputLock.release()

            for testSock in allSocks:
                try:
                    testSock.fileno()
                except socket.error:
                    self.cmdOutputLock.acquire()
                    self.cmdOutputBufs.pop(testSock, None)
                    self.cmdOutputLock.release()
                    if testSock in inputs:
                        inputs.remove(testSock)

        self.Log.log_verbose(self.Log._LOG_DEBUG_CMD_SERVER, "Finished execution of the thread CmdRecv")


    #-------------------------------------------------------------------------------
    #
    #   Command Interface Functions - These functions implement the command interface
    #   to the clagd daemon.
    #
    #-------------------------------------------------------------------------------

    def HandleCmdEcho(self, nlm, cmdStr, ourSock):
        return (cmdStr[5:], chr(0))

    def HandleCmdSummary(self, nlm, cmdStr, ourSock):
        status = {}
        status['peerAlive'] = self.Intf.isPeerAlive()
        if self.Intf.isPeerAlive():
            status['peerPriority'] = self.Parser.peerId[0]
            status['peerId'] = self.Parser.peerId[1]
            status['peerRole'] = self.Parser.peerRole
        else:
            status['peerPriority'] = self.Parser.default_priority
            status['peerId'] = "00:00:00:00:00:00"
            status['peerRole'] = self.Parser.default_role

        status['ourPriority'] = self.Parser.clagId[0]
        status['ourId'] = self.Parser.clagId[1]
        status['ourRole'] = self.Parser.GetCombinedRole()

        status['sysMac'] = self.Parser.args.sysMac
        status['backupIp'] = self.Parser.args.backupIp
        if self.Parser.args.backupVrf:
            status['backupVrf'] = self.Parser.args.backupVrf
        status['backupActive'] = self.HealthCheck.backupActive
        status['backupReason'] = self.HealthCheck.backupReason
        if self.Parser.args.vxlanAnycast:
            status['vxlanAnycast'] = self.Parser.args.vxlanAnycast

        status['peerIf'] = self.Parser.args.peerIf
        status['peerIp'] = self.Parser.remotePeerIp
        status['linklocal'] = (self.Parser.args.peerIp == 'linklocal')

        jsonStr = json.dumps(status, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))

    def HandleCmdBackupInfo(self, nlm, cmdStr, ourSock):
        backup = {}
        backup['backupIp'] = self.Parser.args.backupIp
        if self.Parser.args.backupVrf:
            backup['backupVrf'] = self.Parser.args.backupVrf
        backup['backupActive'] = self.HealthCheck.backupActive
        backup['backupReason'] = self.HealthCheck.backupReason
        backup['backupRole'] = self.HealthCheck.backupRole
        if self.HealthCheck.backupRoleAvailable:
            backup['peerPriority'] = self.HealthCheck.peerId[0]
            backup['peerId'] = self.HealthCheck.peerId[1]
            backup['peerRole'] = self.HealthCheck.peerRole
        else:
            backup['peerPriority'] = self.Parser.default_priority
            backup['peerId'] = "00:00:00:00:00:00"
            backup['peerRole'] = self.Parser.default_role

        jsonStr = json.dumps(backup, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))

    def HandleCmdParams(self, nlm, cmdStr, ourSock):
        '''
        Implements the "params" command. Fill up a string with all of the interesting
        parameters of the clagd program.
        '''
        params = {}
        params['clagVersion'] = self.Parser._ClagVersion
        params['clagDataVersion'] = self.Parser._ClagDataVersion
        params['clagCmdVersion'] = self.Parser._ClagCmdVersion
        params['peerIp'] = self.Parser.args.peerIp
        params['peerIf'] = self.Parser.args.peerIf
        params['sysMac'] = self.Parser.args.sysMac
        params['lacpPoll'] = self.Parser.getCfgLacpPoll()
        params['currLacpPoll'] = self.Parser.getCurrLacpPoll()
        params['peerConnect'] = self.Parser.args.peerConnect
        params['cmdConnect'] = self.Parser.args.cmdConnect
        params['peerLinkPoll'] = self.Parser.getPeerLinkPoll()
        params['switchdReadyTimeout'] = self.Parser.args.switchdReadyTimeout
        params['reloadTimer'] = self.Parser.args.reloadTimer
        params['periodicRun'] = self.Parser.getPeriodicRun()
        params['initDelay'] = self.Parser.args.initDelay
        params['priority'] = self.Parser.args.priority
        params['quiet'] = self.Log.getQuiet()
        params['debug'] = self.Log.getDebug()
        params['verbose'] = self.Log.getVerbose()
        params['log'] = self.Log.getHandlerStr()
        params['vm'] = self.Parser.args.vm
        params['peerPort'] = self.Parser.args.peerPort
        params['peerTimeout'] = self.Parser.getPeerTimeout()
        params['sendTimeout'] = self.Parser.args.sendTimeout
        params['sendBufSize'] = self.Parser.args.sendBufSize
        params['forceDynamic'] = self.Parser.args.forceDynamic
        params['dormantDisable'] = self.Parser.args.dormantDisable
        params['redirectEnable'] = self.Parser.args.redirectEnable
        params['neighSync'] = self.Parser.args.neighSync
        params['permanentMacSync'] = self.Parser.args.permanentMacSync
        params['backupIp'] = self.Parser.args.backupIp
        if self.Parser.args.backupVrf:
            params['backupVrf'] = self.Parser.args.backupVrf
        params['vxlanAnycast'] = self.Parser.args.vxlanAnycast
        params['backupPort'] = self.Parser.args.backupPort
        params['switchdState'] = self.Parser.switchdState
        params['peerlinkLearnEnable'] = self.Parser.args.peerlinkLearnEnable
        params['cmdLine'] = sys.argv

        jsonStr = json.dumps(params, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))


    def HandleCmdBondSetClagId(self, nlm, cmdStr, ourSock):
        cmdList = cmdStr.split()
        try:
            bond = cmdList[1]
            clagId = int(cmdList[2])
        except:
            retStr= "ERROR SetClagId command requires <bond> <0-%d>: %s" % (self.Parser._CLAG_ID_MAX, cmdStr,)
            return (retStr, chr(0))
        return (self.Parser.setBondClagId(nlm, bond, clagId), chr(0))


    def HandleCmdReloadDone(self, nlm, cmdStr, ourSock):
        return (self.Parser.setReloadDone(), chr(0))


    def HandleCmdSetAnycastIp(self, nlm, cmdStr, ourSock):
        cmdList = cmdStr.split()
        try:
            address = cmdList[1]
        except:
            address = None
        return (self.Parser.setAnycastIp(nlm, address), chr(0))


    def HandleCmdSetBackupIp(self, nlm, cmdStr, ourSock):
        cmdList = cmdStr.split()
        try:
            address = cmdList[1]
        except:
            address = None

        vrf = None
        try:
            if cmdList[2] == "vrf":
                vrf = cmdList[3]
        except:
            pass
        return (self.Parser.setBackupIp(nlm, address, vrf), chr(0))

    def HandleCmdDebug(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Log.setDebug(int(cmdList[1],0))
        except:
            retStr = "ERROR Debug command requires an integer argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdVerbose(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Log.setVerbose(cmdList[1].lower() in ["yes", "true", "t", "1"])
        except:
            retStr = "ERROR SetVerbose command requires a True/False argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdQuiet(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Log.setQuiet(cmdList[1].lower() in ["yes", "true", "t", "1"])
        except:
            retStr = "ERROR SetQuiet command requires a True/False argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdLogfile(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Log.setHandler(cmdList[1])
        except:
            retStr = "ERROR SetLogFile command requires the name of a log file: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdLacpPoll(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Parser.args.lacpPoll = int(cmdList[1],0)
        except:
            retStr = "ERROR SetLacpPoll command requires an integer argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdPeerTimeout(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            timeout = int(cmdList[1],0)
            if timeout == 0:
                timeout = None
            self.Parser.setPeerTimeout(timeout)
        except:
            retStr = "ERROR SetPeerTimeout command requires an integer argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdSendTimeout(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Parser.args.sendTimeout = int(cmdList[1],0)
            self.Parser.CloseClientSock()
        except:
             retStr = "ERROR SetSendTimeout command requires an integer argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdSendBufSize(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Parser.args.sendBufSize = int(cmdList[1],0)
            self.Parser.CloseClientSock()
        except:
            retStr = "ERROR SetSendBufSize command requires an integer argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdPriority(self, nlm, cmdStr, ourSock):
        global clagId
        retStr = ""
        cmdList = cmdStr.split()
        try:
            newPriority = int(cmdList[1],0)
            if 0 <= newPriority <= 65535:
                if self.Parser.GetMismatchState():
                    raise Exception
                self.Parser.clagId[0] = newPriority
            else:
                raise ValueError
        except Exception:
            retStr = "ERROR SetPriority command cannot be execute when clag peers are in mismatch state"
        except ValueError:
            retStr = "ERROR SetPriority command requires an integer argument, 0-65535: %s" % (cmdStr,)
        except:
            retStr = "ERROR SetPriority command requires an integer argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdPeerLinkPoll(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            self.Parser.setPeerLinkPoll(int(cmdList[1],0))
        except:
            retStr = "ERROR SetPeerLinkPoll command requires an integer argument: %s" % (cmdStr,)
        return (retStr, chr(0))


    def HandleCmdDualIfs(self, nlm, cmdStr, ourSock):
        dualIfDb = {}
        dualIfs = self.Intf.GetDualIfs()
        for dualIf in dualIfs:
            dualIfInfo={}
            if dualIfs[dualIf]:
                dualIfInfo['peerIf'] = dualIfs[dualIf]
                dualIfInfo['status'] = "dual"
            else:
                dualIfInfo['status'] = "single"
            dualIfInfo['clagId'] = self.Parser.GetBondClagId(dualIf)
            dualIfDb[dualIf] = dualIfInfo
        jsonStr = json.dumps(dualIfDb, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))

    def HandleCmdIdMatchIfs(self, nlm, cmdStr, ourSock):
        dualIfDb = {}
        dualIfs = self.Intf.GetClagIdMatchInfo()
        for dualIf in dualIfs:
            dualIfInfo={}
            if dualIfs[dualIf]:
                dualIfInfo['peerIf'] = dualIfs[dualIf][self.Intf.CLAG_ID_MATCHIF_IDX_RIF]
                dualIfInfo['peerIfOperState'] = dualIfs[dualIf][self.Intf.CLAG_ID_MATCHIF_IDX_RIF_STATE]
            dualIfInfo['clagId'] = self.Parser.GetBondClagId(dualIf)
            dualIfDb[dualIf] = dualIfInfo
        jsonStr = json.dumps(dualIfDb, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))



    def HandleCmdPeerAlive(self, nlm, cmdStr, ourSock):
        return (str(self.Intf.isPeerAlive()), chr(0))


    def HandleCmdClagId(self, cmdStr, ourSock):
        jsonStr = " ".join([str(self.Parser.clagId[0]), str(self.Parser.clagId[1])])
        return (jsonStr, chr(0))


    def HandleCmdClagRole(self, nlm, cmdStr, ourSock):
        return (str(self.Parser.clagRole), chr(0))


    def HandleCmdBackupPeerId(self, nlm, cmdStr, ourSock):
        retStr = "-"
        if self.HealthCheck.backupRoleAvailable:
            retStr = " ".join([str(self.HealthCheck.peerId[0]), str(self.HealthCheck.peerId[1])])
        return (retStr, chr(0))


    def HandleCmdBackupPeerRole(self, nlm, cmdStr, ourSock):
        retStr = "-"
        if self.HealthCheck.backupRoleAvailable:
            retStr = str(self.HealthCheck.peerRole)
        return (retStr, chr(0))

    def HandleCmdPeerId(self, nlm, cmdStr, ourSock):
        retStr = " ".join([str(self.Parser.peerId[0]), str(self.Parser.peerId[1])])
        return (retStr, chr(0))


    def HandleCmdPeerRole(self, nlm, cmdStr, ourSock):
        return (str(self.Parser.peerRole), chr(0))


    def HandleCmdPeerIf(self, nlm, cmdStr, ourSock):
        return (str(self.Parser.args.peerIf), chr(0))


    def HandleCmdPeerIp(self, nlm, cmdStr, ourSock):
        return (str(self.Parser.args.peerIp), chr(0))


    def HandleCmdGetBackupIp(self, nlm, cmdStr, ourSock):
        retStr = str(self.Parser.args.backupIp) if self.Parser.args.backupIp else '-'
        return (retStr, chr(0))


    def HandleCmdBackupActive(self, nlm, cmdStr, ourSock):
        stateStr = 'active'
        if not self.HealthCheck.backupActive:
            stateStr = 'inactive'
            if self.HealthCheck.backupReason:
                stateStr += ' (%s)' % (self.HealthCheck.backupReason,)
        return (stateStr, chr(0))


    def HandleCmdBackupRole(self, nlm, cmdStr, ourSock):
        retStr = 'primary' if self.HealthCheck.backupRole == self.Parser._ROLE_PRIMARY else 'secondary'
        return (retStr, chr(0))


    def HandleCmdSysMac(self, nlm, cmdStr, ourSock):
        return (str(self.Parser.args.sysMac), chr(0))


    def HandleCmdOurLacpDB(self, nlm, cmdStr, ourSock):
        return (self.LacpSync.EncodedOurStr(), chr(0))


    def HandleCmdPeerLacpDB(self, nlm, cmdStr, ourSock):
        return (self.LacpSync.EncodedPeerStr(), chr(0))


    def HandleCmdPeerLacpRate(self, nlm, cmdStr, ourSock):
        self.Parser.peerLacpRateLock.acquire()
        retStr = str(self.Parser.peerLacpRate)
        self.Parser.peerLacpRateLock.release()
        return (retStr, chr(0))


    def HandleCmdOurMacDB(self, nlm, cmdStr, ourSock):
        return (self.FdbSync.EncodedOurStr(), chr(0))


    def HandleCmdGetQInfo(self, nlm, cmdStr, ourSock):
        return (util.queueGetInfo(), chr(0))


    def HandleCmdGetObjectInfo(self, nlm, cmdStr, ourSock):
        jsonStr = json.dumps(self.Parser.getInterestingObjectsSize(),
                             sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))


    def HandleCmdGetDBCheckSum(self, nlm, cmdStr, ourSock):
        checkSumDict = {}
        if self.NeighSync:
            checkSumDict['neighDB'] = self.NeighSync.CreateDBCheckSum()
        jsonStr = json.dumps(checkSumDict, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))


    def HandleCmdGetTimers(self, nlm, cmdStr, ourSock):
        timers = {}
        initDelayRemaining = 0
        if self.Parser.args.initDelay and self.HealthCheck.initDelayStartTime:
            initDelayRemaining = (self.HealthCheck.initDelayStartTime
                                 + self.Parser.args.initDelay
                                 - int(os.times()[4]))
        if initDelayRemaining < 0:
            initDelayRemaining = 0
        helloReloadRemaining = (self.HealthCheck.helloReloadStartTime
                               + self.Parser.args.reloadTimer
                               - int(os.times()[4]))
        if self.Intf.firstInitHandShakeDone or helloReloadRemaining < 0:
            helloReloadRemaining = 0

        peerTimeoutRemaining = (self.Parser.peerTimeoutStart
                               + self.Parser.getPeerTimeout()
                               - int(os.times()[4]))
        if peerTimeoutRemaining < 0:
            peerTimeoutRemaining = 0
        timers['reloadTimer'] = {"value": self.Parser.args.reloadTimer,
                                "remaining": time.strftime('%H:%M:%S', time.gmtime(helloReloadRemaining))}
        timers['initDelay'] = {"value": self.Parser.args.initDelay,
                               "remaining": time.strftime('%H:%M:%S', time.gmtime(initDelayRemaining))}
        timers['sendTimeout'] = {"value": self.Parser.args.sendTimeout,
                                 "remaining": '-'}
        timers['peerTimeout'] = {"value": self.Parser.getPeerTimeout(),
                                 "remaining":time.strftime('%H:%M:%S', time.gmtime(peerTimeoutRemaining))}
        jsonStr = json.dumps(timers, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))


    def HandleCmdClagBondDB(self, nlm, cmdStr, ourSock):
        return self.HandleCmdClagIntfDB(cmdStr, ourSock, True)


    def HandleCmdClagIntfDB(self, nlm, cmdStr, ourSock, justBonds=False):
        intfDb = {}
        dualIfs = self.Intf.GetDualIfs()
        if justBonds:
            clagIntfs = self.Parser.GetClagBondDB()
        else:
            clagIntfs = self.Parser.GetClagIntfDB()
        for intf in clagIntfs.values():
            intfInfo = {}
            intfInfo['clagId'] = intf.clagId
            peerIf = dualIfs.get(intf.intfName)
            intfInfo['operstate'] = self.Intf.GetOperState(intf.intfName)
            if peerIf:
                intfInfo['peerIf'] = peerIf
                intfInfo['status'] = "dual"
            else:
                intfInfo['status'] = "single"
            conflicts = self.Parser.GetClagIntfConflictStr(intf.intfName)
            if conflicts:
                intfInfo['conflicts'] = conflicts
            protoDownReasons = intf.GetProtoDownReasons()
            if protoDownReasons:
                intfInfo['protoDown'] = protoDownReasons
            intfDb[intf.intfName]=intfInfo
        jsonStr = json.dumps(intfDb, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))


    def HandleCmdNlLinkCache(self, nlm, cmdStr, ourSock):
        cmdList = cmdStr.split()
        linkCache = self.Intf.GetKCacheLinkCopy()
        linkCacheDb = {}
        linkCacheNameDb = {}
        if linkCache:
            for ifIndex, iface in linkCache.items():
                name = iface.get(nlmanager.nlpacket.Link.IFLA_IFNAME)
                flags = iface.get(self.Intf.IFLA_FLAGS)
                operState = iface.get(nlmanager.nlpacket.Link.IFLA_OPERSTATE, 0)

                iface['flags'] = flags
                iface['operState'] = nlmanager.nlpacket.Link.oper_to_string.get(operState, "unknown")
                if name in cmdList:
                    linkCacheNameDb[name] = iface
                linkCacheDb[name] = iface
        if linkCacheNameDb:
            linkCacheDb = linkCacheNameDb
        jsonStr = json.dumps(linkCacheDb, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))

    def DumpThreadStacks(self):
        id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
        self.Log.log("DumpThreadStacks - start\n")
        for threadId, stack in sys._current_frames().items():
            self.Log.log("\nthread: %s (%d)\n" % (id2name.get(threadId,""), threadId))
            for filename, lineno, name, line in traceback.extract_stack(stack):
                self.Log.log('  file: %s, line %d, in %s\n' % (filename, lineno, name))
            self.Log.log("\n")
        self.Log.log("DumpThreadStacks - end\n")

    def HandleCmdIntInfo(self, nlm, cmdStr, ourSock):
        intInfo = {}
        intInfo['syncDoneFromPeer'] = self.Parser.syncDoneFromPeer
        intInfo['peerDataVersion'] = self.Parser.peerDataVersion
        intInfo['configReloadDone'] = self.Parser.configReloadDone
        jsonStr = json.dumps(intInfo, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (jsonStr, chr(0))


    def HandleCmdPeerMacDB(self, nlm, cmdStr, ourSock):
        return (self.FdbSync.EncodedPeerStr(), chr(0))


    def HandleCmdPermanentMacDB(self, nlm, cndStr, ourSock):
        return (self.FdbSync.EncodedPermStr(), chr(0))


    def HandleCmdPermanentNeighDB(self, nlm, cndStr, ourSock):
        return (self.NeighSync.EncodedPermNeighStr(), chr(0))


    def HandleCmdOurMdbDB(self, nlm, cmdStr, ourSock):
        return (self.MdbSync.EncodedOurStr(), chr(0))


    def HandleCmdPeerMdbDB(self, nlm, cmdStr, ourSock):
        return (self.MdbSync.EncodedPeerStr(), chr(0))

    def HandleCmdNeighDB(self, nlm, cmdStr, ourSock):
        if self.NeighSync:
            return (self.NeighSync.EncodedStr(), chr(0))
        else:
            return ("[]", chr(0))

    def HandleCmdKernNeighDB(self, nlm, cmdStr, ourSock):
        if self.NeighSync:
            return (self.NeighSync.EncodedKernStr(), chr(0))
        else:
            return ("[]", chr(0))

    def HandleCmdOurRportDB(self, nlm, cmdStr, ourSock):
        return (self.MdbSync.EncodedOurRportStr(), chr(0))


    def HandleCmdPeerRportDB(self, nlm, cmdStr, ourSock):
        return (self.MdbSync.EncodedPeerRportStr(), chr(0))


    def HandleCmdPeerVlanDB(self, nlm, cmdStr, ourSock):
        return (self.VlanSync.EncodedPeerStr(), chr(0))


    def HandleCmdOurVlanDB(self, nlm, cmdStr, ourSock):
        return (self.VlanSync.EncodedOurStr(), chr(0))


    def HandleCmdLogMsg(self, nlm, cmdStr, ourSock):
        self.Log.log(cmdStr[7:])
        return ("", chr(0))


    def HandleCmdGarbage(self, nlm, cmdStr, ourSock):
        return (str(gc.collect()), chr(0))

    
    def HandleCmdCollectObjectCount(self, nlm, cmdStr, ourSock):
        return ((str(len(gc.get_objects()))), chr(0))

    
    def HandleCmdUnCollectableObject(self, nlm, cmdStr, ourSock):
        return (str(gc.garbage), chr(0))


    def HandleCmdGetAll(self, nlm, cmdStr, ourSock):
        termChar = chr(0)
        if ourSock not in self.cmdGetAll:
            retStr = self.HandleMstpState()
            self.cmdGetAll.add(ourSock)
            termChar = ""
        else:
            retStr = "ERROR The GetAll command has already been enabled on this socket."
        return (retStr, termChar)


    def HandleCmdDumpMstp(self, nlm, cmdStr, ourSock):
        return (self.HandleMstpState(), chr(0))


    def HandleCmdNoGetAll(self, nlm, cmdStr, ourSock):
        self.cmdGetAll.discard(ourSock)
        return ("", chr(0))


    def HandleMstpState(self):
        '''
        Dumps the current state of clagd (as far as mstpd is concerned) at the
        beginning of a getall command.
        '''
        paramStr = ""
        peerLink = self.Parser.args.peerIf
        if self.Intf.isSubIfName(peerLink):
            masters = self.Intf.GetLogicalMastersOfSubIf(peerLink)
            if len(masters):
                peerLink = masters[0]
        paramStr += "add peer-link = %s\n" % (peerLink,)
        paramStr += "add clag-state = %s\n" % ("up" if self.Intf.isPeerAlive() and not self.Parser.GetMismatchState() else "down",)
        paramStr += "add system-mac = %s\n" % (self.Parser.args.sysMac,)
        paramStr += "add clag-role = %s\n" % (str(self.Parser.GetCombinedRole()),)

        myDualIfs = self.Intf.GetDualIfs()
        myPeerLacpDB = self.LacpSync.GetPeerLacpDB()

        for dualIf in myDualIfs:
            peerData = myPeerLacpDB.get(myDualIfs[dualIf])
            if peerData is not None:
                paramStr += "add mac-" + str(peerData[0]) + " = " + dualIf + "," + str(peerData[1]) + "\n"

        return paramStr


    def HandleStartProfile(self, nlm, cmdStr, ourSock):
        retStr = ""
        if yappiAvail:
            yappi.clear_stats()
            yappi.set_context_name_callback(lambda: threading.current_thread().name)
            yappi.start()
        else:
            retStr = "ERROR Cannot start profiling. Yappi is not installed."
        return (retStr, chr(0))


    def HandleStopProfile(self, nlm, cmdStr, ourSock):
        retStr = ""
        if yappiAvail:
            yappi.stop()
        else:
             retStr = "ERROR Cannot stop profiling. Yappi is not installed."
        return (retStr, chr(0))


    def HandleDumpProfile(self, nlm, cmdStr, ourSock):
        retStr = ""
        if yappiAvail:
            self.Log.log("%-40s %9s %12s %12s %12s" % ("Function", "ncall", "tsub", "ttot", "tavg"))
            self.Log.log("%-40s %9s %12s %12s %12s" % ("--------", "-----", "----", "----", "----"))
            for stat in yappi.get_func_stats().sort('totaltime'):
                self.Log.log("%-40s %9d %12.6f %12.6f %12.6f" % (stat[0][-40:], stat[3], stat[7], stat[6], stat[11]))

            self.Log.log("%-20s %12s %12s" % ("Thread", "ttot", "schedcnt"))
            self.Log.log("%-20s %12s %12s" % ("------", "----", "--------"))
            for stat in yappi.get_thread_stats().sort('totaltime'):
                self.Log.log("%-20s %12.6f %12d" % (stat[0][-20:], stat[3], stat[4]))
        else:
            retStr = "ERROR Cannot dump profiling. Yappi is not installed."
        return (retStr, chr(0))


    def HandleCmdGetSocket(self, nlm, cmdStr, ourSock):
        retStr = json.dumps(self.Parser.getSocketInfo(), sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return (retStr, chr(0))


    def HandleCmdGetDebugFlags(self, nlm, cmdStr, ourSock):
        retStr = ""
        logList = self.Log.getDebugFlags()
        for arg in logList:
            retStr += "%s " % arg
        return (retStr, chr(0))


    def HandleCmdSetDebugFlags(self, nlm, cmdStr, ourSock):
        retStr = ""
        errStr = ""
        cmdList = cmdStr.split()
        if len(cmdList) < 2:
            retStr = "ERROR setdebugflags command requires one or more arguments"
        del cmdList[0:1]
        logList = self.Log.getDebugFlags()
        for arg in cmdList:
            if arg not in logList:
                try:
                    self.Log.setDebugFlags(arg)
                except:
                    errStr += "%s " % arg
        if errStr:
            retStr = "ERROR setdebugflags command invalid argument: %s" % (errStr,)
        return (retStr, chr(0))


    def HandleCmdClearDebugFlags(self, nlm, cmdStr, ourSock):
        retStr = ""
        errStr = ""
        cmdList = cmdStr.split()
        if len(cmdList) < 2:
            self.Log.clearDebugFlags()
        del cmdList[0:1]
        logList = self.Log.getDebugFlags()
        for arg in cmdList:
            if arg in logList:
                try:
                    self.Log.clearDebugFlags(arg)
                except:
                    errStr += "%s " % arg
        if errStr:
            retStr = "ERROR cleardebugflags command invalid argument: %s" % (errStr,)
        return (retStr, chr(0))


    def HandleCmdSetFastShutDown(self, nlm, cmdStr, ourSock):
        retStr = ""
        try:
            self.Parser.fastShutdown = True
        except:
            retStr = "ERROR setfastshutdown command failed"
        return (retStr, chr(0))


    def HandleCmdSetVxlanLocalIp(self, nlm, cmdStr, ourSock):
        retStr = ""
        cmdList = cmdStr.split()
        try:
            localIp = cmdList[1]
        except:
            retStr= "ERROR SetVxlanLocalIp command requires an argument: %s" % (cmdStr,)
            return (retStr, chr(0))
        return (self.VxLanSync.SetCmdVxlanLocalIp(localIp), chr(0))


    def CmdGetAllOutput(self, outStr, ignoreLocks=False):
        '''
        This function is called by various threads when certain events happen to
        send output to sockets which are monitoring the state of CLAGD.
        '''
        for ourSock in self.cmdGetAll:
            self.SendSock(ourSock, outStr + "\n", ignoreLocks)


    def SendSock(self, outSock, outStr, ignoreLocks):
        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Sending data to a socket: %s" % (outStr,))
        if outStr is not None:
            if not self.cmdOutputBufs.get(outSock):
                try:
                    numSent = outSock.send(outStr)
                except:
                    self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Error sending data to the socket. Closing socket.")
                    numSent = len(outStr)
                    outSock.close()
                outStr = outStr[numSent:]
            if outStr:
                if not ignoreLocks:
                    self.cmdOutputLock.acquire()
                self.cmdOutputBufs.setdefault(outSock, "")
                self.cmdOutputBufs[outSock] += outStr
                if not ignoreLocks:
                    self.cmdOutputLock.release()
        return


    def HandleCommand(self, nlm, cmdStr, ourSock):
        '''
        This function is called from the CmdRecv thread when a command has been
        received. The command is parsed, executed,  and the output data, if any, 
        is added to the output buffer.
        '''

        clagCmds = {
        #   Command              Function                    Priviledged
            "echo"            : (self.HandleCmdEcho,         False),
            "params"          : (self.HandleCmdParams,       False),
            "setclagid"       : (self.HandleCmdBondSetClagId,True ),
            "setanycastip"    : (self.HandleCmdSetAnycastIp, True ),
            "setbackupip"     : (self.HandleCmdSetBackupIp,  True ),
            "reloaddone"      : (self.HandleCmdReloadDone,   True ),
            "debug"           : (self.HandleCmdDebug,        True ),
            "setverbose"      : (self.HandleCmdVerbose,      True ),
            "setquiet"        : (self.HandleCmdQuiet,        True ),
            "setlogfile"      : (self.HandleCmdLogfile,      True ),
            "setlacppoll"     : (self.HandleCmdLacpPoll,     True ),
            "setpeertimeout"  : (self.HandleCmdPeerTimeout,  True ),
            "setsendtimeout"  : (self.HandleCmdSendTimeout,  True ),
            "setsendbufsize"  : (self.HandleCmdSendBufSize,  True ),
            "setpriority"     : (self.HandleCmdPriority,     True ),
            "setpeerlinkpoll" : (self.HandleCmdPeerLinkPoll, True ),
            "getdualifs"      : (self.HandleCmdDualIfs,      False),
            "getidmatchifs"   : (self.HandleCmdIdMatchIfs,   False),
            "getpeeralive"    : (self.HandleCmdPeerAlive,    False),
            "getclagid"       : (self.HandleCmdClagId,       False),
            "getclagrole"     : (self.HandleCmdClagRole,     False),
            "getpeerid"       : (self.HandleCmdPeerId,       False),
            "getpeerrole"     : (self.HandleCmdPeerRole,     False),
            "getbackuppeerid" : (self.HandleCmdBackupPeerId, False),
            "getbackuppeerrole" : (self.HandleCmdBackupPeerRole, False),
            "getpeerif"       : (self.HandleCmdPeerIf,       False),
            "getpeerip"       : (self.HandleCmdPeerIp,       False),
            "getbackupip"     : (self.HandleCmdGetBackupIp,  False),
            "getbackupactive" : (self.HandleCmdBackupActive, False),
            "getbackuprole"   : (self.HandleCmdBackupRole,   False),
            "getsysmac"       : (self.HandleCmdSysMac,       False),
            "getourlacpdb"    : (self.HandleCmdOurLacpDB,    False),
            "getpeerlacpdb"   : (self.HandleCmdPeerLacpDB,   False),
            "getpeerlacprate" : (self.HandleCmdPeerLacpRate, False),
            "getclagbonddb"   : (self.HandleCmdClagBondDB,   False),
            "getclagintfdb"   : (self.HandleCmdClagIntfDB,   False),
            "getourmacdb"     : (self.HandleCmdOurMacDB,     False),
            "getpeermacdb"    : (self.HandleCmdPeerMacDB,    False),
            "getpermanentmacdb" : (self.HandleCmdPermanentMacDB, False),
            "getpermanentneighdb" : (self.HandleCmdPermanentNeighDB, False),
            "getourmcastdb"   : (self.HandleCmdOurMdbDB,     False),
            "getpeermcastdb"  : (self.HandleCmdPeerMdbDB,    False),
            "getneighdb"      : (self.HandleCmdNeighDB,      False),
            "getkernelneighdb": (self.HandleCmdKernNeighDB,  False),
            "getourrportdb"   : (self.HandleCmdOurRportDB,   False),
            "getpeerrportdb"  : (self.HandleCmdPeerRportDB,  False),
            "getpeervlandb"   : (self.HandleCmdPeerVlanDB,   False),
            "getourvlandb"    : (self.HandleCmdOurVlanDB,    False),
            "getnllinkcache"  : (self.HandleCmdNlLinkCache,  False),
            "getintinfo"      : (self.HandleCmdIntInfo,      False),
            "getsummary"      : (self.HandleCmdSummary,      False),
            "getbackupinfo"   : (self.HandleCmdBackupInfo,   False),
            "getqinfo"        : (self.HandleCmdGetQInfo,     False),
            "getdbchecksum"   : (self.HandleCmdGetDBCheckSum, False),
            "getobjectinfo"   : (self.HandleCmdGetObjectInfo, False),
            "gettimers"       : (self.HandleCmdGetTimers, False),
            "logmsg"          : (self.HandleCmdLogMsg,       True ),
            "collectgarbage"  : (self.HandleCmdGarbage,      True ),
            "collectobject"   : (self.HandleCmdCollectObjectCount,  True), 
            "collectuncollectableobject" : (self.HandleCmdUnCollectableObject,  True),
            "getall"          : (self.HandleCmdGetAll,       False),
            "dumpmstpstate"   : (self.HandleCmdDumpMstp,     False),
            "nogetall"        : (self.HandleCmdNoGetAll,     False),
            "startprofile"    : (self.HandleStartProfile,    False),
            "stopprofile"     : (self.HandleStopProfile,     False),
            "dumpprofile"     : (self.HandleDumpProfile,     False),
            "getsocketinfo"   : (self.HandleCmdGetSocket,    False),
            "getdebugflags"   : (self.HandleCmdGetDebugFlags,   True),
            "setdebugflags"   : (self.HandleCmdSetDebugFlags,   True),
            "cleardebugflags" : (self.HandleCmdClearDebugFlags, True),
            "setfastshutdown" : (self.HandleCmdSetFastShutDown, True),
            "setvxlanlocalip" : (self.HandleCmdSetVxlanLocalIp, True),
        }

        ucredStr = ourSock.getsockopt(socket.SOL_SOCKET, cmdsrv._SOCKET_PEERCRED, struct.calcsize("3I"))
        (peerPid, peerUid, peerGid) = struct.unpack("3I", ucredStr)
        canModify = peerUid == 0

        termChar = chr(0)
        cmdList = cmdStr.split()
        if cmdList[0].lower() in clagCmds:
            if clagCmds[cmdList[0].lower()][1] and not canModify:
                outStr = "ERROR You do not have sufficient privilege to execute this command: %s" % (cmdStr,)
            else:
                (outStr, termChar) = clagCmds[cmdList[0].lower()][0](nlm, cmdStr, ourSock)
        else:
             outStr = "ERROR Unknown command: %s" % (cmdStr,)

        self.SendSock(ourSock, outStr + termChar, False)

