#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014-2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import clag.util as util
    import copy
    import ipaddr
    import json
    import nlmanager.nlmanager
    import nlmanager.nlpacket
    import os
    import Queue
    import socket
    import struct
    import threading
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class fdbsync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the Bridge Forwarding
    Database, or FDB, between two CLAG peers.
    """

    # Hardware work queue operation codes.
    _MAC_ADD            = 1
    _MAC_DELETE         = 2
    _MAC_REFRESH        = 3
    _MAC_MOVE           = 4
    _MAC_STATIC_ADD     = 5
    _MAC_STATIC_DEL     = 6

    # FDB Work Queue operation codes
    _WORK_QUEUE_ADD            = 1     # Params: (mac, vlan, if, vni, dst, master)
    _WORK_QUEUE_DELETE         = 2     # Params: (mac, vlan, if, vni, dst, master)
    _WORK_QUEUE_REFRESH        = 3     # Params: ()
    _WORK_QUEUE_NOP            = 4     # Params: ()
    _WORK_QUEUE_PERM_ADD       = 5     # Params: (mac, vlan, if, vni, dst, master)
    _WORK_QUEUE_PERM_DELETE    = 6     # Params: (mac, vlan, if, vni, dst, master)
    _WORK_QUEUE_STATIC_ADD     = 7     # Params: (mac, vlan, if, vni, dst, master)
    _WORK_QUEUE_STATIC_DELETE  = 8     # Params: (mac, vlan, if, vni, dst, master)

    _RECENT         = 10

    def __init__(self, Daemon, Intf, Parser, Log, sendQueue, clagnl):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(fdbsync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon     = Daemon
        self.Intf       = Intf
        self.Parser     = Parser
        self.Log        = Log
        self.sendQueue  = sendQueue
        self.clagnl     = clagnl

        self.peerIf = self.Parser.args.peerIf
        if self.Intf.isSubIfName(self.peerIf):
            self.peerIf = self.Intf.GetMasterOfSubIf(self.peerIf)

        #
        #   Mac Database - The MAC addresses which are present on an interface
        #   of a bridge, are tracked in this database, which is a dictionary.
        #   The key is a tuple containing the MAC address and VLAN ID. On a VLAN
        #   unaware bridge, the VLAN ID will be the peer interface's VLAN ID.
        #   The value of each entry is a list containing six elements: the
        #   local "base" interface name on which the address exists, the remote
        #   "base" interface name on which the address exists or None for
        #   locally learned addresses, the time the entry was last updated, the
        #   time the entry was added, the local IP for VxLAN or None for non-
        #   VxLAN interfaces and the VNI for VxLAN or 0xFFFFFFFF for non-VxLAN.
        #
        #   {
        #                MAC          Vlan       localIf  remoteIf  updated added dst   vni
        #       ("00:00:00:00:00:02", 101 )  : [ "bond1", "bond11",   14,    30,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:03", 101 )  : [ "bond2", None,        0,    10,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:04", 102 )  : [ "bond3", "bond13",    0,    10,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:05", 102 )  : [ "bond4", None,        0,    12,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:06", 101 )  : [ "vx-75", None,      100,   500,  "36.0.0.1",   75 ]
        #   }
        #
        self.macDBLock  = threading.Lock()
        self.macDB      = {}
        self.syncDB     = {}

        # Permanent MAC Database - SVIs permanent MACs are synced to the peer and
        # programmed as static entries pointing to the peerlink on the peer
        # The permanent flag is set when a permanent entry is owned by the local
        # switch and synced to the peer, the static flag is set when an entry is
        # permanent on the peer and programmed as a static entry on the local
        # switch. There are some cases where the entry is permanent on both clag
        # peers, in that case, we don't try to overwrite those entries and we
        # hide those entries (permanent and static flag would both be set),
        # when dumping the permanentMacDB to avoid confusion.
        #
        #   {
        #                MAC          Vlan      localIf    permanent static
        #       ("00:00:00:00:00:13", 11 )  : [ "br0",      True,    False ],
        #       ("00:00:00:00:00:13", 12 )  : [ "br0",      True,    False ],
        #       ("00:00:00:00:00:13", 13 )  : [ "br0",      True,    False ],
        #       ("00:00:00:00:00:19", 11 )  : [ "peerlink", False,   True ],
        #       ("00:00:00:00:00:19", 12 )  : [ "peerlink", False,   True ],
        #       ("00:00:00:00:00:19", 13 )  : [ "peerlink", False,   True ]
        #   }
        #
        self.permanentMacDBLock = threading.Lock()
        self.permanentMacDB     = {}

        #
        #   Hardware Work Queue - This queue contains MAC addresses which need
        #   to be added or deleted to/from the kernel/hardware.  Each entry is a
        #   tuple containing an operation, the MAC address, a VLAN Id, the
        #   "base" interface, and the destination IP address (for vxlan).
        #
        #       ( _MAC_ADD, "00:00:00:00:00:02", 101, "bond1", None )
        #       ( _MAC_ADD, "00:00:00:00:00:02", 102, "vx-10", "36.0.0.1" )
        #
        self.hwWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.hwWorkQueue, 'fdbHwWorkQueue')

        #
        #   Remote Delete Queue - This queue contains remote MAC addresses which
        #   have  been requested by the peer switch to be deleted. Before being
        #   deleted, we need to check if we have seen the address recently. The
        #   format of each queue element is a tuple containing the MAC address,
        #   VLAN Id, vni, "base" interface, and local IP for VxLAN.
        #
        #       ( "00:00:00:00:00:02", 101, 0xFFFFFFF, "bond1", None )
        #
        self.remoteDeleteQueue = Queue.Queue()
        util.queueAddNameMap(self.remoteDeleteQueue, 'fdbRemoteDeleteQueue')

        #
        #   FDB Work Queue - This queue contains fdb operations that are
        #   awaiting processing. Each queue element is a two-element tuple: an
        #   opcode and parameters. The opcode is one of the values, above, and
        #   the parameters are a tuple whose format is dependent on the opcode.
        #
        self.fdbWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.fdbWorkQueue, 'fdbWorkQueue')
        self.clagnl.SetFdbSync(self)

        #
        #   Start threads to handle hardwre updates, kernel notifications, etc.
        #
        self.keep_going = True
        self.stopEvent  = clag.clagthread.Event()

        self.refreshMacsThread = threading.Thread(None, self.RefreshMacsT, "RefreshMacs")
        self.refreshMacsThread.daemon = True
        self.refreshMacsThread.start()

        self.updateHwThread = threading.Thread(None, self.UpdateMacsInHwT, "UpdateMacsInHw")
        self.updateHwThread.daemon = True
        self.updateHwThread.start()

        self.remoteDeleteThread = threading.Thread(None, self.RemoteDeleteT, "RemoteDelete")
        self.remoteDeleteThread.daemon = True
        self.remoteDeleteThread.start()

        self.kernelMsgThread = threading.Thread(None, self.FdbKernelMsgT, "FdbKernelMsg")
        self.kernelMsgThread.daemon = True
        self.kernelMsgThread.start()

    def __del__(self):
        """
        Destructor:
        """
        self.SignalThreads()
        self.JoinThreads()

    def RefreshMacsT(self):
        try:
            self.Log.log("[%d]Init RefreshMacsT" % (self.Parser.getTid()))
            self.RefreshMacs()
        except:
            self.Daemon.DumpTraceback()

    def UpdateMacsInHwT(self):
        try:
            self.Log.log("[%d]Init UpdateMacsInHwT" % (self.Parser.getTid()))
            self.UpdateMacsInHw()
        except:
            self.Daemon.DumpTraceback()

    def RemoteDeleteT(self):
        try:
            self.Log.log("[%d]Init RemoteDeleteT" % (self.Parser.getTid()))
            self.RemoteDelete()
        except:
            self.Daemon.DumpTraceback()

    def FdbKernelMsgT(self):
        try:
            self.Log.log("[%d]Init FdbKernelMsgT" % (self.Parser.getTid()))
            self.FdbKernelMsg()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        try:
            self.hwWorkQueue.put_nowait((None, None, None, None, None))
            util.queueSetMaxLen('fdbHwWorkQueue')
        except Queue.Full:
            pass
        try:
            self.remoteDeleteQueue.put_nowait((None, None, None, None, None))
            util.queueSetMaxLen('fdbRemoteDeleteQueue')
        except Queue.Full:
            pass
        try:
            self.fdbWorkQueue.put_nowait((fdbsync._WORK_QUEUE_NOP, ()))
            util.queueSetMaxLen('fdbWorkQueue')
        except Queue.Full:
            pass

    def JoinThreads(self):
        self.refreshMacsThread.join()
        self.updateHwThread.join()
        self.remoteDeleteThread.join()
        self.kernelMsgThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_FDB_SYNC, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_FDB_SYNC, logStr)

    def HwWorkQueuePut(self, element):
        if self.keep_going:
            self.hwWorkQueue.put(element)
            util.queueSetMaxLen('fdbHwWorkQueue')

    def RemoteDeleteQueuePut(self, element):
        if self.keep_going:
            self.remoteDeleteQueue.put(element)
            util.queueSetMaxLen('fdbRemoteDeleteQueue')

    def GetObjectSize(self):
        macObj = {}
        self.macDBLock.acquire()
        macObj['macDB'] = util.getSize(self.macDB)
        self.macDBLock.release()
        return macObj

    def __str__(self):
        """
        Return local information in human-readable form
        """
        macDb = []
        self.macDBLock.acquire()
        for (mac, vlanId), [lif, rif, updated, added, dst, vni] in self.macDB.iteritems():
            if not rif:
                macInfo = {}
                macInfo['outIf'] = lif
                macInfo['vlanId'] = vlanId
                macInfo['mac'] = mac
                if dst:
                    macInfo['dst'] = dst
                if vni != 0xFFFFFFFF:
                    macInfo['vni'] = vni
                macDb.append(macInfo)
        self.macDBLock.release()
        return json.dumps(macDb, sort_keys=True, indent=4)

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        macDb = []
        self.macDBLock.acquire()
        for (mac, vlanId), [lif, rif, updated, added, dst, vni] in self.macDB.iteritems():
            if rif:
                macInfo = {}
                macInfo['outIf'] = lif
                macInfo['vlanId'] = vlanId
                macInfo['mac'] = mac
                if dst:
                    macInfo['dst'] = dst
                if vni != 0xFFFFFFFF:
                    macInfo['vni'] = vni
                macDb.append(macInfo)
        self.macDBLock.release()
        return json.dumps(macDb, sort_keys=True, indent=4)

    def EncodedPermStr(self):
        """
        Return the permanent macs information in human-readable form
        """
        macDB = []
        self.permanentMacDBLock.acquire()
        for (mac, vlanId), [lif, permanent, static] in self.permanentMacDB.iteritems():
            macInfo = {}
            macInfo['iface'] = lif
            macInfo['vlanId'] = vlanId
            macInfo['mac'] = mac
            macInfo['permanent'] = permanent
            macInfo['static'] = static
            macDB.append(macInfo)
        self.permanentMacDBLock.release()
        return json.dumps(macDB, sort_keys=True, indent=4)

    def CollectLocalInfo(self, nlm):
        """
        This function is called on a periodic interval to gather up any
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """
        pass

    def GetDualIfs(self):
        if self.Parser.args.redirectEnable:
            dualIfs = self.Intf.GetClagIdMatchIfs()
        else:
            dualIfs = self.Intf.GetDualIfs()
        return dualIfs

    def RefreshMacs(self):
        """
        This function executes periodically to refresh the MAC age of remote
        addresses.
        """
        self.log_verbose("Beginning execution of the thread RefreshMacs")
        lastAgeRefresh = int(os.times()[4])
        while self.keep_going:

            shortestMaxAge = 300
            for br in self.Intf.GetAllBridges():
                maxAge = self.Intf.GetBridgeAgeingTime(br)
                if maxAge is not None and maxAge < shortestMaxAge:
                    shortestMaxAge = maxAge

            now = int(os.times()[4])
            if (now - lastAgeRefresh) > (shortestMaxAge * 0.45):
                self.log_debug("Adding entry to the fdbWorkQueue: %s" % ("DoRefreshMacs",))
                self.fdbWorkQueue.put((fdbsync._WORK_QUEUE_REFRESH, ()))
                lastAgeRefresh = now

            self.log_verbose("Waiting %d seconds before continuing execution of the thread RefreshMacs" % (self.Parser.getPeriodicRun(),))
            self.stopEvent.wait(self.Parser.getPeriodicRun())
            self.log_verbose("Executing the thread RefreshMacs")

        self.log_verbose("Finished execution of the thread RefreshMacs")

    def DoRefreshMacs(self, nlm, now, lastAgeRefresh):
        self.log_verbose("Beginning RefreshMacs")
        dualIfs = self.GetDualIfs()
        k_macDB = self.GetKernelFdb(nlm)
        self.SyncCurrentState(k_macDB)
        self.macDBLock.acquire()
        c_macDB = copy.deepcopy(self.macDB)
        self.macDBLock.release()

        for (k_mac, k_vlanId), [k_if, k_age, k_dst, k_vni, k_bif] in k_macDB.items():
            [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = c_macDB.get((k_mac, k_vlanId), [None, None, None, None, None, None])
            if c_rif and (now - lastAgeRefresh) < (k_age + fdbsync._RECENT):
                if k_if not in dualIfs:
                    k_if = self.peerIf
                self.log_verbose("Refreshing age of remote address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, k_if, k_dst))
                self.macDBLock.acquire()
                self.macDB[(k_mac, k_vlanId)] = [k_if, c_rif, now, c_added, k_dst, k_vni]
                self.macDBLock.release()
                self.HwWorkQueuePut((fdbsync._MAC_REFRESH, k_mac, k_vlanId, k_if, k_dst))
            c_macDB.pop((k_mac, k_vlanId), None)

        # Any addresses left are missed kernel deletes, re-add them to the kernel
        for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in c_macDB.items():
            if c_rif:
                self.log_debug("Sync re-adding removed remote address %s vlan %s vni %s dev %s dst %s" % (c_mac, c_vlanId, c_vni, c_lif, c_dst))
                self.macDBLock.acquire()
                self.macDB[(c_mac, c_vlanId)] = [c_lif, c_rif, now, c_added, c_dst, c_vni]
                self.macDBLock.release()
                self.HwWorkQueuePut((fdbsync._MAC_ADD, c_mac, c_vlanId, c_lif, c_dst))

        self.log_verbose("Ending RefreshMacs")

    def SyncCurrentState(self, k_macDB):
        '''
        Given the current state of the FDB, make sure that the clag database
        is in sync with it. Also, make sure that the peer is in sync with these
        addresses.
        '''
        self.log_verbose("Beginning sync of local addresses")

        self.macDBLock.acquire()
        c_macDB = copy.deepcopy(self.macDB)
        self.macDBLock.release()
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        ET.SubElement(root, "fdb-start-sync")
        xmlEntries = 0

        now = int(os.times()[4])
        for (k_mac, k_vlanId), [k_if, k_age, k_dst, k_vni, k_bif] in k_macDB.items():
            [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = c_macDB.get((k_mac, k_vlanId), [None, None, now, now, None, None])
            # Is it a local address, on an interface with learning enabled?
            if not c_rif and self.Intf.isBridgeLearning(k_bif) and self.isMacNeedPeerSync(k_vlanId):
                # Update our clag database with this address
                self.log_debug("Syncing local address %s vlan %s vni %s dev %s dst %s, c_lif %s" % (k_mac, k_vlanId, k_vni, k_if, k_dst, c_lif))
                self.macDBLock.acquire()
                self.macDB[(k_mac, k_vlanId)] = [k_if, None, c_updated, c_added, k_dst, k_vni]
                self.macDBLock.release()
                c_macDB.pop((k_mac, k_vlanId), None)

                # Inform the peer of this address.
                ET.SubElement(root, "fdb", {"op" : "sync", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                xmlEntries += 1
                if xmlEntries >= 2000:
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)
                    root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                    xmlEntries = 0

        # Any addresses left are missed kernel deletes.
        self.macDBLock.acquire()
        if len(c_macDB):
            brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
        for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in c_macDB.items():
            brIf = self.Intf.GetBridgeOfMember(c_lif)
            if not c_rif and (self.Intf.isVlanFilteringBridge(brIf) or brIf in brPeerVlanMap):
                self.log_debug("Sync removing local address %s vlan %s vni %s dev %s dst %s" % (c_mac, c_vlanId, c_vni, c_lif, c_dst))
                self.macDB.pop((c_mac, c_vlanId), None)
        self.macDBLock.release()

        ET.SubElement(root, "fdb-stop-sync")
        xmlStr = ET.tostring(root)
        self.sendQueue.put(xmlStr)

        self.log_verbose("Ending sync of local addresses")

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML for sending to the peer on a
        periodic basis.
        """
        if self.Parser.syncDoneToPeer:
            return

        self.permanentMacDBLock.acquire()
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0
        for (mac, vlanId), [lif, permanent, static] in self.permanentMacDB.iteritems():
            if permanent:
                ET.SubElement(root, "fdb-permanent", {"op" : "permanent-add", "mac" : mac, "vlanId" : str(vlanId), "if" : lif}) # if not need?
                xmlEntries += 1
            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
        self.permanentMacDBLock.release()

        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(root)
            self.sendQueue.put(xmlStr)
        self.log_debug("Initial permanent mac sync to peer done.")

    def PrintElement(self, node, level):
        self.log_debug("%s%s: %s: %s" % (" " * (level*3), node.tag, node.text, str(node.attrib)))
        for child in node.findall('*'):
            self.PrintElement(child, level+1)

    def GetWorkQueueDelayed(self, delay=0.001):
        '''
        Returns the next entry on the FDB work queue, waiting up to the specified
        delay (default 1 millisecond) for an entry.
        '''
        queueEmpty = False
        try:
            (opcode, params) = self.fdbWorkQueue.get(True, delay)
        except Queue.Empty:
            opcode = fdbsync._WORK_QUEUE_NOP
            params = ()
            queueEmpty = True
        return (opcode, params, queueEmpty)

    def isExternalDev(self, k_if, k_bif):
        """
        Some interfaces are outside clagd's domain. If a mac is installed
        in the kernel on such a interface that mac should be removed from
        clagd's macDb and treated as an "external mac".

        A vxlan/network port in a EVPN topology is one such "external interface"
        """
        return k_if != self.peerIf and not self.Intf.isBridgeLearning(k_bif)

    def isMacNeedPeerSync(self, vlan):
        """
        It decides whether mac learnt on vlan need to be synced with the peer:
        if peer interface is part of this vlan, then it need to sync the mac
        on this vlan with the peer. This will handle the scenario
        of peer interface in clag vlans and as a normal trunk interface for non
        clag vlans. This will prevent syncing of any macs on vlans in which peer
        interface is not part of it.
        """
        vlanIfs = self.Intf.GetMembersOfVlan(vlan)
        if vlanIfs:
            # vlan aware case
            peerIf = self.Parser.args.peerIf
            if self.Intf.isSubIfName(peerIf):
                peerIf = self.Intf.GetMasterOfSubIf(peerIf)
            return True if peerIf in vlanIfs else False
        else:
            # vlan unaware case
            peer_vlans = self.Intf.GetBridgePeerVlanMap()
            return True if vlan in peer_vlans.values() else False

    def isValidMacEntry(self, k_add, vlanMap, brPeerVlanMap, pvidMap, params):
        (k_mac, k_vlanId, k_bif, k_vni, k_dst, k_master) = params
        if k_bif is None:
            self.log_debug("Ignoring MAC %s because there is no interface" % (k_mac,))
            return (False, None, None, None, None)
        k_if = vlanMap.get(k_bif, (k_bif, None))[0]

        # Get remote IP address (vxlan)
        # XXX: Check if dst/vni is always included for vxlan interaces
        vxlanInfo = self.Parser.GetVxLanInfo(k_if)
        if k_dst is not None:
            k_dst = str(k_dst)
            if not vxlanInfo or not k_add:
                self.log_debug("Ignoring MAC %s because there is no vxlan information or this is a delete from the vxlan driver (k_add=%s)" % (k_mac, str(k_add)))
                return (False, None, None, None, None)
            else:
                k_vni = vxlanInfo[0]
        else:
            if vxlanInfo:
                if k_add: # Ignore VxLAN on bridge driver for adds
                    self.log_debug("Ignoring MAC %s because this is an add on the bridge driver" % (k_mac,))
                    return (False, None, None, None, None)
                else:
                    k_dst = vxlanInfo[1]
                    k_vni = vxlanInfo[0]
            else:
                k_vni = 0xFFFFFFFF

        # Get the vlanId
        if k_vlanId is None:
            k_vlanId = 0 if not k_dst else 0xFFFF

        if not k_vlanId:
            if not k_master:
                self.log_debug("Ignoring MAC %s because there is no vlanid and no master interface" % (k_mac,))
                return (False, None, None, None, None)
            k_vlanId = brPeerVlanMap.get(k_master, None)
            if k_vlanId is None:
                self.log_debug("Ignoring MAC %s because there is no vlanid and no peerlink vlanid for master %s" % (k_mac, k_master))
                return (False, None, None, None, None)
        elif k_vlanId == 0xFFFF:
            k_master = self.Intf.GetBridgeOfMember(k_if)
            k_vlanId = brPeerVlanMap.get(k_master, pvidMap.get(k_if))
            if k_vlanId is None:
                self.log_debug("Ignoring MAC %s because there is no vlanid and no peerlink vlanid for master %s on vxlan" % (k_mac, k_master))
                return (False, None, None, None, None)

        if k_add == fdbsync._WORK_QUEUE_PERM_ADD or k_add == fdbsync._WORK_QUEUE_PERM_DELETE:
            if k_add == fdbsync._WORK_QUEUE_PERM_ADD:
                self.macDBLock.acquire()
                if (k_mac, k_vlanId) in self.macDB:
                    self.log_debug("Permanent Mac %s vlan %s exists in macDB(dynamic). Deleting the entry" % (k_mac, k_vlanId))
                    del self.macDB[(k_mac, k_vlanId)]
                self.macDBLock.release()
            return (True, k_vlanId, k_if, k_vni, k_dst)

        if k_add and self.isExternalDev(k_if, k_bif):
            # If DYNAMIC mac add is happening on an externalIntf (such as vxlan
            # network port in an evpn setup) we remove the entry from the
            # clagd macDb (if one exists).
            self.macDBLock.acquire()
            if (k_mac, k_vlanId) in self.macDB:
                self.log_debug("Removing existing address %s vlan %s; now it is ext" % (k_mac, k_vlanId))
                del self.macDB[(k_mac, k_vlanId)]
            else:
                self.log_debug("Ignoring MAC %s %s as it is ext" % (k_mac, k_vlanId))
            self.macDBLock.release()
            return (False, None, None, None, None)

        # If we have received mac address addition for an entry which is already in permanent mac Db. 
        # Remove the permanent Entry
        if k_add == fdbsync._WORK_QUEUE_ADD:
            self.permanentMacDBLock.acquire()
            if (k_mac, k_vlanId) in self.permanentMacDB:
                self.log_debug("Dynamic address %s vlan %s received which already has existing permanent entry. Deleting permanent fdb entry" % (k_mac, k_vlanId))
                del self.permanentMacDB[(k_mac, k_vlanId)]
            self.permanentMacDBLock.release()
        return (True, k_vlanId, k_if, k_vni, k_dst)

    def FdbKernelMsg(self):
        """
        This is an independent thread of the script. It is responsible for
        processing entries on the fdbWorkQueue, which may result in modifications
        to the macDB and messages sent to the peer.
        """
        self.log_verbose("Beginning execution of the thread FdbKernelMsg")
        if self.clagnl.nlm is None:
            self.clagnl.kernelDumpReadyEvent.wait()

        self.clagnl.DumpMacs()
        nlm = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_KERNEL_MSG)

        lastAgeRefresh = int(os.times()[4])
        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution of the thread FdbKernelMsg")
            (opcode, params) = self.fdbWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                vlanMap = self.Intf.GetSubIfsVlanMap()
                brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
                pvidMap = self.Intf.GetPvidMap()
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
                peerDown = self.Intf.GetOperStateFromSysFs(self.peerIf) != "up"
                while not queueEmpty and self.keep_going:
                    if opcode == fdbsync._WORK_QUEUE_REFRESH:
                        now = int(os.times()[4])
                        self.DoRefreshMacs(nlm, now, lastAgeRefresh)
                        lastAgeRefresh = now

                    elif opcode == fdbsync._WORK_QUEUE_PERM_ADD or opcode == fdbsync._WORK_QUEUE_PERM_DELETE:
                        self.log_debug("Processing permanent entry: %d %s" % (opcode, str(params)))
                        k_add = opcode
                        (k_mac, k_vlanId, k_bif, k_vni, k_dst, k_master) = params
                        (valid, k_vlanId, k_if, k_vni, k_dst) = self.isValidMacEntry(k_add, vlanMap, brPeerVlanMap, pvidMap, params)

                        if not k_vlanId or k_vlanId == 0xFFFF or k_mac == self.Parser.args.sysMac:
                            (opcode, params, queueEmpty) = self.GetWorkQueueDelayed()
                            continue
                        if opcode == fdbsync._WORK_QUEUE_PERM_ADD:
                            self.permanentMacDBLock.acquire()
                            if (k_mac, k_vlanId) not in self.permanentMacDB:
                                self.log_debug("Adding permanent MAC %s %s %s" % (str(k_mac), k_bif, str(k_vlanId)))
                                self.permanentMacDB[(k_mac, k_vlanId)] = [k_bif, True, False]
                                ET.SubElement(root, "fdb-permanent", {"op" : "permanent-add", "mac" : k_mac, "vlanId" : str(k_vlanId)})
                                xmlEntries += 1
                            elif self.permanentMacDB[(k_mac, k_vlanId)][1] != True:
                                self.log_debug("Adding permanent MAC %s %s %s" % (str(k_mac), k_bif, str(k_vlanId)))
                                self.permanentMacDB[(k_mac, k_vlanId)][1] = True
                                ET.SubElement(root, "fdb-permanent", {"op" : "permanent-add", "mac" : k_mac, "vlanId" : str(k_vlanId)})
                                xmlEntries += 1
                            self.permanentMacDBLock.release()
                        else:
                            self.permanentMacDBLock.acquire()
                            if (k_mac, k_vlanId) in self.permanentMacDB:
                                if self.permanentMacDB[(k_mac, k_vlanId)][2] != True:
                                    self.log_debug("Deleting permanent MAC %s %s %s" % (str(k_mac), k_bif, str(k_vlanId)))
                                    del self.permanentMacDB[(k_mac, k_vlanId)]
                                else:
                                    self.permanentMacDB[(k_mac, k_vlanId)][1] = False
                                    self.log_debug("Re-adding as static MAC %s %s %s" % (str(k_mac), k_if, str(k_vlanId)))
                                    self.HwWorkQueuePut((fdbsync._MAC_STATIC_ADD, k_mac, k_vlanId, self.peerIf, None))

                            self.permanentMacDBLock.release()
                            ET.SubElement(root, "fdb-permanent", {"op" : "permanent-del", "mac" : k_mac, "vlanId" : str(k_vlanId)})
                            xmlEntries += 1

                    elif opcode == fdbsync._WORK_QUEUE_STATIC_ADD or opcode == fdbsync._WORK_QUEUE_STATIC_DELETE:
                        self.log_debug("Processing static entry: %d %s" % (opcode, str(params)))
                        (k_mac, k_vlanId, k_bif, k_vni, k_dst, k_master) = params
                        if k_bif is None:
                            (opcode, params, queueEmpty) = self.GetWorkQueueDelayed()
                            self.log_verbose("Ignoring MAC %s because there is no interface" % (k_mac,))
                            continue
                        k_if = vlanMap.get(k_bif, (k_bif, None))[0]
                        self.permanentMacDBLock.acquire()
                        if (k_mac, k_vlanId) in self.permanentMacDB:
                            if opcode == fdbsync._WORK_QUEUE_STATIC_ADD:
                                self.log_debug("Adding static MAC %s %s %s" % (str(k_mac), k_if, str(k_vlanId)))
                                # update the permanentMacDB, now static
                                self.permanentMacDB[(k_mac, k_vlanId)][2] = True
                            else:
                                # deleted static entry still in the DB, should be re-added
                                self.log_debug("Re-adding static MAC %s %s %s" % (str(k_mac), k_if, str(k_vlanId)))
                                self.HwWorkQueuePut((fdbsync._MAC_STATIC_ADD, k_mac, k_vlanId, self.peerIf, None))
                        self.permanentMacDBLock.release()

                    elif opcode == fdbsync._WORK_QUEUE_ADD or opcode == fdbsync._WORK_QUEUE_DELETE:
                        self.log_debug("Processing entry: %d %s" % (opcode,str(params)))
                        k_add = (opcode == fdbsync._WORK_QUEUE_ADD)

                        # VLAN Unaware:
                        # 00:aa:00:00:00:01 dev vx-14567103 dst 36.0.0.13 vlan 65535 self
                        # 00:aa:00:00:00:01 dev vx-14567103 vlan 0 master br-14567103
                        # 00:aa:00:00:00:01 dev t2h_c-4.1010 vlan 0 master br-14567103
                        #
                        # VLAN Aware:
                        # 00:aa:00:00:00:01 dev vx-14567103 dst 36.0.0.13 vlan 0 self
                        # 00:aa:00:00:00:01 dev vx-14567103 vlan 1003 master vx_aware-1
                        # 00:aa:00:00:00:02 dev t2h_c-4 vlan 1003 master vx_aware-1

                        (k_mac, k_vlanId, k_bif, k_vni, k_dst, k_master) = params
                        (valid, k_vlanId, k_if, k_vni, k_dst) = self.isValidMacEntry(k_add, vlanMap, brPeerVlanMap, pvidMap, params)

                        if not valid:
                            (opcode, params, queueEmpty) = self.GetWorkQueueDelayed()
                            continue

                        if not self.isMacNeedPeerSync(k_vlanId):
                            (opcode, params, queueEmpty) = self.GetWorkQueueDelayed()
                            continue

                        self.macDBLock.acquire()
                        now = int(os.times()[4])
                        # Is this a new address we don't know anything about yet?
                        if (k_mac, k_vlanId) not in self.macDB:
                            if k_add and self.Intf.isBridgeLearning(k_bif):
                                self.log_debug("Adding new local address %s vlan %s vni %s dev %s" % (k_mac, k_vlanId, k_vni, k_if))
                                self.macDB[(k_mac, k_vlanId)] = [k_if, None, now, now, k_dst, k_vni]
                                ET.SubElement(root, "fdb", {"op" : "add", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                                xmlEntries += 1
                        else:
                            # We've seen this address before. Is it a local address?
                            [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = self.macDB[(k_mac, k_vlanId)]

                            if not c_rif:
                                if k_add:
                                    # Is the address moving to a different port?
                                    if k_if != c_lif or k_dst != c_dst:
                                        if k_if == self.peerIf:
                                            if (now - c_added) > fdbsync._RECENT:
                                                self.log_debug("Removing existing address (moved to peer) %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                                del self.macDB[(k_mac, k_vlanId)]
                                            else:
                                                self.log_debug("Re-adding existing local address %s vlan %s vni %s dev %s" % (k_mac, k_vlanId, k_vni, c_lif))
                                                self.macDB[(k_mac, k_vlanId)] = [c_lif, None, c_updated, c_added, c_dst, c_vni]
                                                self.HwWorkQueuePut((fdbsync._MAC_ADD, k_mac, k_vlanId, c_lif, c_dst))
                                        else:
                                            self.log_debug("Moving existing local address %s vlan %s vni %s dst %s from %s to %s" % (k_mac, k_vlanId, k_vni, k_dst, c_lif, k_if))
                                            self.macDB[(k_mac, k_vlanId)] = [k_if, None, now, now, k_dst, k_vni]
                                            ET.SubElement(root, "fdb", {"op" : "add", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                                            xmlEntries += 1
                                else:
                                    self.log_debug("Removing existing local address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                    del self.macDB[(k_mac, k_vlanId)]
                                    ET.SubElement(root, "fdb", {"op" : "delete", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : c_lif, "dst" : str(c_dst)})
                                    xmlEntries += 1
                            else:
                                # Remote address.
                                if k_add:
                                    if (now - c_added) > fdbsync._RECENT and k_if != self.peerIf:
                                        self.log_debug("Moving existing remote address to local %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, k_if, k_dst))
                                        self.macDB[(k_mac, k_vlanId)] = [k_if, None, c_updated, c_added, k_dst, k_vni]
                                        ET.SubElement(root, "fdb", {"op" : "add", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                                        xmlEntries += 1
                                    else:
                                        if k_if != c_lif:
                                            self.log_debug("Re-adding existing remote address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                            self.macDB[(k_mac, k_vlanId)] = [c_lif, c_rif, c_updated, c_added, c_dst, c_vni]
                                            self.HwWorkQueuePut((fdbsync._MAC_ADD, k_mac, k_vlanId, c_lif, c_dst))
                                else:
                                    if k_if == self.peerIf and peerDown:
                                        self.log_debug("Removing remote address peer-down %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                        del self.macDB[(k_mac, k_vlanId)]
                                    else:
                                        if self.Intf.GetOperStateFromSysFs(k_bif) == "up":
                                            self.log_debug("Re-adding deleted remote address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                            self.macDB[(k_mac, k_vlanId)] = [c_lif, c_rif, now, c_added, c_dst, c_vni]
                                            self.HwWorkQueuePut((fdbsync._MAC_ADD, k_mac, k_vlanId, c_lif, c_dst))
                                        else:
                                            if k_vni == 0xFFFFFFFF:
                                                self.log_debug("Remote address, interface down, move address %s vlan %s vni %s dev %s to the peer" % (k_mac, k_vlanId, c_vni, c_lif))
                                                self.macDB[(k_mac, k_vlanId)] = [self.peerIf, c_rif, now, now, k_dst, k_vni]
                                                self.HwWorkQueuePut((fdbsync._MAC_MOVE, k_mac, k_vlanId, self.peerIf, None))
                                            else:
                                                if c_lif == k_bif:
                                                    self.log_debug("Removing remote address intf not up %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                                    del self.macDB[(k_mac, k_vlanId)]

                        self.macDBLock.release()

                    if xmlEntries >= 2000 and self.keep_going:
                        xmlStr = ET.tostring(root)
                        self.sendQueue.put(xmlStr)
                        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                        xmlEntries = 0

                    (opcode, params, queueEmpty) = self.GetWorkQueueDelayed()

                if xmlEntries and self.keep_going:
                    #self.PrintElement(root, 0)
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)

        self.log_verbose("Finished execution of the thread FdbKernelMsg")

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.

        root - The root element of the XML tree
        """
        # Get the current list of dual connected hosts
        self.log_verbose("Beginning DecodeXmlInfo")
        dualIfs    = self.GetDualIfs()
        revDualIfs = { v : k for k , v in dualIfs.items() }
        new_root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0

        if self.Parser.args.permanentMacSync:
            self.permanentMacDBLock.acquire()
            self.macDBLock.acquire()
            for fdb in root.findall('fdb-permanent'):
                p_add = True if fdb.get('op') == "permanent-add" else None
                r_mac = fdb.get('mac')
                r_vlanId = fdb.get('vlanId')
                if r_vlanId == "None":
                    r_vlanId = None
                if r_vlanId is not None:
                    r_vlanId = int(r_vlanId)
                # if we have received permanent mac update from the peer and if the mac 
                # exists in our dynamic mac Db delete it .
                if (r_mac, r_vlanId) in self.macDB and p_add:
                    self.log_debug("Permanent MAC add. Entry exists in dynamic mac DB, Deleting it")
                    del self.macDB[(r_mac, r_vlanId)]
                if (r_mac, r_vlanId) not in self.permanentMacDB:
                    if p_add:
                        self.log_debug("Permanent MAC (%s, %s) added on peer, updating permanentMacDB" % (str(r_mac), (r_vlanId)))
                        self.permanentMacDB[(r_mac, r_vlanId)] = [self.peerIf, False, True]
                        self.HwWorkQueuePut((fdbsync._MAC_STATIC_ADD, r_mac, r_vlanId, self.peerIf, None))
                elif self.permanentMacDB[(r_mac, r_vlanId)][1] == True:
                    if p_add:
                        self.log_debug("Same permanent MAC (%s, %s) added on peer, updating permanentMacDB" % (str(r_mac), str(r_vlanId)))
                        self.permanentMacDB[(r_mac, r_vlanId)][2] = True
                    else:
                        self.log_debug("Same permanent MAC (%s, %s) deleted on peer, updating permanentMacDB" % (str(r_mac), str(r_vlanId)))
                        self.permanentMacDB[(r_mac, r_vlanId)][2] = False
                else:
                    if not p_add:
                        self.log_debug("Permanent MAC (%s, %s) deleted on peer, removing from permanentMacDB" % (str(r_mac), str(r_vlanId)))
                        del self.permanentMacDB[(r_mac, r_vlanId)]
                        self.HwWorkQueuePut((fdbsync._MAC_STATIC_DEL, r_mac, r_vlanId, self.peerIf, None))
            self.macDBLock.release()
            self.permanentMacDBLock.release()

        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(new_root)
            self.sendQueue.put(xmlStr)
        self.log_verbose("Ending DecodeXmlInfo")
        xmlEntries = 0

        self.macDBLock.acquire()
        self.permanentMacDBLock.acquire()
        if len(root.findall('fdb-start-sync')):
            self.log_debug("Starting sync of remote addresses")
            self.syncDB = copy.deepcopy(self.macDB)

        now = int(os.times()[4])

        for fdb in root.findall('fdb'):
            r_add = fdb.get('op') in [ "add", "sync" ]
            r_mac = fdb.get('mac')
            r_vlanId = fdb.get('vlanId')
            if r_vlanId == "None":
                r_vlanId = None
            if r_vlanId is not None:
                r_vlanId = int(r_vlanId)
            r_vni = fdb.get('vni')
            if r_vni == "None":
                r_vni = None
            if r_vni is not None:
                r_vni = int(r_vni)
            r_dst = fdb.get('dst')
            if r_dst == "None":
                r_dst = None
            r_peerIf = fdb.get('if')
            r_if = revDualIfs.get(r_peerIf, self.peerIf)
            if fdb.get('op') == "sync":
                self.syncDB.pop((r_mac, r_vlanId), None)

            if (r_mac, r_vlanId) not in self.macDB:
                #If we have received a dynamic add for a mac which is permanent or interface mac
                is_permanent_mac = False
                is_intf_mac = False
                if (r_mac, r_vlanId) in self.permanentMacDB and \
                        self.permanentMacDB[(r_mac, r_vlanId)][1] == True:
                    is_permanent_mac = True
                sviIf = self.Intf.GetLocalIfBasedOnPeerVlanId(r_vlanId)
                sviMac = self.Intf.GetIfMac(sviIf)
                if r_mac == sviMac:
                    is_intf_mac = True
                if r_add and not is_permanent_mac and not is_intf_mac:
                    self.log_debug("Adding new remote address %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, r_if, r_dst))
                    self.macDB[(r_mac, r_vlanId)] = [r_if, r_peerIf, now, now, r_dst, r_vni]
                    if (r_mac, r_vlanId) in self.permanentMacDB:
                        self.log_debug("Entry exists in permanentDB.Deleting it from permanentDB")
                        del self.permanentMacDB[(r_mac, r_vlanId)]
                    self.HwWorkQueuePut((fdbsync._MAC_ADD, r_mac, r_vlanId, r_if, r_dst))
                elif is_permanent_mac:
                    self.log_debug("Remote mac address %s vlan %d is found permanent-mac DB .Skipping remote-add. Resyncing the permanent mac entry to the peer" % (r_mac, r_vlanId,))
                    ET.SubElement(new_root, "fdb-permanent", {"op" : "permanent-add", "mac" : r_mac, "vlanId" : str(r_vlanId)})
                    xmlEntries += 1
                elif is_intf_mac:
                    self.log_debug("Svi Intf mac address %s vlan %d. Skipping remote add." % (r_mac, r_vlanId,))
            else:
                [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = self.macDB[(r_mac, r_vlanId)]
                if not c_rif:
                    if r_add:
                        if (now - c_added) > fdbsync._RECENT or self.Parser.clagRole != self.Parser._ROLE_PRIMARY:
                            self.log_debug("Moving existing local address to remote %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, c_lif, r_dst))
                            self.macDB[(r_mac, r_vlanId)] = [r_if, r_peerIf, now, now, r_dst, r_vni]
                            if c_lif != r_if or c_dst != r_dst:
                                self.HwWorkQueuePut((fdbsync._MAC_MOVE, r_mac, r_vlanId, r_if, r_dst))
                    else:
                        self.log_debug("Telling peer to re-add address he deleted %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, c_lif, c_dst))
                        ET.SubElement(new_root, "fdb", {"op" : "add", "mac" : r_mac, "vlanId" : str(r_vlanId), "vni" : str(r_vni), "if" : c_lif, "dst" : str(c_dst)})
                        xmlEntries += 1
                else:
                    if r_add:
                        if c_lif != r_if or c_dst != r_dst:
                            self.log_debug("Moving remote address %s vlan %s vni %s dev %s dst %s to %s" % (r_mac, r_vlanId, r_vni, c_lif, r_dst, r_if))
                            self.macDB[(r_mac, r_vlanId)] = [r_if, r_peerIf, now, now, r_dst, r_vni]
                            self.HwWorkQueuePut((fdbsync._MAC_MOVE, r_mac, r_vlanId, r_if, r_dst))
                    else:
                        if r_if == self.peerIf and r_if == c_lif:
                            self.log_debug("Removing remote address %s vlan %s vni %s dev %s" % (r_mac, r_vlanId, r_if, r_vni))
                            del self.macDB[(r_mac, r_vlanId)]
                            self.HwWorkQueuePut((fdbsync._MAC_DELETE, r_mac, r_vlanId, r_if, r_dst))
                        else:
                            self.log_debug("Enqueueing remote delete %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, c_lif, r_dst))
                            self.RemoteDeleteQueuePut((r_mac, r_vlanId, r_vni, c_lif, r_dst))


        if len(root.findall('fdb-stop-sync')):
            self.log_debug("Stoping sync of remote addresses")
            for (s_mac, s_vlanId), [s_lif, s_rif, s_updated, s_added, s_dst, s_vni] in self.syncDB.items():
                if s_rif:
                    self.log_debug("Sync Removing remote address %s vlan %s vni %s dev %s" % (s_mac, s_vlanId, s_vni, s_lif))
                    self.macDB.pop((s_mac, s_vlanId), None)
                    self.HwWorkQueuePut((fdbsync._MAC_DELETE, s_mac, s_vlanId, s_lif, s_dst))

        self.permanentMacDBLock.release()
        self.macDBLock.release()

        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(new_root)
            self.sendQueue.put(xmlStr)
        self.log_verbose("Ending DecodeXmlInfo")

    def GetKernelFdb(self, nlm):
        '''
        Retrieve the kernel's forwarding database. The database is a python
        dictionary whose key is a tuple of the mac address and vlan id. The
        data of each entry is a list, with the first element the interface
        on which the mac/vlan exists and the second element is the number of
        seconds since the address was last seen.
        '''
        vlanMap = self.Intf.GetSubIfsVlanMap()
        brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
        pvidMap = self.Intf.GetPvidMap()
        k_macDB = {}

        debug = nlmanager.nlpacket.RTM_GETNEIGH in nlm.debug
        macList = nlm.request_dump(nlmanager.nlpacket.RTM_GETNEIGH, socket.AF_BRIDGE, debug)

        for mac in macList:
            if mac.family == socket.AF_BRIDGE and mac.state & mac.NUD_REACHABLE:
                k_mac    = mac.get_attribute_value(mac.NDA_LLADDR)
                k_vlanId = mac.get_attribute_value(mac.NDA_VLAN)
                k_bif    = self.Intf.GetIfName(mac.ifindex)
                k_vni    = mac.get_attribute_value(mac.NDA_VNI)
                k_dst    = mac.get_attribute_value(mac.NDA_DST)
                k_master = self.Intf.GetIfName(mac.get_attribute_value(mac.NDA_MASTER))
                k_cache  = mac.get_attribute_value(mac.NDA_CACHEINFO)

                if k_bif is None:
                    continue
                k_if = vlanMap.get(k_bif, (k_bif, None))[0]

                # Get remote IP address (vxlan)
                # XXX: Check if dst/vni is always included for vxlan interaces
                vxlanInfo = self.Parser.GetVxLanInfo(k_if)
                if k_dst is not None:
                    k_dst = str(k_dst)
                    if not vxlanInfo:
                        continue
                    else:
                        k_vni = vxlanInfo[0]
                else:
                    if vxlanInfo:
                        continue
                    else:
                        k_vni = 0xFFFFFFFF

                # Get the vlanId
                if k_vlanId is None:
                    k_vlanId = 0 if not k_dst else 0xFFFF

                if not k_vlanId:
                    if not k_master:
                        continue
                    k_vlanId = brPeerVlanMap.get(k_master, None)
                    if k_vlanId is None:
                        continue
                elif k_vlanId == 0xFFFF:
                    k_master = self.Intf.GetBridgeOfMember(k_if)
                    k_vlanId = brPeerVlanMap.get(k_master, pvidMap.get(k_if))
                    if k_vlanId is None:
                        continue

                try:
                    k_age = k_cache[2] / 100
                except (TypeError, IndexError):
                    k_age = 0

                k_macDB[(k_mac, k_vlanId)] = [k_if, k_age, k_dst, k_vni, k_bif]

        return k_macDB

    def RemoteDelete(self):
        """
        This is an independent thread of the script. It is responsible for
        processing entries on the RemoteDeleteQueue, which handles requests from
        the peer to delete an address.
        """
        self.log_verbose("Beginning execution of the thread RemoteDelete")
        nlm = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_REMOTE_DELETE)

        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution of the thread RemoteDelete")
            (r_mac, r_vlanId, r_vni, r_if, r_dst) = self.remoteDeleteQueue.get()

            if self.keep_going:
                queueEmpty = False
                dualIfs = self.GetDualIfs()
                now = int(os.times()[4])
                k_macDB = self.GetKernelFdb(nlm)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
                while not queueEmpty and self.keep_going:
                    self.macDBLock.acquire()
                    [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = self.macDB.get((r_mac, r_vlanId), [None, None, None, None, None, None])
                    if c_rif:
                        [k_if, k_age, k_dst, k_vni, k_bif] = k_macDB.get((r_mac, r_vlanId), [None, 0, None, 0xFFFFFFFF, None])
                        self.log_debug("Processing delete candidate: %s vlan %s vni %s dev %s. c_lif: %s, c_rif: %s, c_updated: %s, k_if: %s, k_age: %s" % (r_mac, r_vlanId, r_vni, r_if, c_lif, c_rif, c_updated, k_if, k_age))
                        if k_if in dualIfs:
                            if (k_age + fdbsync._RECENT) < (now - c_updated):
                                self.log_debug("Telling peer to re-add address (now local) %s vlan %s vni %s dev %s. k_age: %s, now: %s, c_updated: %s" % (r_mac, r_vlanId, r_vni, r_if, k_age, now, c_updated))
                                self.macDB[(r_mac, r_vlanId)] = [k_if, None, c_updated, c_added, k_dst, c_vni]
                                ET.SubElement(root, "fdb", {"op" : "add", "mac" : r_mac, "vlanId" : str(r_vlanId), "vni" : str(c_vni), "if" : k_if, "dst" : str(k_dst)})
                                xmlEntries += 1
                            else:
                                self.log_debug("Removing remote address %s vlan %s vni %s dev %s" % (r_mac, r_vlanId, r_vni, k_if))
                                del self.macDB[(r_mac, r_vlanId)]
                                self.HwWorkQueuePut((fdbsync._MAC_DELETE, r_mac, r_vlanId, k_if, k_dst))
                        else:
                            self.log_debug("Removing non-dual remote address %s vlan %s vni %s dev %s" % (r_mac, r_vlanId, r_vni, r_if))
                            del self.macDB[(r_mac, r_vlanId)]
                            self.HwWorkQueuePut((fdbsync._MAC_DELETE, r_mac, r_vlanId, r_if, r_dst))

                    self.macDBLock.release()

                    try:
                        (r_mac, r_vlanId, r_vni, r_if, r_dst) = self.remoteDeleteQueue.get(True, 0.001)
                    except Queue.Empty:
                        queueEmpty = True

                if xmlEntries and self.keep_going:
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)

        self.log_verbose("Finished execution of the thread RemoteDelete")

    def FlushMacsOnDownIf(self, iface):
        # clag bond is down on both sides; flush macs locally
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0
        self.macDBLock.acquire()
        for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in self.macDB.items():
            if c_lif == iface and c_rif is None:
                self.log_debug("interface down on both sides delete"
                               "local address %s vlan %s vni %s dev %s" %
                               (c_mac, c_vlanId, c_vni, c_lif))
                del self.macDB[(c_mac, c_vlanId)]
                self.HwWorkQueuePut((fdbsync._MAC_DELETE, c_mac, c_vlanId, c_lif, None))
                if c_vni == 0xFFFFFFFF:
                    ET.SubElement(root, "fdb", {"op" : "delete", "mac" : c_mac, "vlanId" : str(c_vlanId), "vni" : str(c_vni), "if" : c_lif, "dst" : "None"})
                    xmlEntries += 1
                    if xmlEntries >= 2000 and self.keep_going:
                        xmlStr = ET.tostring(root)
                        self.sendQueue.put(xmlStr)
                        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                        xmlEntries = 0
        self.macDBLock.release()
        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(root)
            self.sendQueue.put(xmlStr)


    def DoDualIfChange(self, ifs, enable, matchIdChange):
        """
        Called when an interface becomes or is no longer dual connected
        """
        if self.keep_going:
            root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
            xmlEntries = 0
            if enable:
                dualIfs    = self.GetDualIfs()
                revDualIfs = { v : k for k , v in dualIfs.items() }

                self.macDBLock.acquire()
                now = int(os.times()[4])
                for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in self.macDB.items():
                    r_lif = revDualIfs.get(c_rif, None)
                    if r_lif in ifs:
                        self.log_debug("Dual connected, Moving address %s vlan %s vni %s dev %s to %s" % (c_mac, c_vlanId, c_vni, c_lif, r_lif))
                        self.macDB[(c_mac, c_vlanId)] = [r_lif, c_rif, now, now, c_dst, c_vni]
                        self.HwWorkQueuePut((fdbsync._MAC_MOVE, c_mac, c_vlanId, r_lif, c_dst))
                    if c_lif in ifs:
                        self.log_debug("Telling peer to re-add address (now dual connected) %s vlan %s vni %s dev %s." % (c_mac, c_vlanId, c_vni, c_lif))
                        ET.SubElement(root, "fdb", {"op" : "add", "mac" : c_mac, "vlanId" : str(c_vlanId), "vni" : str(c_vni), "if" : c_lif, "dst" : str(c_dst)})
                        xmlEntries += 1
                        if xmlEntries >= 2000 and self.keep_going:
                            xmlStr = ET.tostring(root)
                            self.sendQueue.put(xmlStr)
                            root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                            xmlEntries = 0
                self.macDBLock.release()

            else:
                # Interfaces are becoming singly connected
                self.macDBLock.acquire()
                operState = {}
                for s_if in ifs:
                    operState[s_if] = self.Intf.GetOperStateFromSysFs(s_if)
                now = int(os.times()[4])
                for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in self.macDB.items():
                    if c_lif in ifs:
                        #The  peerAlive Flag is not set immediately after peerlink is down. Avoiding this with an additional check
                        #of peer is not alive
                        if not self.Intf.isPeerAlive() or self.Intf.isPeerLinkDown():
                            # If peer not alive and we are primary, make all addresses local.
                            if self.Parser.clagRole == self.Parser._ROLE_PRIMARY:
                                self.log_debug("Not dual connected, primary, peer down, make local %s vlan %s vni %s dev %s" % (c_mac, c_vlanId, c_vni, c_lif))
                                self.macDB[(c_mac, c_vlanId)] = [c_lif, None, c_updated, c_added, c_dst, c_vni]
                            # If peer not alive and we are secondary, flush all addresses.
                            else:
                                self.log_debug("Not dual connected, secondary, peer down, delete address %s vlan %s vni %s dev %s" % (c_mac, c_vlanId, c_vni, c_lif))
                                del self.macDB[(c_mac, c_vlanId)]
                                self.HwWorkQueuePut((fdbsync._MAC_DELETE, c_mac, c_vlanId, c_lif, c_dst))
                        else:
                            if operState[c_lif] != "up" or matchIdChange:
                                if c_rif is not None:
                                    self.log_debug("Not dual connected, down or non-clag-dev, move address %s vlan %s vni %s dev %s to the peer" % (c_mac, c_vlanId, c_vni, c_lif))
                                    self.macDB[(c_mac, c_vlanId)] = [self.peerIf, c_rif, now, now, c_dst, c_vni]
                                    self.HwWorkQueuePut((fdbsync._MAC_MOVE, c_mac, c_vlanId, self.peerIf, None))
                                else:
                                    self.log_debug("Not dual connected, down or non-clag-dev, delete local address %s vlan %s vni %s dev %s" % (c_mac, c_vlanId, c_vni, c_lif))
                                    del self.macDB[(c_mac, c_vlanId)]
                                    self.HwWorkQueuePut((fdbsync._MAC_DELETE, c_mac, c_vlanId, c_lif, None))
                                    if c_vni == 0xFFFFFFFF:
                                        ET.SubElement(root, "fdb", {"op" : "delete", "mac" : c_mac, "vlanId" : str(c_vlanId), "vni" : str(c_vni), "if" : c_lif, "dst" : "None"})
                                        xmlEntries += 1
                                        if xmlEntries >= 2000 and self.keep_going:
                                            xmlStr = ET.tostring(root)
                                            self.sendQueue.put(xmlStr)
                                            root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                                            xmlEntries = 0
                self.macDBLock.release()

            if xmlEntries and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)

    def DualIfChange(self, ifs, enable):
        if self.Parser.args.redirectEnable:
            return
        self.DoDualIfChange(ifs, enable, matchIdChange=False)

    def MatchIdIfChange(self, ifs, enable):
        if not self.Parser.args.redirectEnable:
            return
        self.DoDualIfChange(ifs, enable, matchIdChange=True)

    def ClearPeerInfo(self):
        """
        Removes permanent MAC addresses which are owned by the peer.
        """
        self.permanentMacDBLock.acquire()
        for (mac, vlanId), [lif, permanent, static] in self.permanentMacDB.items():
            if static and not permanent:
                del self.permanentMacDB[(mac, vlanId)]
                self.HwWorkQueuePut((fdbsync._MAC_STATIC_DEL, mac, vlanId, self.peerIf, None))
        self.permanentMacDBLock.release()

    def ClearLocalInfo(self):
        """
        Removes MAC addresses which are owned by this switch.
        """
        self.macDBLock.acquire()
        for (mac, vlanId), [lif, rif, updated, added, dst, vni] in self.macDB.items():
            if not rif:
                self.HwWorkQueuePut((fdbsync._MAC_DELETE, mac, vlanId, lif, dst))
                del self.macDB[(mac, vlanId)]
        self.macDBLock.release()

    def UpdateMacsInHw(self):
        '''
        This function is an independent thread of the script. It compares the MAC
        address data base received from the peer (peerMacDB) with the mac addresses
        we have already programmed into the hardware. Any new addresses are added and
        any addresses no longer present on the peer are removed.
        '''
        fdb_cmd = {
            fdbsync._MAC_ADD     : ( nlmanager.nlpacket.RTM_NEWNEIGH, nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_REPLACE, "added to"     ),
            fdbsync._MAC_STATIC_ADD : ( nlmanager.nlpacket.RTM_NEWNEIGH, nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_REPLACE, "added (static) to"),
            fdbsync._MAC_REFRESH : ( nlmanager.nlpacket.RTM_NEWNEIGH, nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_REPLACE, "refreshed on" ),
            fdbsync._MAC_MOVE    : ( nlmanager.nlpacket.RTM_NEWNEIGH, nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_REPLACE, "moved to"     ),
            fdbsync._MAC_DELETE  : ( nlmanager.nlpacket.RTM_DELNEIGH, 0, "removed from" ),
            fdbsync._MAC_STATIC_DEL : ( nlmanager.nlpacket.RTM_DELNEIGH, 0, "removed (static) from")
        }

        self.log_verbose("Beginning execution of the thread UpdateMacsInHw")
        nlm = fdb_nlmanager(self.Log)

        while self.keep_going:
            # Wait for a change in mac address data base received from the peer
            self.log_verbose("Waiting for an event before continuing execution of the thread UpdateMacsInHw")
            (op, mac, vlanId, lif, dst) = self.hwWorkQueue.get()

            if self.keep_going:
                masterVlanMap = self.Intf.GetMasterIfVlanMap()
                peerVlanMap = self.Intf.GetPeerBridgeVlanMap()
                queueEmpty = False
                params = []
                vlanFilteringIf = {}
                while not queueEmpty and self.keep_going:
                    (rtm_command, nlm_flags, fdb_str) = fdb_cmd.get(op, (None, None, None))
                    if rtm_command is not None:
                        pvlanId = peerVlanMap.get(vlanId, {}).get(lif)
                        intfStr = masterVlanMap.get(lif,{}).get(pvlanId,lif)
                        intf = self.Intf.GetIfIndex(intfStr)
                        if intf is not None:
                            try:
                                vlanFiltering = vlanFilteringIf[intf]
                            except KeyError:
                                br = self.Intf.GetBridgeOfMember(intfStr)
                                if br is None:
                                    try:
                                        (op, mac, vlanId, lif, dst) = self.hwWorkQueue.get(True, 0.001)
                                    except Queue.Empty:
                                        queueEmpty = True
                                    continue
                                vlanFiltering = self.Intf.isVlanFilteringBridge(br)
                                vlanFilteringIf[intf] = vlanFiltering
                            vlanId = vlanId if vlanId and not pvlanId and vlanFiltering else 0
                            params += [(rtm_command, nlm_flags, mac, intf, vlanId, dst, op)]
                            self.log_verbose("MAC address %s has been %s %s in VLAN %d dst %s." % (mac, fdb_str, intfStr, vlanId, dst))
                    else:
                        self.Log.log_error("Unknown op code in work queue: %s" % (op,))

                    try:
                        (op, mac, vlanId, lif, dst) = self.hwWorkQueue.get(True, 0.001)
                    except Queue.Empty:
                        queueEmpty = True

                nlm.mac_batch(params)

        self.log_verbose("Finished execution of the thread UpdateMacsInHw")


class fdb_nlmanager(nlmanager.nlmanager.NetlinkManager):

    def __init__(self, Log):
        super(fdb_nlmanager, self).__init__(util.CLAG_NLM_FDB_ID)
        self.Log = Log

    def tx_or_concat_message(self, total_message, neigh):
        """
        Adding a mac only takes approx 60 bytes, if we are adding thousands
        of them this can add up to a lot of send calls.  Concat several of
        them together before TXing.
        """
        PACKET_CONCAT_SIZE = 16384

        if not total_message:
            total_message = neigh.message
        else:
            total_message += neigh.message

        if len(total_message) >= PACKET_CONCAT_SIZE:
            self.tx_nlpacket_raw(total_message)
            total_message = None

        return total_message

    def mac_batch(self, macs):

        total_message = None

        for (rtm_command, nlm_flags, mac, intf, vlanid, dst, op) in macs:
            debugFlags = self.Log._LOG_DEBUG_NETLINK | self.Log._LOG_DEBUG_FDB_SYNC
            debug = (self.Log.getDebug() & debugFlags) == debugFlags
            neigh = nlmanager.nlpacket.Neighbor(rtm_command, debug, self.Log.logger, self.Log.getHandlerStr() != "syslog")
            neigh.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlm_flags
            neigh.family = socket.AF_BRIDGE

            ntf = neigh.NTF_MASTER
            nud_state = neigh.NUD_REACHABLE
            if op == fdbsync._MAC_REFRESH:
                ntf |= neigh.NTF_USE
            elif op == fdbsync._MAC_STATIC_ADD or op == fdbsync._MAC_STATIC_DEL:
                # adding/deleting a static entry
                nud_state = neigh.NUD_NOARP
            if dst:
                ntf |= neigh.NTF_SELF
            neigh.body = struct.pack(neigh.PACK, socket.AF_BRIDGE, intf, nud_state, ntf, 0)
            neigh.add_attribute(neigh.NDA_LLADDR, mac)
            if vlanid:
                neigh.add_attribute(neigh.NDA_VLAN, vlanid)
            if dst:
                neigh.add_attribute(neigh.NDA_DST, ipaddr.IPv4Address(dst))

            neigh.build_message(self.sequence.next(), self.pid)
            total_message = self.tx_or_concat_message(total_message, neigh)

        if total_message:
            self.tx_nlpacket_raw(total_message)
