#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2015-2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.util
    import nlmanager.nlmanager
    import socket
    import threading
    import select
    import os
    import xml.etree.ElementTree as ET
    from ctypes import *
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

#-------------------------------------------------------------------------------
# This module checks the health of the peer switch via a backup IP address.
# 1. This IP address is typically the mgmt. IP address (but could also be an 
# inband/frontpanel IP). The backup IP config is optional (on account of upgrade
# from previous release but is recommended to avoid traffic loss on peer switch 
# death without "goodbye").
# 2. UDP is used for these health checks as UDP is connectionless (i.e. light 
# weight) and typically not blocked by mgmt network firewalls (on account of 
# traceroute UDP probes).
# 3. The healthcheck request is a light weight periodic hello request which is 
# turned around as a reply by clagd on the peer switch.
# 4. The backup state is only used if the backup was active at the time of peer 
# link down on the secondary switch. This takes care of the upgrade case where i
# the peer is not responding to the "new" hello requests. And the case were the 
# peer's mgmt IP was changed independently without updating the backup IP 
# config.
# 5. To avoid overwhelming the mgmt network the frequency of the 
# steady-state/periodic hellos has been kept low, one every 4 seconds. On a 
# potential peer death event (peer link down/peer-connection timeout) 
# healthCheck is moved to a rapid-detection mode (and the peer-down event is 
# deferred). In this mode 6 requests are sent 100ms apart. Even if one of the
# requests is responded too clagd assumes that the peer switch is alive (and 
# does not take over as primary; unless a goodbye is rxed). If no responses are
# rxed during rapid-detection the peer is considered dead and clagd takes over 
# the primary role.
#-------------------------------------------------------------------------------

libc_handle = cdll.LoadLibrary('libc.so.6')

class HealthCheck():

    # MAX counter to control the socket error logging while HelloRx and HelloTx
    _MAX_SOCKET_ERROR = 100

    def __init__(self, Daemon, Parser, Log, Intf):
        self.Daemon = Daemon
        self.Parser = Parser
        self.Log = Log
        self.Intf = Intf

        # params that cannot be changed on the fly
        self.peerHealthPort = self.Parser.args.backupPort

        # clag hello id - changes everytime clag is restarted
        self.helloId = os.getpid()
        # seq increments for every txed hello
        self.helloSeq = 0
        self.helloOutMaxReqs = 6
        # we can make the intervals configurable if needed; for now 4 seconds
        # and will take up to 24 seconds to declare peer dead. The time taken 
        # to declare backup inactive should be higher than the time it takes to 
        # detect peer link going down otherwise we will end up ignoring the 
        # backup state change on peer switch death.
        self.helloRegularInterval = 4
        self.helloRapidInterval = 0.1

        # Data shared by all the threads
        self.helloOutReqs = 0
        self.helloOutReqsLock = threading.Lock()
        self.backupActive = False
        self.backupReason = ""
        self.goodbyeSent = False
        self.backupVrfUnbind = False
        self.backupVrfBindSuccess = False

        # Election is run over the backup independently
        self.backupRoleAvailable = False
        self.backupRole = self.Parser.default_role
        self.peerId = [self.Parser.default_priority, "00:00:00:00:00:00"]
        self.peerRole = None
        self.initState = False
        self.peerInitState = False

        # peerlink down triggers rapid detection (provided health is good at the
        # time of the event)
        self.helloRapidDetection = False
        self.keep_going = True

        # Reload timer (5min) - moves the clag bond out of error disabled state 
        # on clagd startup (if nothing is heard from the peer)
        # XXX - any way to avoid this thread? 
        self.helloReloadEvent = clag.clagthread.Event()
        self.helloReloadThread = threading.Thread(None, self.HelloReloadT, "HelloReloadT")
        self.helloReloadThread.daemon = True
        self.helloReloadStartTime = int(os.times()[4])
        self.helloReloadThread.start()

        # Temporary thread to keep clag bonds in protodown for x seconds
        if self.Parser.args.initDelay:
            self.initDelayEvent = clag.clagthread.Event()
            self.initDelayTimerStartEvent = clag.clagthread.Event()
            self.initDelayThread = threading.Thread(None, self.InitDelayT, "InitDelayT")
            self.initDelayThread.daemon = True
            self.initDelayThread.start()

        # helloTxT - tx thread - 
        # 1. Sends hello/echo-requests periodically
        # 2. Declares peerSwitch dead if outstanding (not replied to)
        #    requests reach max (helloOutMaxReqs)
        self.helloTxEvent = clag.clagthread.Event()
        try:
            self.helloTxSock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            # Even though it is udp socket introducing the timeout here, since 
            # due to buffer issues sendto call had gone to infinite wait state 
            # and was not sending hello requests.
            self.helloTxSock.settimeout(self.Parser.args.sendTimeout)
        except socket.error:
            self.Log.log_error("HealthCheck: Failed to create a udp socket")
            self.Daemon.SignalAllThreadsToStop()
            return

        # helloRxT - rx thread - 
        # 1. Listens on peerHealthPort
        # 2. Replies to hello/echo-requests (simple turn around)
        # 3. Declares peerSwitch alive when a valid reply is rxed.
        #
        # Note: Using separate sockets for rx and tx (they are by seperate 
        # threads)
        # create a wake-up eventfd that can be used by the main thread to
        # wakeup the helloRx thread in case it needs to exit. Otherwise
        # the helloRx thread is blocked listening on the peerHealthPort
        self.helloRxWkupFd = libc_handle.eventfd(0, 0)
        try:
            self.helloRxSock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        except socket.error:
            self.Log.log_error("HealthCheck: Failed to create a udp socket")
            self.Daemon.SignalAllThreadsToStop()
            return

        # Try to bind the socket to the VRF if configured.
        self.AttemptVrfBind()

        self.helloRxSocErrCounter = 0
        self.helloTxSocErrCounter = 0

        # Start tx/rx threads after tx/rx sockets creation
        self.helloTT = threading.Thread(None, self.HelloTxT, "HelloTxT")
        self.helloTT.daemon = True
        self.helloTT.start()

        try:
            self.helloRxSock.bind(('', self.peerHealthPort))
            self.helloRT = threading.Thread(None, self.HelloRxT, "HelloRxT")
            self.helloRT.daemon = True
            self.helloRT.start()
        except socket.error:
            self.Log.log_error("HealthCheck: Failed to bind to the backup udp port %d" % (self.peerHealthPort))
            self.Daemon.SignalAllThreadsToStop()
            return

    def __del__(self):
        # Signal the update thread to wake up and exit
        self.SignalThreads()
        self.JoinThreads()

    def HelloRxT(self):
        try:
            self.Log.log("[%d]Init HelloRxT" % (self.Parser.getTid()))
            self.HelloRx()
            self.helloRxSock.close()
        except:
            self.Daemon.DumpTraceback()

    def HelloTxT(self):
        try:
            self.Log.log("[%d]Init HelloTxT" % (self.Parser.getTid()))
            self.HelloTx()
        except:
            self.Daemon.DumpTraceback()

    def HelloReloadT(self):
        try:
            self.Log.log("[%d]Init HelloReloadT" % (self.Parser.getTid()))
            self.HelloReload()
        except:
            self.Daemon.DumpTraceback()

    def InitDelayT(self):
        try:
            self.Log.log("[%d]Init InitDelayT" % (self.Parser.getTid()))
            self.InitDelay()
        except:
            self.Daemon.DumpTraceback()


    def SignalThreads(self):
        self.keep_going = False
        self.helloTxEvent.set()
        try:
            self.helloReloadEvent.set()
        except AttributeError:
            pass
        try:
            self.initDelayEvent.set()
        except AttributeError:
            pass
        try:
            os.write(self.helloRxWkupFd, c_ulonglong(1))
        except:
            self.Log.log_error("helloRx: failed to wakeup thread")

    def JoinThreads(self):
        try:
            self.helloRT.join()
            self.helloTT.join()
            self.helloReloadThread.join()
            self.InitDelayThread.join()
        except AttributeError:
            pass

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_HEALTH_CHECK, logStr)

    def PrintElement(self, node, level):
        self.log_debug("%s%s: %s: %s" % (" " * (level*3), node.tag, node.text, str(node.attrib)))
        for child in node.findall('*'):
            self.PrintElement(child, level+1)

    def AttemptVrfBind(self):
        if self.backupVrfUnbind:
            try:
                self.helloTxSock.setsockopt(socket.SOL_SOCKET, 25, "")
                self.helloRxSock.setsockopt(socket.SOL_SOCKET, 25, "")
            except socket.error:
                self.Log.log_error("HealthCheck: Failed to clear socket device binding")
            else:
                self.backupVrfUnbind = False

        if self.backupVrfBindSuccess or not self.Parser.args.backupVrf:
            return True

        self.log_debug("HealthCheck: bind socket to vrf %s" % self.Parser.args.backupVrf)
        try:
            self.helloTxSock.setsockopt(socket.SOL_SOCKET, 25, self.Parser.args.backupVrf)
            self.helloRxSock.setsockopt(socket.SOL_SOCKET, 25, self.Parser.args.backupVrf)
        except socket.error:
            self.backupVrfBindSuccess = False
            self.Log.log_error("HealthCheck: Failed to bind socket to vrf %s" % self.Parser.args.backupVrf)
            return False

        self.backupVrfBindSuccess = True
        return True

    def HelloClearOutReqs(self):
        self.helloOutReqsLock.acquire()
        self.helloOutReqs = 0
        self.helloOutReqsLock.release()

    def HelloIncOutReqs(self):
        self.helloOutReqsLock.acquire()
        self.helloOutReqs = self.helloOutReqs + 1
        self.helloOutReqsLock.release()

    def HelloGetOutReqs(self):
        self.helloOutReqsLock.acquire()
        outReqs = self.helloOutReqs
        self.helloOutReqsLock.release()
        return outReqs

    def HelloBuildXmlMessage(self, echoType, echoId, seq):
        '''
        <clag_hello_data version="0.0.0">
            <echo type="request" id=helloId seq=1/>
            <echo type="request" id=helloId seq=1/>
            <i8023ad system_id="00:00:00:01:01:01"/>
            <clagid uniqueId="44:38:39:00:11:16" priority=32768 role=Primary/>
        </clag_hello_data>
        '''
        clagDataAttr = {'version' : self.Parser._ClagHelloVersion}
        root = ET.Element("clag_hello_data", clagDataAttr)
        ET.SubElement(root, "echo", {"type" : echoType, "id" : str(echoId), "seq" : str(seq)})
        ET.SubElement(root, "i8023ad", {"system_id" : self.Parser.args.sysMac})
        ET.SubElement(root, "clagid", {"priority" : str(self.Parser.clagId[0]), "uniqueId" : self.Parser.clagId[1], "role" : self.Parser.clagRole})
        ET.SubElement(root, "initState", {"state" : str(self.initState)})

        #self.PrintElement(root, 0)
        return(ET.tostring(root))
        
    def HelloParseXmlMessage(self, nlm, xmlStr, addr):
        # If we're in the process of quitting, don't do any processing
        if self.goodbyeSent is True:
            return

        nakLogs = []
        # Convert the XML string into an ElementTree element
        try:
            root = ET.fromstring(xmlStr)
        except ET.ParseError:
            self.Log.log_error("HealthCheck: data from the peer is invalid:\n%s" % (xmlStr,))
            return

        # Is the root element named properly?
        if root is None or root.tag != "clag_hello_data":
            self.Log.log_error("HealthCheck: data from the peer has an invalid root tag:\n%s" % (xmlStr,))
            return

        # Is there a version attribute with the root element?
        verStr = root.get("version")
        if verStr is None:
            self.Log.log_error("HealthCheck: data from the peer does not include a version:\n%s" % (xmlStr,))
            return

        # Is the version something we understand?
        (peerMajVer, peerMinVer, peerSubVer) = verStr.split(".")
        (myMajVer,   myMinVer,   mySubVer)   = self.Parser._ClagHelloVersion.split(".")
        if int(peerMajVer) > int(myMajVer):
            self.Log.log_error("HealthCheck: version mismatch; My version: %s, Peer version: %s" % (self.Parser._ClagHelloVersion, verStr))
            return

        # Is there a goodbye attribute with the root element?
        byeStr = root.get("goodbye")
        if byeStr == 'yes':
            self.Parser.ClearConfigMismatch(nlm)
            self.Intf.SetClagRole(nlm, self.Parser._ROLE_PRIMARY, "peer sent goodbye via backup link")
            self.peerInitState = False
            self.SetBackupActive(nlm, False)
            self.SetBackupRole(nlm, self.Parser._ROLE_PRIMARY)
            self.HelloClearOutReqs()
            return

        # Is there a peerIpMismatch attribute with the root element?
        peerIpMismatch = root.get("peerIpMismatch")
        if peerIpMismatch == 'yes':
            self.Intf.SetClagRole(nlm, self.Parser._ROLE_PRIMARY, "peer sent peerIpMismatch message via backup link")
            self.Parser.SetRemoteIpMismatchOnPeer()
            self.Parser.SetPeerConfigMismatch()
            self.Parser.ListenersStateUpdate("down")
            return

        # Just parse clagId. We will only use it for election if it
        # was in a valid reply message.
        peerId = []
        peerRole = None
        peerInitState = False
        
        clagIdNode = root.find('clagid')
        if clagIdNode is not None:
            peerId.append(int(clagIdNode.get('priority')))
            peerId.append(clagIdNode.get('uniqueId'))
            peerRole = clagIdNode.get('role')

        peerInitStr = root.find('initState')
        if peerInitStr is not None:
            peerInitState = True if peerInitStr.get('state') == "True" else False

        echoNode = root.find('echo')
        if echoNode is not None:
            echoType = echoNode.get('type')
            echoId = int(echoNode.get('id'), 0)
            seq = int(echoNode.get('seq'), 0)
            self.HelloProcEchoFromPeer(nlm, echoType, echoId, seq, addr, peerId, peerRole, peerInitState, nakLogs)

    def SetBackupRole(self, nlm, role):
        if self.backupRoleAvailable and self.backupRole == role:
            return
        self.Intf.protoDownLock.acquire()
        self.backupRole = role
        self.Log.log('HealthCheck: role via backup is %s' % \
                ('primary' if self.backupRole==self.Parser._ROLE_PRIMARY else 'secondary'))

        if not self.backupRoleAvailable:
            self.backupRoleAvailable = True

        # backupRole may be set before backupActive; in that case
        # SetBackupActive will update the proto_down state on all the 
        # interfaces
        if self.backupActive:
            # if backup active and peerlink down and role via peerlink is primary, change role to secondary
            nlmsgs = ""
            for intf in self.Parser.GetClagIntfDB().values():
                nlmsgs += intf.UpdateIntfProtoDown(nlm, batch=True)
            nlm.tx_nlpacket_raw(nlmsgs)
        self.Intf.protoDownLock.release()

    def IsBackupRoleAvailable(self):
        return self.backupActive and self.backupRoleAvailable

    def GetBackupRole(self):
        return self.backupRole

    def GetBackupActive(self):
        return self.backupActive

    def SetBackupActive(self, nlm, active, reason=""):
        self.Intf.protoDownLock.acquire()
        self.backupActive = active
        self.backupReason = reason
        if not self.backupActive:
            self.backupRoleAvailable = False
        self.Intf.SetProtoPeerOrBackupState(nlm)
        # backup role availability may have changed so re-evaluate
        nlmsgs = ""
        for intf in self.Parser.GetClagIntfDB().values():
            nlmsgs += intf.UpdateIntfProtoDown(nlm, batch=True)
        nlm.tx_nlpacket_raw(nlmsgs)
        self.Intf.protoDownLock.release()

    def HelloBackupReset(self, nlm):
        self.HelloClearOutReqs()
        self.backupReason = ""
        if self.backupActive:
            self.Log.log('HealthCheck: backup ip changed')
            self.peerInitState = False
            self.SetBackupActive(nlm, False)
        self.HelloRapidDetectionDone(nlm)

    def PeerRoleToVal(self, role):
        return 0 if role == self.Parser._ROLE_PRIMARY else 1

    def RunBackupElection(self, nlm):
        peerRoleVal = self.PeerRoleToVal(self.peerRole)
        roleVal = self.PeerRoleToVal(self.Parser.clagRole)
        # Same criteria for election as the one over the ISL. Only
        # difference is that we also include the peerRole determined over
        # ISL as a deciding criterion.
        if roleVal < peerRoleVal or \
            (roleVal == peerRoleVal and \
                (self.Parser.clagId[0] < self.peerId[0] or \
                (self.Parser.clagId[0] == self.peerId[0] and self.Parser.clagId[1] < self.peerId[1]))):
            self.SetBackupRole(nlm, self.Parser._ROLE_PRIMARY)
        else:
            self.SetBackupRole(nlm, self.Parser._ROLE_SECONDARY)

    def HelloProcEchoRep(self, nlm, peerId, peerRole, peerInitState):
        # On rxing an acceptable response we clear all pending requests
        self.HelloClearOutReqs()
        peerInitStateChange = False
        
        if peerInitState != self.peerInitState:
            self.peerInitState = peerInitState
            peerInitStateChange = True

        # perform election if the reply has the relevant info
        if peerId:
            self.peerId = peerId
            self.peerRole = peerRole
            if peerId[1] == self.Parser.clagId[1]:
                self.peerInitState = False
                self.SetBackupActive(nlm, False, "self")
                self.HelloRapidDetectionDone(nlm)
                return
            self.RunBackupElection(nlm)

        if not self.backupActive or peerInitStateChange:
            self.SetBackupActive(nlm, True)
            if peerInitStateChange:
                self.Log.log('HealthCheck: peerInitState changed')
            else:
                self.Log.log('HealthCheck: backup active')

        self.HelloRapidDetectionDone(nlm)
        
    def SwitchdIsNotReady(self):
        '''
        Check if switchd is ready. Use the status provided by systemd. This is
        really a check to see if the CPU can send and receive packets through
        the SWP ports.
        '''
        # Are we using knet interfaces? If so we are not ready when switchd is
        # activating (started but not ready)
        bcmKnetDir = "/sys/module/user_bde"
        if os.path.isdir(bcmKnetDir):
            return self.Parser.switchdState == self.Parser._SWITCHD_STARTING
        # Whithout knet, we are not ready in any other state than active.
        return self.Parser.switchdState != self.Parser._SWITCHD_READY

    def HelloProcEchoFromPeer(self, nlm, echoType, echoId, seq, addr, peerId, peerRole, peerInitState, nakLogs):
        if echoType == 'request':
            if self.SwitchdIsNotReady() and not self.Parser.args.vm:
                self.log_debug("RX message: ignored because switchd is not ready.")
                self.Intf.SetClagRole(nlm, self.Parser._ROLE_SECONDARY, "switchd not ready", True)
                return
            # reply to requests - simple turnaround
            xmlStr = self.HelloBuildXmlMessage('reply', echoId, seq)
            if self.keep_going:
                # if previous vrf binds had failed we try again; if it fails we will
                # supress transmission
                if not self.AttemptVrfBind():
                    return

                try:
                    self.helloRxSock.sendto(xmlStr, (addr[0], self.peerHealthPort))
                    self.log_debug("TX message: %s to %r" % (xmlStr, (addr[0], self.peerHealthPort)))
                except socket.error as e:
                    # Log the socket error and also control the log rate if this exception continuously happens
                    if self.helloRxSocErrCounter == HealthCheck._MAX_SOCKET_ERROR:
                        self.Log.log_error("helloRxSock send reply failed to dest %r, socket error : %s " % (addr[0], (e)))
                        self.helloRxSocErrCounter = 0
                    self.helloRxSocErrCounter += 1
        elif echoType == 'reply':
            if nakLogs:
                #Some thing is wrong with the reply
                for nakLog in nakLogs:
                    self.Log.log_error(nakLog)
                self.Daemon.SignalAllThreadsToStop()
                return
            if echoId == self.helloId and seq == self.helloSeq:
                self.HelloProcEchoRep(nlm, peerId, peerRole, peerInitState)
        else:
            self.Log.log_error("HealthCheck: Rxed %s message from peer" % echoType)

    def HelloRx(self):
        # just receive all packets sent to the "helloPort"
        nlm = nlmanager.nlmanager.NetlinkManager(clag.util.CLAG_NLM_HELLO_RX_ID)
        inputs = [self.helloRxSock, self.helloRxWkupFd]
        while self.keep_going and inputs:
            (readable, writeable, exceptional) = select.select(inputs, [], inputs)
            for errSock in exceptional:
                # TODO - what kind of UDP exceptions can we get
                self.Log.log_error('HealthCheck: %r exception ignored' % errSock)

            if self.keep_going:
                for readSock in readable:
                    try:
                        data, addr = readSock.recvfrom(1024)
                    except:
                        self.Log.log_error("Unexpected error while receiving data via backup link")
                        if self.Parser.args.backupVrf:
                            self.backupVrfUnbind = True
                            self.backupVrfBindSuccess = False
                        (data, addr) = [0, None] 
                    if data:
                        self.log_debug("RX message: %s from %r" % (data, addr))
                        self.HelloParseXmlMessage(nlm, data, addr)

    def HelloTx(self):
        nlm = nlmanager.nlmanager.NetlinkManager(clag.util.CLAG_NLM_HELLO_TX_ID)
        while self.keep_going:
            self.log_debug("Executing thread HelloTx")
            # Check if the peer health has timed out
            outReqs = self.HelloGetOutReqs()
            if outReqs >= self.helloOutMaxReqs:
                if self.backupActive:
                    logstr = 'rapid' if self.helloRapidDetection else 'regular'
                    self.Log.log('HealthCheck: backup connection lost; %s' % logstr)
                    self.peerInitState = False
                    self.SetBackupActive(nlm, False)
                    #If backup has gone down and peerlink was already down change role to primary
                    if self.Intf.isPeerLinkDown():
                        self.Intf.SetClagRole(nlm, self.Parser._ROLE_PRIMARY, "backup connection is lost while peer-link down", False)
                        if not self.Parser.args.backupIp:
                            self.Intf.SetAllClagBondsSysMac(nlm, self.Parser.args.sysMac, [self.Parser.args.peerIf])
                self.HelloRapidDetectionDone(nlm)

            # Send echo request
            healthIp = self.Parser.args.backupIp
            # if previous vrf binds had failed we try again; if it fails we will
            # supress transmission
            if healthIp and self.AttemptVrfBind():
                self.helloSeq = self.helloSeq + 1
                xmlStr = self.HelloBuildXmlMessage('request', self.helloId, self.helloSeq)
                try:
                    self.helloTxSock.sendto(xmlStr, (healthIp, self.peerHealthPort))
                except socket.error as e:
                    # Log the socket error and also control the log rate if this exception continuously happens
                    if self.helloTxSocErrCounter == HealthCheck._MAX_SOCKET_ERROR:
                        self.Log.log_error("helloTxSock send reply failed to dest %r, socket error : %s " % (healthIp, (e)))
                        self.helloTxSocErrCounter = 0
                    self.helloTxSocErrCounter += 1
                self.HelloIncOutReqs()
                self.log_debug("TX message: %s to %r" % (xmlStr, (healthIp, self.peerHealthPort)))

            if self.helloRapidDetection:
                helloInterval = self.helloRapidInterval
            else:
                helloInterval = self.helloRegularInterval
            self.helloTxEvent.wait(helloInterval)
            self.helloTxEvent.clear()

    def HelloRapidDetectionSetup(self):
        if not self.helloRapidDetection:
            self.Log.log('HealthCheck: Rapid detection start')
            self.HelloClearOutReqs()
            self.helloRapidDetection = True
            self.helloTxEvent.set()

    def HelloRapidDetectionDone(self, nlm):
        if self.helloRapidDetection:
            logstr = 'active' if self.backupActive else 'inactive'
            self.Log.log('HealthCheck: Rapid detection done; backup %s\n' % logstr)
            self.helloRapidDetection = False
            self.Intf.ClearPeerDeathPending(nlm)

    def HelloSendPeerIpMismatch(self):
        # When peer-ip is not correctly configured, send peeripmismatch
        # message via backup link
        healthIp = self.Parser.args.backupIp
        if healthIp and self.AttemptVrfBind():
            clagDataAttr = {'version' : self.Parser._ClagHelloVersion, 'peerIpMismatch' : 'yes'}
            root = ET.Element("clag_hello_data", clagDataAttr)
            xmlStr = ET.tostring(root)
            try:
                self.helloTxSock.sendto(xmlStr, (healthIp, self.peerHealthPort))
            except:
                self.log_debug("helloTxSock send backup peerIpMismatch message failed")
            self.log_debug("TX message: %s to %r" % (xmlStr, (healthIp, self.peerHealthPort)))
            self.HelloClearOutReqs()
            self.helloTxEvent.set()

    def HelloSendGoodbye(self):
        # Send goodbye via backup link
        healthIp = self.Parser.args.backupIp
        if healthIp and self.AttemptVrfBind():
            clagDataAttr = {'version' : self.Parser._ClagHelloVersion, 'goodbye' : 'yes'}
            root = ET.Element("clag_hello_data", clagDataAttr)
            xmlStr = ET.tostring(root)
            try:
                self.helloTxSock.sendto(xmlStr, (healthIp, self.peerHealthPort))
            except:
                self.log_debug("helloTxSock send backup goodbye failed")
            self.log_debug("TX message: %s to %r" % (xmlStr, (healthIp, self.peerHealthPort)))
            self.HelloClearOutReqs()
            self.helloTxEvent.set()

    def isPeerLinkDown(self):
        peerIf = self.Parser.args.peerIf
        if self.Intf.isSubIfName(peerIf):
            peerIf = self.Intf.GetMasterOfSubIf(peerIf)
        return self.Intf.GetOperStateFromSysFs(peerIf) != "up"

    def initDelayBringup(self):
        if self.Parser.args.initDelay:
            return not self.initDelayEvent.is_set()

    def InitDelay(self):
        nlm = nlmanager.nlmanager.NetlinkManager(clag.util.CLAG_NLM_INIT_DELAY_ID)
        while self.keep_going:
            self.initDelayEvent.clear()
            self.initDelayStartTime = int(os.times()[4])
            self.initState = True
            self.initDelayEvent.wait(self.Parser.args.initDelay)
            self.initDelayTimerStartEvent.clear()
            self.initDelayEvent.set()
            if self.keep_going:
                self.Log.log('HealthCheck: Delayed bring up of clag bonds for %d seconds' % self.Parser.args.initDelay)
                nlmsgs = ""
                self.Intf.protoDownLock.acquire()
                for intf in self.Parser.GetClagIntfDB().values():
                    if intf.protoDownFlags & intf._PROTO_DOWNF_INIT_DELAY:
                        clearFlags = intf._PROTO_DOWNF_INIT_DELAY
                        nlmsgs += intf.UpdateProtoDownFlags(nlm, 0, clearFlags, batch=True)
                nlm.tx_nlpacket_raw(nlmsgs)
                self.Intf.protoDownLock.release()
                if self.Intf.anycastIpAddDelayed:
                    self.Intf.anycastIpAddDelayed = False
                    anycastIp = self.Parser.args.vxlanAnycast
                    if anycastIp:
                        self.Log.log('HealthCheck: lo addr %s delayed add' % str(anycastIp))
                        self.Parser.args.vxlanAnycastDeleted = False
                        self.Intf.AddIpAddress(nlm, "lo", str(anycastIp), 32)
                self.initState = False
            self.initDelayTimerStartEvent.wait()

    def reloadExpired(self):
        return self.helloReloadEvent.is_set()

    def HelloReload(self):
        nlm = nlmanager.nlmanager.NetlinkManager(clag.util.CLAG_NLM_RELOAD_ID)
        while self.keep_going:
            self.helloReloadEvent.wait(self.Parser.args.reloadTimer)
            self.helloReloadEvent.set()
            if self.keep_going:
                firstTime = True
                nlmsgs = ""
                self.Intf.protoDownLock.acquire()
                for intf in self.Parser.GetClagIntfDB().values():
                    if intf.protoDownFlags & intf._PROTO_DOWNF_INIT:
                        clearFlags = intf._PROTO_DOWNF_INIT
                        if firstTime:
                            self.Log.log("HealthCheck: reload timeout.")
                            if self.backupActive and self.backupRole == "secondary":
                                self.Intf.SetProtoPeerOrBackupState(nlm)
                            else:
                                self.Intf.SetClagRole(nlm, self.Parser._ROLE_PRIMARY, "Reload timeout", False)
                            firstTime = False
                        nlmsgs += intf.UpdateProtoDownFlags(nlm, 0, clearFlags, batch=True)
                nlm.tx_nlpacket_raw(nlmsgs)
                self.Intf.protoDownLock.release()
                return
