#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014-2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import copy
    import nlmanager.nlmanager
    import os
    import re
    import threading
    import json
    import clag.util as util
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class lacpsync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the LACP information
    between two CLAG peers.
    """

    _LACP_DB_FLAGS_NONE     = 0
    _LACP_DB_FLAGS_DUAL     = (1 << 0)

    LACP_DB_IDX_AD_PARTNER_MAC = 0
    LACP_DB_IDX_BRIDGE_PORT_ID = 1
    LACP_DB_IDX_CLAG_ID = 2
    LACP_DB_IDX_OPER_STATE = 3
    LACP_DB_IDX_FLAGS = 4
    LACP_DB_IDX_LACP_BYPASS = 5

    def __init__(self, Daemon, Intf, Parser, Log, peerSendQueue, FdbSync):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(lacpsync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon = Daemon
        self.Intf   = Intf
        self.Parser = Parser
        self.Log    = Log
        self.Cmd    = None
        self.DataSyncs = []
        self.peerSendQueue = peerSendQueue
        self.FdbSync = FdbSync

        self.ignoreClagId = Parser.args.forceDynamic

        #
        #   LACP Database - There are two LACP databases, one for our switch and one for
        #   the peer switch. Each has the same structure. The database is a dictionary
        #   whose keys are the interface names of all bonds in the system with valid
        #   LACP partners, except for the peer link, and the value is a tuple with five
        #   elements. The first element is a string representation of the ad_partner_mac,
        #   the second element is the bridge port ID, the third is the clagId, the fourth
        #   is the operational state, and the fifth is the flags (dual connected). If
        #   the spanning tree is not enabled, or single instance spanning tree is not
        #   enabled, or the port is not in the bridge, then the bridge ID is None.
        #   {
        #       "bond2" : ( "00:00:00:00:00:01", None, 1, "up", 1 ),
        #       "bond5" : ( "00:00:00:00:00:02", "8.003", 2, "down", 0 )
        #   }
        #
        # if you add a new field to the lacpDb tuple please update LACP_DB_IDX_
        self.ourLacpDBLock      = threading.Lock()
        self.ourLacpDB          = {}
        self.peerLacpDBLock     = threading.Lock()
        self.peerLacpDB         = {}

        #
        #   Start a thread to handle updates from the peer switch
        #
        self.keep_going         = True
        self.stopEvent          = clag.clagthread.Event()
        self.newLacpDataEvent   = clag.clagthread.Event()

        self.updateThread = threading.Thread(None, self.UpdateLacpFromPeerT, "UpdateLacpFromPeer")
        self.updateThread.daemon = True
        self.updateThread.start()

        self.configThread = threading.Thread(None, self.UpdateLacpConfigT, "UpdateLacpConfig")
        self.configThread.daemon = True
        self.configThread.start()

    def __del__(self):
        """
        Destructor:
        """
        #
        #   Signal the update thread to wake up and exit
        #
        self.SignalThreads()
        self.JoinThreads()

    def UpdateLacpFromPeerT(self):
        try:
            self.Log.log("[%d]Init UpdateLacpFromPeerT" % (self.Parser.getTid()))
            self.UpdateLacpFromPeer()
        except:
            self.Daemon.DumpTraceback()

    def UpdateLacpConfigT(self):
        try:
            self.Log.log("[%d]Init UpdateLacpConfigT" % (self.Parser.getTid()))
            self.UpdateLacpConfig()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        self.newLacpDataEvent.set()

    def JoinThreads(self):
        self.updateThread.join()
        self.configThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_LACP_SYNC, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_LACP_SYNC, logStr)

    def OperStToDispStr(self, operSt):
        if operSt == "dormant" or operSt == "up":
            operStr = operSt
        else:
            operStr = "down"
        return operStr

    def FlagsToDispStr(self, flags):
        flagDisp = []
        if (flags & self._LACP_DB_FLAGS_DUAL):
            flagDisp.append("dual")
        return flagDisp

    def GetObjectSize(self):
        lacpObj = {}
        self.ourLacpDBLock.acquire()
        lacpObj['ourlacpDB'] = util.getSize(self.ourLacpDB)
        self.ourLacpDBLock.release()
        self.peerLacpDBLock.acquire()
        lacpObj['peerlacpDB'] = util.getSize(self.peerLacpDB)
        self.peerLacpDBLock.release()
        return lacpObj

    def __str__(self):
        """
        Return local information in human-readable form
        """
        lacpDb = {}
        self.ourLacpDBLock.acquire()
        for bond in self.ourLacpDB:
            lacpDbInfo = {}
            lacpInfo = self.ourLacpDB[bond]
            lacpDbInfo['operState'] = self.OperStToDispStr(lacpInfo[3])
            lacpDbInfo['flags'] = self.FlagsToDispStr(lacpInfo[4])
            lacpDbInfo['partnerMac'] = lacpInfo[0]
            if lacpInfo[1]:
                lacpDbInfo['cistPortId'] = lacpInfo[1]
            lacpDbInfo['clagId'] = lacpInfo[2]
            lacpDbInfo['lacpBypass'] = lacpInfo[self.LACP_DB_IDX_LACP_BYPASS]
            lacpDb[bond] = lacpDbInfo
        self.ourLacpDBLock.release()
        return json.dumps(lacpDb, sort_keys=True, indent=4)

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        lacpDb = {}
        self.peerLacpDBLock.acquire()
        for bond in self.peerLacpDB:
            lacpDbInfo = {}
            lacpInfo = self.peerLacpDB[bond]
            lacpDbInfo['operState'] = self.OperStToDispStr(lacpInfo[3])
            lacpDbInfo['flags'] = self.FlagsToDispStr(lacpInfo[4])
            lacpDbInfo['partnerMac'] = lacpInfo[0]
            if lacpInfo[1]:
                lacpDbInfo['cistPortId'] = lacpInfo[1]
            lacpDbInfo['clagId'] = lacpInfo[2]
            lacpDbInfo['lacpBypass'] = lacpInfo[self.LACP_DB_IDX_LACP_BYPASS]
            lacpDb[bond] = lacpDbInfo
        self.peerLacpDBLock.release()
        return json.dumps(lacpDb, sort_keys=True, indent=4)

    def GetOurLacpDB(self):
        """
        Return a copy of the current LACP data base on this switch.
        """
        self.ourLacpDBLock.acquire()
        lacpDB = copy.deepcopy(self.ourLacpDB)
        self.ourLacpDBLock.release()
        return lacpDB

    def GetOurLacpInfo(self, bond):
        """
        Return a copy of the current LACP data for a specific bond
        """
        self.ourLacpDBLock.acquire()
        if bond in self.ourLacpDB:
            lacpInfo = copy.deepcopy(self.ourLacpDB[bond])
        else:
            lacpInfo = None
        self.ourLacpDBLock.release()
        return lacpInfo

    def GetPeerLacpDB(self):
        """
        Return a copy of the current LACP data base on the peer switch.
        """
        self.peerLacpDBLock.acquire()
        lacpDB = copy.deepcopy(self.peerLacpDB)
        self.peerLacpDBLock.release()
        return lacpDB

    def GetPeerLacpInfo(self, bond):
        """
        Return a copy of the peer LACP data for a specific bond
        """
        self.peerLacpDBLock.acquire()
        if bond in self.peerLacpDB:
            lacpInfo = copy.deepcopy(self.peerLacpDB[bond])
        else:
            lacpInfo = None
        self.peerLacpDBLock.release()
        return lacpInfo

    def GetPeerLacpInfoByClagId(self, clagId):
        """
        Return a copy of the peer LACP data for a specific bond
        """
        lacpInfo = None
        self.peerLacpDBLock.acquire()
        for bond in self.peerLacpDB:
            if clagId == self.peerLacpDB[bond][2]:
                lacpInfo = copy.deepcopy(self.peerLacpDB[bond])
                break
        self.peerLacpDBLock.release()
        return lacpInfo

    def SetOurLacpDB(self, lacpDB):
        """
        Set the current value of the LACP data base on this switch.
        """
        self.ourLacpDBLock.acquire()
        if self.ourLacpDB != lacpDB:
            self.ourLacpDB = copy.deepcopy(lacpDB)
            self.newLacpDataEvent.set()
        self.ourLacpDBLock.release()

    def SetDormantBondOperState(self, nlm, bondName):
        if not self.IsOperStatePeerCheckNeeded(bondName):
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'Move %s to up immediately' % bondName)
            self.Intf.SetOperStateUp(nlm, bondName)
            return False

        updateDone = False
        self.ourLacpDBLock.acquire()
        if bondName in self.ourLacpDB and self.ourLacpDB[bondName][3] != "dormant":
            # XXX - cleaner way to do this?
            oldInfo = copy.deepcopy(self.ourLacpDB[bondName])
            lacpInfo = (oldInfo[0], oldInfo[1], oldInfo[2], "dormant",
                        oldInfo[4], oldInfo[self.LACP_DB_IDX_LACP_BYPASS])
            self.ourLacpDB[bondName] = lacpInfo
            updateDone = True
            # XXX - need a way to avoid complete table analysis eveytime
            # a single bond is updated
            self.newLacpDataEvent.set()
        self.ourLacpDBLock.release()
        if updateDone:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, '%s dormant update sent to peer' % bondName)
            self.SendBondAdd(bondName)
        return True

    def SetBondFlags(self, bondName, flags):
        updateDone = False
        self.ourLacpDBLock.acquire()
        if bondName in self.ourLacpDB and not (self.ourLacpDB[bondName][4] & flags):
            oldInfo = copy.deepcopy(self.ourLacpDB[bondName])
            lacpInfo = (oldInfo[0], oldInfo[1], oldInfo[2], oldInfo[3],
                        oldInfo[4]|flags, oldInfo[self.LACP_DB_IDX_LACP_BYPASS])
            self.ourLacpDB[bondName] = lacpInfo
            updateDone = True
        self.ourLacpDBLock.release()
        if updateDone:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, '%s dual-connect update sent to peer' % bondName)
            self.SendBondAdd(bondName)
        return updateDone

    def SetPeerLacpDB(self, lacpDB):
        """
        Return a copy of the current LACP data base on the peer switch.
        """
        self.peerLacpDBLock.acquire()
        if self.peerLacpDB != lacpDB:
            self.peerLacpDB = copy.deepcopy(lacpDB)
            self.newLacpDataEvent.set()
        self.peerLacpDBLock.release()

    def AddBondToPeerLacpDB(self, bondName, bondInfo):
        """
        Adds/updates a single element to the peer lacp db
        """
        updateDone = False
        self.peerLacpDBLock.acquire()
        if bondName not in self.peerLacpDB or self.peerLacpDB[bondName] != bondInfo:
            self.peerLacpDB[bondName] = copy.deepcopy(bondInfo)
            updateDone = True
            # XXX - need a way to avoid complete table analysis eveytime
            # a single bond is updated
            self.newLacpDataEvent.set()
        self.peerLacpDBLock.release()
        if updateDone:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer bond %s add %s 0x%x' % (bondName, bondInfo[3], bondInfo[4]))

    def IsOperStatePeerCheckNeeded(self, bond):
        '''
        In the following cases we don't need to co-ordinate bond state bringup
        with the peer -
            1. peer is not alive
            2. peer is alive but doesn't understand the new state machine
            3. clagId is not present in the peerDb
            4. peer-bond is down on the peer
            5. peer has already dual connected the bond - this is a tricky one
               and would result in problem windows when the oper state changes
               on the local bond repeatedly and rapidly - XXX need to revist
               once everything becomes event driven.
            6. LACP bond is up and ad_partner_mac is 00:00:00:00:00:00. This is
               the LACP bypass mode and the bonds are kept single-connected.
            7. partner mac mismatch - this is an interesting case; we should
               hold one side down ideally - till that is done the bond will
               be stuck in a dormant state - XXX.
        '''
        if not self.Intf.isPeerAlive():
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer dead; oper-up %s' % (bond))
            return False

        if not self.Parser.isPeerStaticClagStateAware():
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer is older version; oper-up %s' % (bond))
            return False

        clagId = self.Parser.GetBondClagId(bond)
        peerInfo = self.GetPeerLacpInfoByClagId(clagId) if clagId else None
        if not peerInfo:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'no peer bond for clagId %d; oper-up %s' % (clagId, bond))
            return False

        if not self.IsOperReadyForDualConnect(peerInfo[3]):
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'clagId %d peer bond oper %s; oper-up %s' % (clagId, peerInfo[3], bond))
            return False

        if self.IsOperReadyForDualConnect(peerInfo[3]) and\
                     peerInfo[self.LACP_DB_IDX_LACP_BYPASS]:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'clagId %d peer bond oper %s peer partner mac %s; oper-up %s' % (clagId, peerInfo[3], peerInfo[0], bond))
            return False

        if (peerInfo[4] & self._LACP_DB_FLAGS_DUAL):
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'clagId %d peer bond already dual-conn 0x%x; oper-up %s' % (clagId, peerInfo[4], bond))
            return False

        lacpInfo = self.GetOurLacpInfo(bond)
        # lacp bypass case - bypass bonds are not going to be dual connected
        # as long as the partner mac is null
        if lacpInfo:
            if self.IsOperReadyForDualConnect(lacpInfo[3]) and\
                 lacpInfo[self.LACP_DB_IDX_LACP_BYPASS]:
                return False
        else:
            lacpBypass = 0
            for slave in self.Intf.GetMembersOfBond(bond):
                lacpBypass |= self.Intf.GetLacpBypass(slave)
            if lacpBypass:
                return False

        return True

    def IsPartnerMacConflicting(self, lacpInfo, peerLacpInfo):
        return self.IsOperReadyForDualConnect(lacpInfo[3]) and \
            self.IsOperReadyForDualConnect(peerLacpInfo[3]) and \
            lacpInfo[0]     != '00:00:00:00:00:00' and \
            peerLacpInfo[0] != '00:00:00:00:00:00' and \
            lacpInfo[0]     != peerLacpInfo[0]

    def IsOperReadyForDualConnect(self, operState):
        return operState in ["up", "dormant"] and self.Intf.isPeerAlive() \
            and self.Parser.isClientSockOpen()

    def IsPartnerMacReadyForDualConnect(self, partnerMac):
        if partnerMac and partnerMac != "00:00:00:00:00:00":
            return True
        return False

    def BuildMacWithClagId(self, clagId):
        if not clagId:
            return None

        # We build the partner MAC for static bond by overwritting the last two
        # bytes of the sysMac with the clag id and setting the second bit of the
        # most significant byte (to 1) to make it a locally administered MAC
        mac = self.Parser.args.sysMac
        tmp = mac.split(':')
        tmp[5] = '%02x'%(clagId&0xff)
        tmp[4] = '%02x'%(clagId>>8&0xff)
        msb = int(tmp[0],16)|2
        tmp[0] = '%02x'%msb
        return ':'.join(tmp)

    def CollectLocalInfo(self, nlm, SendBondUpdate=None):
        """
        This function is called on a periodic interval to gather up any 
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """
        # Get a list of all of the CLAG bonds in the system, except for the peer interface
        bonds = self.Intf.GetAllClagBonds([self.Parser.args.peerIf])

        # Get the LACP neighbor information for each member that has such information
        newLacpDB = {}
        for bond in bonds:
            operState = self.Intf.GetOperState(bond)
            clagId = self.Parser.GetBondClagId(bond)
            mode = self.Intf.GetBondMode(bond)
            if mode == 4:
                partnerMac = self.Intf.GetBondPartnerMac(bond)
            else:
                partnerMac = self.BuildMacWithClagId(clagId)
            if not partnerMac:
                partnerMac = "00:00:00:00:00:00"
            flags = self._LACP_DB_FLAGS_NONE
            if self.Intf.isDualIf(bond):
                flags |= self._LACP_DB_FLAGS_DUAL
            lacpBypass = 0
            for slave in self.Intf.GetMembersOfBond(bond):
                tmpLacpByPass = self.Intf.GetLacpBypass(slave)
                lacpBypass |= tmpLacpByPass
                self.log_debug("%s:%s: %d" % (bond, slave, tmpLacpByPass))
            newLacpDB[bond] = (partnerMac, None, self.Parser.GetBondClagId(bond), operState, flags, lacpBypass)

        self.log_debug("The partner MACs for the bonds are:")
        for bond in newLacpDB:
            self.log_debug("    " + bond + ": " + newLacpDB[bond][0] + " " + str(newLacpDB[bond][1]) + " " + str(newLacpDB[bond][2]) + " " + str(newLacpDB[bond][3]) + " " + str(newLacpDB[bond][4]) + " " + str(newLacpDB[bond][5]),)
        self.SetOurLacpDB(newLacpDB)
        if SendBondUpdate:
            # temporary workaround to send update on a bond oper up/down event (i.e: do not wait for periodic update)
            bond = SendBondUpdate
            self.SendBondAdd(bond)

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML.

        parent - An xml.etree.ElementTree.Element instance which is the parent
            element under which all of the information should be encoded.

        <bonds>
            <bond name="bond4" ad_partner_mac="00:02:00:00:00:09" portId="8.003" clagId=1/>
            <bond name="bond5" ad_partner_mac="00:02:00:00:00:0b" portId="8.005" />
            <bond name="bond0" ad_partner_mac="00:02:00:00:00:24" portId="8.009" />
            <bond name="bond1" ad_partner_mac="00:02:00:00:00:03" portId="8.001" />
            <bond name="bond2" ad_partner_mac="00:02:00:00:00:05" portId="8.004" />
            <bond name="bond3" ad_partner_mac="00:02:00:00:00:07" />
        </bonds>
        """
        lacpDB = self.GetOurLacpDB()
        bonds = ET.SubElement(parent, "bonds")
        newPeer = self.Parser.isPeerStaticClagStateAware()
        for bond in lacpDB:
            bondInfo = lacpDB[bond]
            # check if the peer can handle down bonds
            if not self.IsOperReadyForDualConnect(bondInfo[3]) and not newPeer:
                continue
            self.EncodeBondXmlInfo(bonds, bond, bondInfo, "sync")
         

    def EncodeBondXmlInfo(self, bonds, bond, bondInfo, op):
        bondAttrib = { "name" : bond, "ad_partner_mac" : bondInfo[0] }
        if bondInfo[1] is not None:
            bondAttrib["portId"] = str(bondInfo[1])
        bondAttrib["clagId"] = str(bondInfo[2])
        bondAttrib["operState"] = str(bondInfo[3])
        bondAttrib["flags"] = str(bondInfo[4])
        bondAttrib["lacpBypass"] = str(bondInfo[self.LACP_DB_IDX_LACP_BYPASS])
        bondAttrib["op"] = op
        ET.SubElement(bonds, "bond", bondAttrib)

    def SendBondAdd(self, bond):
        if not self.Intf.isPeerAlive() or not self.Parser.isPeerStaticClagStateAware():
            return
        bondInfo = self.GetOurLacpInfo(bond)
        if bondInfo:
            clagDataAttr = {'version' : self.Parser._ClagDataVersion}
            root = ET.Element("clag_data", clagDataAttr)
            bonds = ET.SubElement(root, "bond-updates")
            self.EncodeBondXmlInfo(bonds, bond, bondInfo, "add")
            xmlStr = ET.tostring(root)
            self.peerSendQueue.put(xmlStr)
            util.queueSetMaxLen('peerSendQueue')

    def DecodeBondXmlInfo(self, bonds):
        lacpDB = {}
        for bond in bonds.findall('bond'):
            name = bond.get('name')
            partner = bond.get('ad_partner_mac')
            portid = bond.get('portId')
            clagId = int(bond.get('clagId', 0))
            operState = bond.get('operState', "up")
            flags = int(bond.get('flags', 0))
            lacpBypass = int(bond.get('lacpBypass', 0))
            op = bond.get('op', "sync")
            # XXX - should we remove the "partner not None" condition in
            # prep for future static bond support
            if name is not None and partner is not None:
                lacpInfo = ( partner, portid, clagId, operState, flags, lacpBypass )
                if op == "sync":
                    lacpDB[name] = lacpInfo
                elif op == "add":
                    self.AddBondToPeerLacpDB(name, lacpInfo)
        return lacpDB

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.

        root - The root element of the XML tree
        """
        bonds = root.find('bonds')
        if bonds is not None:
            lacpDB = self.DecodeBondXmlInfo(bonds)
            self.SetPeerLacpDB(lacpDB)

        bonds = root.find('bond-updates')
        if bonds is not None:
            self.DecodeBondXmlInfo(bonds)
        if bonds:
            del bonds

    def DualIfChange(self, dualIfs, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        # XXX - currently we only care when a bond becomes dual connected
        # this needs to change to take care of rapid link state changes.
        if enable:
            for bond in dualIfs:
                self.SetBondFlags(bond, self._LACP_DB_FLAGS_DUAL)

    def MatchIdIfChange(self, dualIf, enable):
        """
        Called when an interface gets or loses a matching clag-id on the peer
        """
        pass

    def ClearPeerInfo(self):
        """
        Removes information which has been obtained from the peer.
        """
        if self.peerLacpDBLock.acquire(False):
            if self.peerLacpDB != {}:
                self.peerLacpDB = {}
                self.newLacpDataEvent.set()
            self.peerLacpDBLock.release()

    def ClearLocalInfo(self):
        """
        Removes information which has been obtained locally.
        """
        self.SetOurLacpDB({})

    def PeerLacpDualConnProc(self, nlm, clagId):
        '''
        Peer has dual connected this clag bond. We can move the local bond to
        oper up state if it is in a dormant state.
        '''
        bond = self.Parser.GetBondFromClagId(clagId)
        if bond and self.Intf.GetOperStateFromKCache(bond) == "dormant":
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer dual connected clagId %d; oper-up %s' % (clagId, bond))
            self.Intf.SetOperStateUp(nlm, bond)

    def PeerLacpOperDownProc(self, nlm, clagId):
        '''
        Peer bond is not ready for dual connection (not up or dormant). We can 
        move the local bond to oper up state if it is in a dormant state.
        '''
        bond = self.Parser.GetBondFromClagId(clagId)
        if bond and self.Intf.GetOperStateFromKCache(bond) == "dormant":
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer clagId %d oper down; oper-up %s' % (clagId, bond))
            self.Intf.SetOperStateUp(nlm, bond)

    def EvaluateClagIntfConflicts(self, bond, clagId, lacpDB=None):
        setConflicts = self.Parser._CLAG_INTF_CONFLICT_NONE 
        lacpInfo = self.GetOurLacpInfo(bond)
        if lacpInfo and self.IsOperReadyForDualConnect(lacpInfo[3]) and\
                self.IsPartnerMacReadyForDualConnect(lacpInfo[0]):
            if lacpInfo[0] == self.Parser.args.sysMac:
                setConflicts |= self.Parser._CLAG_INTF_CONFLICT_LACP_MAC_EQ_SYS_MAC
            if not lacpDB:
                lacpDB = self.GetOurLacpDB()
            for k in lacpDB:
                if bond != k and lacpInfo[0] == lacpDB[k][0]:
                    setConflicts |= self.Parser._CLAG_INTF_CONFLICT_LACP_MAC_DUP

        # The remaining conflict checks can be skipped if the bond is already
        # dual connected
        if self.Intf.isDualIf(bond):
            return setConflicts

        peerLacpInfo = self.GetPeerLacpInfoByClagId(clagId)
        if peerLacpInfo:
            if lacpInfo and\
                    self.IsPartnerMacConflicting(lacpInfo, peerLacpInfo):
                setConflicts |= self.Parser._CLAG_INTF_CONFLICT_LACP_PARTNER_MAC_MISMATCH
        else:
            setConflicts |= self.Parser._CLAG_INTF_CONFLICT_NO_PEER
        return setConflicts

    def EvaluateAndUpdateConflicts(self):
        if not self.Intf.ReadyToRunClagConflictChecks() or self.Parser.sendGoodbye:
            return

        # If an interface is not dual-connected we need to look for
        # config/topological inconsistencies and report them to the admin
        clagBonds = self.Parser.GetClagBondDB()
        lacpDB = self.GetOurLacpDB()
        # list of conflicts that we are evaluating here
        for bond in clagBonds:
            clagId = self.Parser.GetBondClagId(bond)
            setConflicts = self.EvaluateClagIntfConflicts(bond, clagId, lacpDB)
            self.Parser.UpdateClagConflicts(bond, setConflicts,
                 self.Parser._CLAG_INTF_CONFLICT_LACP_ALL&~setConflicts)

    def FlushMacsOnDownIf(self, iface):
        if not self.Parser.args.redirectEnable:
            return

        lacpInfo = self.GetOurLacpInfo(iface)
        clagId = self.Parser.GetBondClagId(iface)
        if not lacpInfo or not clagId:
            return
        operSt = lacpInfo[self.LACP_DB_IDX_OPER_STATE]
        if operSt in ["up", "dormant"]:
            return

        peerInfo = self.GetPeerLacpInfoByClagId(clagId)
        if not peerInfo:
            return
        operSt = peerInfo[self.LACP_DB_IDX_OPER_STATE]
        if operSt in ["up", "dormant"]:
            return

        # if local and remote bonds are down flush macs in our-db
        self.Log.log_debug(self.Log._LOG_DEBUG_FSM,\
                           "clag %d down on both sides; flush macs on %s" %\
                           (clagId, iface))
        self.FdbSync.FlushMacsOnDownIf(iface)

    def ProcPeerOperStateChanges(self, currClagIdMatchIfs, newClagIdMatchIfs):
        for iface in newClagIdMatchIfs:
            currIf = currClagIdMatchIfs.get(iface)
            if not currIf:
                continue

            newIf = newClagIdMatchIfs.get(iface)
            if currIf[self.Intf.CLAG_ID_MATCHIF_IDX_RIF] ==\
                newIf[self.Intf.CLAG_ID_MATCHIF_IDX_RIF] and\
                currIf[self.Intf.CLAG_ID_MATCHIF_IDX_RIF_STATE] !=\
                newIf[self.Intf.CLAG_ID_MATCHIF_IDX_RIF_STATE]:
                self.log_debug("iface %s:%s oper state changed on peer" %\
                       (iface, newIf[self.Intf.CLAG_ID_MATCHIF_IDX_RIF]))
                self.FlushMacsOnDownIf(iface)

    def UpdateLacpFromPeer(self):
        '''
        This function is an independent thread of the script. It compares the MAC
        address data base received from the peer (peerMacDB) with the mac addresses
        we have already programmed into the hardware. Any new addresses are added and
        any addresses no longer present on the peer are removed.
        '''
        self.log_verbose("Beginning execution of the thread UpdateLacpFromPeer")

        # Initially, no interfaces are dual connected
        nlm = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_LACP_PEER_ID)
        self.DualHomeIf(nlm, self.Intf.GetAllBonds([self.Parser.args.peerIf]), False)
        currDualIfs = {}
        currClagIdMatchIfs = {}

        while self.keep_going:
            # Wait for a change in our bonds or the peer's bonds
            self.log_verbose("Waiting for an event before continuing execution of the thread UpdateLacpFromPeer")
            self.newLacpDataEvent.wait()
            self.newLacpDataEvent.clear()
            self.log_verbose("Executing the thread UpdateLacpFromPeer.")

            # Sets are used to determine which interfaces are newly or no longer dual conencted
            lacpDB = self.GetOurLacpDB()
            peerDB = self.GetPeerLacpDB()
            newClagIdMatchIfs = {}
            for lacpIntf in lacpDB:
                if lacpDB[lacpIntf][self.LACP_DB_IDX_LACP_BYPASS]:
                    continue
                for peerIntf in peerDB:
                    if peerDB[peerIntf][self.LACP_DB_IDX_LACP_BYPASS]:
                        continue
                    if lacpDB[lacpIntf][2] == peerDB[peerIntf][2]:
                        newClagIdMatchIfs[lacpIntf] = (peerIntf,\
                            peerDB[peerIntf][self.LACP_DB_IDX_OPER_STATE])
                        break
            nowClagIdMatchIfs = set(newClagIdMatchIfs) - set(currClagIdMatchIfs)
            notClagIdMatchIfs = set(currClagIdMatchIfs) - set(newClagIdMatchIfs)
            self.Intf.AddClagIdMatchIfs(newClagIdMatchIfs)
            self.Intf.DelClagIdMatchIfs(notClagIdMatchIfs)

            # Filter out bonds that are not eligible for dual connection
            if not self.Intf.isPeerAlive() and self.Parser.isLacpFlushPending():
                self.log_debug('In the process of going down. Ignoring local LACP updates.')
                ourLacpDB = lacpDB
            else:
                ourLacpDB = {k:lacpDB[k] for k in lacpDB if self.IsOperReadyForDualConnect(lacpDB[k][3]) and self.IsPartnerMacReadyForDualConnect(lacpDB[k][0])}

            # On the peer - If the operState of the bond is down or if it has
            # become dual connected we can move the corresponding local bond
            # to an oper up state (provided it was oper down)
            for bond in peerDB:
                if (peerDB[bond][4] & self._LACP_DB_FLAGS_DUAL):
                    self.PeerLacpDualConnProc(nlm, peerDB[bond][2])
                elif not self.IsOperReadyForDualConnect(peerDB[bond][3]):
                    self.PeerLacpOperDownProc(nlm, peerDB[bond][2])
            # XXX - currently there is no easy way to get a list of deleted
            # peer bonds so need to do a blanket update. This is inefficient and
            # must be removed once everything becomes event driven
            self.Intf.UpdatePendingDormantBonds(nlm)

            # Filter out bonds in the peerDB that are not eligible for dual 
            # connection
            if not self.Intf.isPeerAlive() and self.Parser.isLacpFlushPending():
                self.log_debug('In the process of going down. Ignoring peer LACP updates.')
                peerLacpDB = peerDB
            else:
                peerLacpDB = {k:peerDB[k] for k in peerDB if self.IsOperReadyForDualConnect(peerDB[k][3]) and self.IsPartnerMacReadyForDualConnect(peerDB[k][0])}

            if self.ignoreClagId:
                ourMacs = { v[0] for v in ourLacpDB.values() }
                peerMacs = { v[0] for v in peerLacpDB.values() }
                dualMacs = ourMacs & peerMacs
                newDualIfs = { i for i in ourLacpDB if ourLacpDB[i][0] in dualMacs }
            else:
                ourMacs = { (v[0], v[2]) for v in ourLacpDB.values() }
                peerMacs = { (v[0], v[2]) for v in peerLacpDB.values() }
                dualMacs = ourMacs & peerMacs
                newDualIfs = { i for i in ourLacpDB if (ourLacpDB[i][0], ourLacpDB[i][2]) in dualMacs }

            nowDualIfs = newDualIfs - set(currDualIfs)
            notDualIfs = set(currDualIfs) - newDualIfs

            self.log_debug("currDualIfs : %s" % (currDualIfs,))
            self.log_debug("newDualIfs  : %s" % (newDualIfs,))
            self.log_debug("nowDualIfs  : %s" % (nowDualIfs,))
            self.log_debug("notDualIfs  : %s" % (notDualIfs,))

            # Handle the new dually connected interfaces
            if nowDualIfs:
                self.DualHomeIf(nlm, nowDualIfs, True)

            # Handle the interfaces which are no longer dual connected
            if notDualIfs:
                self.Intf.DelDualIfs(notDualIfs)
                self.DualHomeIf(nlm, notDualIfs, False)

            # Update the current notion of dual connected hosts in the global database.
            theDualIfs = {}
            for nowDualIf in nowDualIfs:
                for peerIntf in peerLacpDB:
                    if ourLacpDB[nowDualIf][0] == peerLacpDB[peerIntf][0] and \
                            (self.ignoreClagId or ourLacpDB[nowDualIf][2] == peerLacpDB[peerIntf][2]):
                        theDualIfs[nowDualIf] = peerIntf
                        break
                self.Log.log("%s is now dual connected." % (nowDualIf,))
                peerData = peerLacpDB[peerIntf]
                self.Cmd.CmdGetAllOutput("add mac-%s = %s,%s" % (peerData[0], nowDualIf, peerData[1]))
            self.Intf.AddDualIfs(theDualIfs)

            for notDualIf in notDualIfs:
                self.Log.log("%s is no longer dual connected." % (notDualIf,))
                self.Cmd.CmdGetAllOutput("del mac-%s = %s,%s" % (ourLacpDB.get(notDualIf,["00:00:00:00:00:00"])[0], notDualIf, "None"))

            for sync in self.DataSyncs:
                sync.DualIfChange(nowDualIfs, True)

            for sync in self.DataSyncs:
                sync.DualIfChange(notDualIfs, False)

            for sync in self.DataSyncs:
                sync.MatchIdIfChange(nowClagIdMatchIfs, True)

            for sync in self.DataSyncs:
                sync.MatchIdIfChange(notClagIdMatchIfs, False)

            # Re-run the consistency checks
            self.EvaluateAndUpdateConflicts()
            currDualIfs = newDualIfs
            self.ProcPeerOperStateChanges(currClagIdMatchIfs, newClagIdMatchIfs)
            currClagIdMatchIfs = newClagIdMatchIfs

        self.log_verbose("Finished execution of the thread UpdateLacpFromPeer")

    def DualHomeIf(self, nlm, dualIfs, enable):
        '''
        Enable or disable the forwarding logic for a dual homed interface by
        setting or clearing the duallink attribute on the bridge.
        '''
        # Create a set of all dual connected sub interfaces
        masterVlanMap = self.Intf.GetMasterIfVlanMap()
        dualSet = set(dualIfs)
        for dualIf in dualIfs:
            dualSet |= set(masterVlanMap.get(dualIf, {}).values())

        # Change the role of the interface to be dual connected or not.
        self.Intf.bridgeAttrLock.acquire()
        self.Intf.SetBridgeDualLink(nlm, dualSet, enable)
        self.Intf.bridgeAttrLock.release()

    def UpdateLacpConfig(self):
        '''
        This function is an independent thread of the script. It periodically 
        runs to check if the learning priority and learning filter are properly
        set. It is possible that due to links going up and down outside of the
        control of CLAG, these settings can be modified.
        '''
        self.log_verbose("Beginning execution of the thread UpdateLacpConfig")
        runCnt = 0
        nlm = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_LACP_ID)

        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution of the thread UpdateLacpConfig")
            self.stopEvent.wait(self.Parser.getPeriodicRun())
            self.log_verbose("Executing the thread UpdateLacpConfig")

            if self.keep_going:
                runCnt += 1
                # Update the kernel link_mode in case it has fallen out of 
                # sync with clagd
                if not runCnt % 3:
                    self.Intf.CheckAndUpdateIntfModeDormant(nlm)

                # Update the proto_down in the interface driver in case it has
                # fallen out of sync with clagd
                if not runCnt % 4:
                    nlmsgs = ""
                    self.Intf.protoDownLock.acquire()
                    for intf in self.Parser.GetClagBondDB().values():
                        nlmsgs += intf.UpdateIntfProtoDown(nlm, batch=True)
                    nlm.tx_nlpacket_raw(nlmsgs)
                    self.Intf.protoDownLock.release()


                # Get the current dual interface list and all sub-interfaces
                self.Intf.bridgeAttrLock.acquire()
                masterVlanMap = self.Intf.GetMasterIfVlanMap()
                currDualIfs = self.Intf.GetDualIfs()
                dualSet = set(currDualIfs.keys())
                for dualIf in dualSet.copy():
                    dualSet |= set(masterVlanMap.get(dualIf, {}).values())

                # Get the current clag bonds and all the sub-interfaces
                clagBonds = self.Parser.GetClagBondDB()
                clagBondSet = set(clagBonds)
                for clagBond in clagBonds:
                    clagBondSet |= set(masterVlanMap.get(clagBond, {}).values())

                # Get the current peer interface, and all sub-interfaces
                peerMasters = set([self.Parser.args.peerIf])
                if self.Intf.isSubIfName(self.Parser.args.peerIf):
                    peerMasters = set(self.Intf.GetLogicalMastersOfSubIf(self.Parser.args.peerIf))
                peerSet = peerMasters.copy()
                for peerMaster in peerMasters:
                    peerSet |= set(masterVlanMap.get(peerMaster, {}).values())

                dualLinkEnables = []
                learningDisables = []
                peerLinkEnables = []
                noFlushEnables = []
                for brName in self.Intf.GetAllBridges():
                    for brIf in self.Intf.GetMembersOfBridge(brName):
                        # Is it dual interface with duallink not set to "on"?
                        if brIf in dualSet and not self.Intf.isBridgeDualLink(brIf):
                            dualLinkEnables += [brIf]
                        # Is it peer interface with learning not set to "off"?
                        if brIf in peerSet and self.Intf.isBridgeLearning(brIf):
                            learningDisables += [brIf]
                        # Is it peer interface with peerlink not set to "on"?
                        if brIf in peerSet and not self.Intf.isBridgePeerLink(brIf):
                            peerLinkEnables += [brIf]
                        # Is it a clag bond with no_flush disabled?
                        if brIf in clagBondSet and not self.Intf.isBridgeNoFlush(brIf):
                            noFlushEnables += [brIf]
                self.Intf.SetBridgeDualLink(nlm, dualLinkEnables, True)
                self.Intf.SetBridgeLearning(nlm, learningDisables, False)
                self.Intf.SetBridgePeerLink(nlm, peerLinkEnables, True)
                self.Intf.SetBridgeNoFlush(nlm, noFlushEnables, True)
                self.Intf.bridgeAttrLock.release()

        self.log_verbose("Finished execution of the thread UpdateLacpConfig")

