#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014-2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import clag.util as util
    import hashlib
    import ipaddr
    import json
    import nlmanager.nlmanager
    import nlmanager.nlpacket
    import Queue
    import socket
    import struct
    import threading
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError(str(e) + "- required module not found")

class neighsync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the Neighbor
    Database, between two CLAG peers.
    """

    # Hardware work queue operation codes.
    _NEIGH_ADD = 1
    _NEIGH_DELETE = 2
    _NEIGH_PERM_ADD = 3
    _NEIGH_PERM_DEL = 4
    _NEIGH_ADD_EXCL = 5

    # Neigh Work Queue operation codes
    _WORK_QUEUE_ADD = 1
    _WORK_QUEUE_DELETE = 2
    _WORK_QUEUE_NOP = 3
    _WORK_QUEUE_DUMP_DONE = 4
    _WORK_QUEUE_ADDR_ADD = 5
    _WORK_QUEUE_ADDR_DEL = 6

    # Neigh Data
    NEIGH_DB_FAMILY = 0
    NEIGH_DB_MAC    = 1
    NEIGH_DB_IFACE  = 2
    NEIGH_DB_OURS   = 3
    NEIGH_DB_PEERS  = 4
    NEIGH_DB_STALE  = 5
    NEIGH_DB_ROUTER  = 6

    # Interesting Neigh States/Flags
    NUD_REACHABLE = 0x02
    NUD_STALE = 0x04
    NUD_PROBE = 0x10
    NUD_FAILED = 0x20
    NUD_NOARP = 0x40
    NUD_PERMANENT = 0x80
    NTF_EXT_LEARNED = 0x10
    NTF_ROUTER = 0x80

    def __init__(self, Daemon, Intf, Parser, Log, sendQueue, clagnl):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(neighsync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon = Daemon
        self.Intf = Intf
        self.Parser = Parser
        self.Log = Log
        self.sendQueue = sendQueue
        self.clagnl = clagnl

        #
        #   Neigh Database - Only local (i.e. non offload) and non-permanent
        #   entries on SVIs will be considered for this database.
        #   A neigh is added to that database when it is reachable on at least
        #   one of the two clag peers. A neigh will be owned by both sides when
        #   once it becomes reachable on both clag peer.
        #   (IP, vlanId) : [family MAC localIf ourNeigh peerNeigh, STALE]
        #
        #   {                         0             1                2       3     4
        #    ("11.4.1.10", 104) : [AF_INET, "00:00:00:00:00:02", "vlan104", True, True, False]
        #    ("11.4.1.11", 104) : [AF_INET, "00:00:00:00:00:04", "vlan104", True, False, False]
        #    ("11.4.1.12", 104) : [AF_INET, "00:00:00:00:00:07", "vlan104", False, True, False]
        #   }
        #

        self.neighDBLock = threading.Lock()
        self.neighDB = {}

        self.kernDBLock = threading.Lock()
        self.kernDB = {}

        self.permanentNeighDBLock = threading.Lock()
        self.permanentNeighDB = {}

        # netlink notification sent to kernel are processed by this queue
        self.toKernWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.toKernWorkQueue, 'toKernWorkQueue')

        # netlink notification received by kernel are processed by this queue
        self.fromKernWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.fromKernWorkQueue, 'fromKernWorkQueue')
        self.clagnl.SetNeighSync(self)

        self.keep_going = True
        self.stopEvent = clag.clagthread.Event()

        self.updateFromKernelThread = threading.Thread(None, self.UpdateFromKernelT,
                                                "UpdateFromKernel")
        self.updateFromKernelThread.daemon = True
        self.updateFromKernelThread.start()

        self.updateToKernelThread = threading.Thread(None, self.UpdateToKernelT,
                                               "UpdateToKernel")
        self.updateToKernelThread.daemon = True
        self.updateToKernelThread.start()

        if self.Parser.args.initDelay:
            self.SyncDelayEvent = clag.clagthread.Event()
            self.SyncDelayThread = threading.Thread(None, self.SyncDelayT, "SyncDelayT")
            self.SyncDelayThread.daemon = True
            self.SyncDelayThread.start()

    def __del__(self):
        """
        Destructor:
        """
        self.SignalThreads()
        self.JoinThreads()

    def UpdateFromKernelT(self):
        try:
            self.Log.log("[%d]Init UpdateFromKernelT" % (self.Parser.getTid()))
            self.UpdateFromKernel()
        except:
            self.Daemon.DumpTraceback()

    def UpdateToKernelT(self):
        try:
            self.Log.log("[%d]Init UpdateToKernelT" % (self.Parser.getTid()))
            self.UpdateToKernel()
        except:
            self.Daemon.DumpTraceback()

    def SyncDelayT(self):
        try:
            self.Log.log("[%d]Init SyncDelayT" % (self.Parser.getTid()))
            self.SyncDelay()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        try:
            self.toKernWorkQueue.put_nowait((None, None, None, None, None, None, None, None))
            util.queueSetMaxLen('toKernWorkQueue')
        except Queue.Full:
            pass
        try:
            self.fromKernWorkQueue.put_nowait((neighsync._WORK_QUEUE_NOP, ()))
            util.queueSetMaxLen('fromKernWorkQueue')
        except Queue.Full:
            pass
        try:
            self.SyncDelayEvent.set()
        except AttributeError:
            pass

    def JoinThreads(self):
        self.updateFromKernelThread.join()
        self.updateToKernelThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_NEIGH_SYNC, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_NEIGH_SYNC, logStr)

    def KernWorkQueuePut(self, element):
        if self.keep_going:
            self.toKernWorkQueue.put(element)
            util.queueSetMaxLen('toKernWorkQueue')

    def GetObjectSize(self):
        neighObj = {}
        self.neighDBLock.acquire()
        neighObj['neighDB'] = util.getSize(self.neighDB)
        self.neighDBLock.release()
        self.kernDBLock.acquire()
        neighObj['kernNeighDB'] = util.getSize(self.kernDB)
        self.kernDBLock.release()
        return neighObj

    def EncodedStr(self):
        """
        Return local information in human-readable form
        """
        neighDb = []
        self.neighDBLock.acquire()
        for (dst, vlanId), neigh in self.neighDB.iteritems():
            neighInfo = {}
            neighInfo['dst'] = dst
            neighInfo['mac'] = neigh[self.NEIGH_DB_MAC]
            neighInfo['dev'] = neigh[self.NEIGH_DB_IFACE]
            neighInfo['vid'] = vlanId
            neighInfo['ourNeigh'] = neigh[self.NEIGH_DB_OURS]
            neighInfo['peerNeigh'] = neigh[self.NEIGH_DB_PEERS]
            neighDb.append(neighInfo)
        self.neighDBLock.release()
        return json.dumps(neighDb, sort_keys=True, indent=4)

    def EncodedKernStr(self):
        """
        Return kernel neigh information in human-readable form
        """
        kernNeighDb = []
        self.kernDBLock.acquire()
        for (dst, idx), [family, mac, state, flags] in self.kernDB.iteritems():
            kernNeighInfo = {}
            kernNeighInfo['dst'] = dst
            kernNeighInfo['idx'] = idx
            kernNeighInfo['family'] = family
            kernNeighInfo['mac'] = mac
            kernNeighInfo['state'] = state
            kernNeighInfo['flags'] = flags
            kernNeighDb.append(kernNeighInfo)
        self.kernDBLock.release()
        return json.dumps(kernNeighDb, sort_keys=True, indent=4)

    def EncodedPermNeighStr(self):
        """
        Return permanent neigh DB information in human-readable form
        """
        permanentNeighDb = []
        self.permanentNeighDBLock.acquire()
        for (addr, vlanId), [family, mac, lif, local] in self.permanentNeighDB.iteritems():
            permanentNeighInfo = {}
            permanentNeighInfo['addr'] = addr
            permanentNeighInfo['dev'] = lif
            permanentNeighInfo['family'] = family
            permanentNeighInfo['mac'] = mac
            permanentNeighInfo['local'] = local
            permanentNeighInfo['vid'] = vlanId
            permanentNeighDb.append(permanentNeighInfo)
        self.permanentNeighDBLock.release()
        return json.dumps(permanentNeighDb, sort_keys=True, indent=4)

    def CreateDBCheckSum(self):
        self.neighDBLock.acquire()
        hlist = []
        sortedDB =  dict(sorted(self.neighDB.items()))
        for (dst,vlanId), neigh in sortedDB.iteritems():
            if self.Parser.clagRole == self.Parser._ROLE_PRIMARY:
                ourNeigh = "ownedByPrimary" if neigh[self.NEIGH_DB_OURS] else None
                peerNeigh = "ownedBySecondary" if neigh[self.NEIGH_DB_PEERS] else None
                hlist.extend([dst, neigh[self.NEIGH_DB_FAMILY],
                              neigh[self.NEIGH_DB_MAC], vlanId, ourNeigh, peerNeigh])
            else:
                ourNeigh = "ownedBySecondary" if neigh[self.NEIGH_DB_OURS] else None
                peerNeigh = "ownedByPrimary" if neigh[self.NEIGH_DB_PEERS] else None
                hlist.extend([dst, neigh[self.NEIGH_DB_FAMILY],
                              neigh[self.NEIGH_DB_MAC], vlanId, peerNeigh, ourNeigh])
        self.neighDBLock.release()
        return hashlib.md5(''.join(str(e) for e in hlist)).hexdigest()

    def SyncDelay(self):
        if self.Parser.args.initDelay:
            self.SyncDelayEvent.wait(self.Parser.args.initDelay/2)
            self.SyncDelayEvent.set()
            vlanMap = self.Intf.GetSubIfsVlanMap()
            for iface, (master, vid) in vlanMap.iteritems():
                if self.Intf.GetOperState(iface) == "up":
                    if self.Intf.isBridgeName(master) and self.Intf.isVlanFilteringBridge(master):
                        self.SyncOnIfaceUp(iface, True)
            bridgeMap = self.Intf.GetAllBridges()
            for bridge in bridgeMap:
                if not self.Intf.isVlanFilteringBridge(bridge):
                    self.SyncOnIfaceUp(bridge, False)

    def SyncOnIfaceUp(self, iface, vlanAware):
        if self.Parser.args.initDelay:
            if not self.SyncDelayEvent.is_set():
                return
        if vlanAware:
            vid = self.Intf.GetVlanIdOfSubIf(iface)
        else:
            brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
            vid = brPeerVlanMap.get(iface, None)
        self.log_debug("%s up, syncing neighs reachable on peer" % iface)
        self.neighDBLock.acquire()
        for (dst, vlanId), neigh in self.neighDB.iteritems():
            if vid == vlanId and not neigh[self.NEIGH_DB_OURS] and neigh[self.NEIGH_DB_PEERS]:
                self.IsUpdateToKernelNeeded(dst, neigh[self.NEIGH_DB_FAMILY], neigh[self.NEIGH_DB_MAC],
                                            iface, vlanId, neigh[self.NEIGH_DB_ROUTER])
        self.neighDBLock.release()
        self.permanentNeighDBLock.acquire()
        self.log_debug("%s up, syncing peer SVI addresses" % iface)
        for (addr, vlanId), [family, mac, lif, local] in self.permanentNeighDB.iteritems():
            #TODO: Need to add support for vlan unaware bridges
            if vlanAware and vid == vlanId and local == False:
                self.KernWorkQueuePut((neighsync._NEIGH_PERM_ADD, addr, family, mac, vlanId, lif, self.NUD_PERMANENT, 0))
        self.permanentNeighDBLock.release()

    def SyncOnIfaceMacChange(self, vid, newMac):
        self.log_debug("Permanent Neigh Update for Svi Mac Change on %d mac %s" % (vid, newMac))
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0
            
        self.permanentNeighDBLock.acquire()
        for (addr, vlanId), [family, oldMac, lif, local] in self.permanentNeighDB.iteritems():
            if vid == vlanId and local == True:
                self.permanentNeighDB[(addr, vlanId)] = [family, newMac, lif, True]
                self.AddXmlEntry(root, "addr_add", addr, family, newMac, vlanId, None)
                xmlEntries += 1

            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
        self.permanentNeighDBLock.release()

        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(root)
            self.sendQueue.put(xmlStr)

    def DelNeighEntries(self, iface, vid):
        """
        This function is called when an interface is deleted from the netlink
        cache, when this happens, we need to clear the neighs owned by the
        interface which is being deleted.
        """
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0
        self.log_debug("%s being deleted, removing reachable neighs from neighDB" % iface)
        self.neighDBLock.acquire()
        ignoreLock = True
        for (dst, vlanId), neigh in self.neighDB.copy().iteritems():
            family = neigh[self.NEIGH_DB_FAMILY]
            mac = neigh[self.NEIGH_DB_MAC]
            lif = neigh[self.NEIGH_DB_IFACE]
            routerFlag = neigh[self.NEIGH_DB_ROUTER]
            if vid == vlanId:
                xmlEntries += self.ClearLocalOwnershipFlag(root, dst, family, mac, lif,
                                                           vlanId, routerFlag, ignoreLock)
            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
        self.neighDBLock.release()
        self.permanentNeighDBLock.acquire()
        for (addr, vlanId), [family, mac, lif, local] in self.permanentNeighDB.copy().iteritems():
            if local == True and vlanId == vid:
                self.AddXmlEntry(root, "addr_del", addr, family, mac, vlanId, None)
                xmlEntries += 1
                self.permanentNeighDB.pop((addr, vlanId), None)
            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
        self.permanentNeighDBLock.release()
        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(root)
            self.sendQueue.put(xmlStr)

    def CollectLocalInfo(self, nlm):
        """
        This function is called on a periodic interval to gather up any
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """
        pass

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML for sending to the peer on a
        periodic basis.
        """
        if self.Parser.syncDoneToPeer:
            return

        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0
        self.neighDBLock.acquire()
        for (dst, vlanId), neigh in self.neighDB.iteritems():
            if neigh[self.NEIGH_DB_OURS]:
                self.AddXmlEntry(root, "sync", dst, neigh[self.NEIGH_DB_FAMILY],
                                 neigh[self.NEIGH_DB_MAC], vlanId, neigh[self.NEIGH_DB_ROUTER])
                xmlEntries += 1
            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
        self.neighDBLock.release()
        self.permanentNeighDBLock.acquire()
        for (addr, vlanId), [family, mac, lif, local] in self.permanentNeighDB.copy().iteritems():
            if local == True:
                self.AddXmlEntry(root, "addr_add", addr, family, mac, vlanId, None)
                xmlEntries += 1
            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
        self.permanentNeighDBLock.release()
        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(root)
            self.sendQueue.put(xmlStr)
        self.Log.log("Initial neigh sync to peer done.")


    def PrintElement(self, node, level):
        self.log_verbose("%s%s: %s: %s" % (" " * (level*3), node.tag, node.text,
                                           str(node.attrib)))
        for child in node.findall('*'):
            self.PrintElement(child, level+1)

    def GetWorkQueueDelayed(self, delay=0.001):
        """
        Returns the next entry on the neigh work queue, waiting up to the
        specified delay (default 1 millisecond) for an entry.
        """
        queueEmpty = False
        try:
            (opcode, params) = self.fromKernWorkQueue.get(True, delay)
        except Queue.Empty:
            opcode = neighsync._WORK_QUEUE_NOP
            params = ()
            queueEmpty = True
        return (opcode, params, queueEmpty)

    def AddXmlEntry(self, element, op, dst, family, mac, vlanId, flags=None):
        """
        Adds an Xml entry to the root element
        """
        ET.SubElement(element, "neigh", {"op" : op, "dst" : str(dst),
                                         "mac" : mac, "vlanId" : str(vlanId),
                                         "family" : str(family), "flags" : str(flags)})

    def ClearPeerInfo(self):
        """
        Remove the peer ownership flag from all the entries in the neigh DB
        """
        self.neighDBLock.acquire()
        for (dst, vlanId), neigh in self.neighDB.copy().iteritems():
            if neigh[self.NEIGH_DB_OURS] == True:
                if neigh[self.NEIGH_DB_PEERS] != False:
                    self.log_debug("Clearing peer ownership for %s mac %s dev %s "
                                   % (dst, neigh[self.NEIGH_DB_MAC], neigh[self.NEIGH_DB_IFACE]))
                    neigh[self.NEIGH_DB_PEERS] = False
            else:
                # we don't even own it so remove the entry
                self.log_debug("Deleting neighbor %s mac %s dev %s from neighDB"
                               % (dst, neigh[self.NEIGH_DB_MAC], neigh[self.NEIGH_DB_IFACE]))
                self.neighDB.pop((dst, vlanId), None)
        self.neighDBLock.release()
        self.permanentNeighDBLock.acquire()
        for (addr, vlanId), [family, mac, lif, local] in self.permanentNeighDB.copy().iteritems():
            if local == False:
                self.log_debug("Deleting peer SVI address from the permanentNeighDB: %s %s %s" % (lif, addr, mac))
                self.KernWorkQueuePut((neighsync._NEIGH_PERM_DEL, addr, family, mac, vlanId, lif, self.NUD_PERMANENT, 0))
        self.permanentNeighDBLock.release()

    def GetVlanIdFromValidEntry(self, k_dst, k_if):
        """
        This function checks if the neigh entry is valid. If not, do not process
        it. If it is a valid entry, it returns the vlan ID.
        """
        k_vlanId = None
        if not k_if:
            return
        if self.Intf.isSubIfName(k_if):
            k_vlanId = self.Intf.GetVlanIdOfSubIf(k_if)
            k_master = self.Intf.GetMasterOfSubIf(k_if)
            if not k_master:
                self.log_verbose("Ignoring neigh %s because interface %s "
                               "has no master" % (k_dst, k_if))
                return
            if not self.Intf.isBridgeName(k_master):
                self.log_verbose("Ignoring neigh %s because master "
                               "of %s is not a bridge" % (k_dst, k_if))
                return
            if not self.Intf.isVlanFilteringBridge(k_master):
                self.log_verbose("Ignoring neigh %s because master of %s is not "
                               "a vlan aware bridge" % (k_dst, k_if))
                return
        elif self.Intf.isBridgeName(k_if) and not self.Intf.isVlanFilteringBridge(k_if):
            brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
            k_vlanId = brPeerVlanMap.get(k_if, None)
        else:
            self.log_verbose("Ignoring neigh %s dev %s" % (k_dst, k_if))
            return
        return k_vlanId

    def SetLocalOwnershipFlag(self, element, dst, family, mac, lif, vlanId, flags):
        # this neigh is reachable now, we now own it
        xmlEntry = 0
        routerFlag = False
        self.neighDBLock.acquire()
        if (dst, vlanId) not in self.neighDB:
            self.log_debug("Adding neighbor %s mac %s dev %s in neighDB, now "
                           "reachable" % (dst, mac, lif))
            self.neighDB[(dst, vlanId)] = [family, mac, lif, True, False, False, flags]
        else:
            self.neighDB[(dst, vlanId)][self.NEIGH_DB_STALE] = False
            c_mac = self.neighDB[(dst, vlanId)][self.NEIGH_DB_MAC]
            c_if = self.neighDB[(dst, vlanId)][self.NEIGH_DB_IFACE]
            # is it a different mac/iface/vid now?
            if mac != c_mac:
                self.log_debug("Updating neighbor %s mac %s dev %s in neighDB"
                               % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][self.NEIGH_DB_MAC] = mac
            elif lif != c_if:
                # only lif changed
                self.neighDB[(dst, vlanId)][self.NEIGH_DB_IFACE] = lif
            if self.neighDB[(dst, vlanId)][self.NEIGH_DB_OURS] != True:
                self.log_debug("Neighbor %s mac %s dev %s is now reachable, "
                               "updating neighDB" % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][self.NEIGH_DB_OURS] = True
        if family == socket.AF_INET6 and flags & self.NTF_ROUTER:
            routerFlag = True
        self.neighDB[(dst, vlanId)][self.NEIGH_DB_ROUTER] = routerFlag
        self.AddXmlEntry(element, "add", dst, family, mac, vlanId, flags)
        xmlEntry = 1
        self.neighDBLock.release()
        return xmlEntry

    def ClearLocalOwnershipFlag(self, element, dst, family, mac, lif, vlanId, r_flag, ignoreLocks=False):
        # no longer reachable, we don't own this neigh anymore
        xmlEntry = 0
        if not ignoreLocks:
            self.neighDBLock.acquire()
        if (dst, vlanId) in self.neighDB:
            self.neighDB[(dst, vlanId)][self.NEIGH_DB_STALE] = False
            c_mac = self.neighDB[(dst, vlanId)][self.NEIGH_DB_MAC]
            if self.neighDB[(dst, vlanId)][self.NEIGH_DB_PEERS] == True:
                if mac and mac != c_mac:
                    # mac changed and neigh no longer reachable
                    self.neighDB[(dst, vlanId)][self.NEIGH_DB_OURS] = False
                elif self.neighDB[(dst, vlanId)][self.NEIGH_DB_OURS] != False:
                    # owned by peer so try to re-add
                    self.log_debug("Neighbor %s mac %s dev %s no longer reachable, "
                                   "updating neighDB" % (dst, c_mac, lif))
                    self.neighDB[(dst, vlanId)][self.NEIGH_DB_OURS] = False
                mac = c_mac if not mac else mac
                self.IsUpdateToKernelNeeded(dst, family, mac, lif, vlanId, r_flag)
            else:
                # not owned by peer so remove the entry
                self.log_debug("Deleting neighbor %s mac %s dev %s in neighDB"
                               % (dst, c_mac, lif))
                self.neighDB.pop((dst, vlanId), None)
            self.AddXmlEntry(element, "delete", dst, family, c_mac, vlanId)
            xmlEntry = 1
        if not ignoreLocks:
            self.neighDBLock.release()
        return xmlEntry

    def ClearLocalStaleEntries(self):
        neigh_del_list = []
        self.neighDBLock.acquire()
        for (dst, vlanId), neigh in self.neighDB.iteritems():
            neigh_del_list.append((dst, vlanId))
        self.neighDBLock.release()
        self.ClearLocalNeighOwnership(neigh_del_list)

    def ClearLocalNeighOwnership(self, neighs):
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0
        for neigh in neighs:
            dst = neigh[0]
            family = self.neighDB[neigh][self.NEIGH_DB_FAMILY]
            mac = self.neighDB[neigh][self.NEIGH_DB_MAC]
            iface = self.neighDB[neigh][self.NEIGH_DB_IFACE]
            vlanId = neigh[1]
            flags = None
            xmlEntries += self.ClearLocalOwnershipFlag(root, dst, family, mac, iface, vlanId, flags)

            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0

        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(root)
            self.sendQueue.put(xmlStr)

    def SetPeerOwnershipFlag(self, dst, family, mac, lif, vlanId, r_flag):
        self.neighDBLock.acquire()
        if (dst, vlanId) not in self.neighDB:
            self.log_debug("Adding neighbor %s mac %s dev %s in neighDB, reachable "
                           "on peer" % (dst, mac, lif))
            self.neighDB[(dst, vlanId)] = [family, mac, lif, False, True, False, r_flag]
            self.IsUpdateToKernelNeeded(dst, family, mac, lif, vlanId, r_flag)
        else:
            c_mac = self.neighDB[(dst, vlanId)][self.NEIGH_DB_MAC]
            c_flag = self.neighDB[(dst, vlanId)][self.NEIGH_DB_ROUTER]
            # is it a different mac or vid now?
            if mac != c_mac:
                self.log_debug("Updating neighbor %s mac %s dev %s in neighDB"
                               % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][self.NEIGH_DB_MAC] = mac
            if self.neighDB[(dst, vlanId)][self.NEIGH_DB_PEERS] != True:
                self.log_debug("Neighbor %s mac %s dev %s now reachable on "
                               "peer, updating neighDB" % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][self.NEIGH_DB_PEERS] = True
            if (self.neighDB[(dst, vlanId)][self.NEIGH_DB_OURS] != True or mac != c_mac or r_flag != c_flag):
                self.IsUpdateToKernelNeeded(dst, family, mac, lif, vlanId, r_flag)
        self.neighDBLock.release()

    def ClearPeerOwnershipFlag(self, dst, family, mac, lif, vlanId, flags):
        self.neighDBLock.acquire()
        if (dst, vlanId) in self.neighDB:
            c_mac = self.neighDB[(dst, vlanId)][self.NEIGH_DB_MAC]
            if self.neighDB[(dst, vlanId)][self.NEIGH_DB_OURS] == True:
                if self.neighDB[(dst, vlanId)][self.NEIGH_DB_PEERS] != False:
                    self.log_debug("Neighbor %s mac %s dev %s not reachable on "
                                   "peer, updating neighDB" % (dst, mac, lif))
                    self.neighDB[(dst, vlanId)][self.NEIGH_DB_PEERS] = False
            else:
                # we don't even own it so remove the entry
                self.log_debug("Deleting neighbor %s mac %s dev %s from neighDB"
                               % (dst, c_mac, lif))
                self.neighDB.pop((dst, vlanId), None)
        self.neighDBLock.release()

    def LocalPermanentNeighUpdate(self, element, opcode, params):
        xmlEntry = 0
        k_add = (opcode == neighsync._WORK_QUEUE_ADDR_ADD)
        (k_addr, k_family, k_null, k_idx, k_null, k_null) = params
        k_bif  = self.Intf.GetIfName(k_idx)
        master = self.Intf.GetMasterOfSubIf(k_bif)
        vlanId = self.Intf.GetVlanIdOfSubIf(k_bif)
        mac = self.Intf.GetIfMac(k_bif)
        if k_add and (not master or not self.Intf.isVlanFilteringBridge(master) or not vlanId \
           or not self.Intf.isBridgeName(master) or not self.Intf.isSubIfName(k_bif)):
            return xmlEntry
        self.permanentNeighDBLock.acquire()
        if k_add:
            self.log_debug("Adding SVI address to the permanentNeighDB: %s %s %s" % (k_bif, k_addr, mac))
            self.permanentNeighDB[(k_addr, vlanId)] = [k_family, mac, k_bif, True]
            self.AddXmlEntry(element, "addr_add", k_addr, k_family, mac, vlanId, None)
            xmlEntry = 1
        elif not k_add and (k_addr, vlanId) in self.permanentNeighDB:
            family = self.permanentNeighDB[(k_addr, vlanId)][0]
            mac = self.permanentNeighDB[(k_addr, vlanId)][1]
            lif = self.permanentNeighDB[(k_addr, vlanId)][2]
            self.log_debug("Deleting SVI address from the permanentNeighDB: %s %s %s" % (lif, k_addr, mac))
            self.AddXmlEntry(element, "addr_del", k_addr, family, mac, vlanId, None)
            xmlEntry = 1
            self.permanentNeighDB.pop((k_addr, vlanId), None)
        self.permanentNeighDBLock.release()
        return xmlEntry

    def PeerPermanentNeighUpdate(self, opcode, addr, family, mac, vlanId, lif):
        r_add = (opcode == "addr_add")
        self.permanentNeighDBLock.acquire()
        if r_add:
            self.log_debug("Adding peer SVI address to the permanentNeighDB: %s %s %s" % (lif, addr, mac))
            self.permanentNeighDB[(addr, vlanId)] = [family, mac, lif, False]
            self.KernWorkQueuePut((neighsync._NEIGH_PERM_ADD, addr, family, mac, vlanId, lif, self.NUD_PERMANENT, 0))
        else:
            self.log_debug("Deleting peer SVI address from the permanentNeighDB: %s %s %s" % (lif, addr, mac))
            self.permanentNeighDB.pop((addr, vlanId), None)
            self.KernWorkQueuePut((neighsync._NEIGH_PERM_DEL, addr, family, mac, vlanId, lif, self.NUD_PERMANENT, 0))
        self.permanentNeighDBLock.release()

    def UpdateKernelDB(self, opcode, params):
        (k_dst, k_family, k_mac, k_idx, k_state, k_flags) = params
        self.kernDBLock.acquire()
        if opcode == neighsync._WORK_QUEUE_ADD:
            self.kernDB[(k_dst, k_idx)] = [k_family, k_mac, k_state, k_flags]
        else:
            self.kernDB.pop((k_dst, k_idx), None)
        self.kernDBLock.release()

    def IsUpdateToKernelNeeded(self, dst, family, mac, lif, vlanId, r_flag):
        """
        Re-add an entry to kernel when required
        """
        self.kernDBLock.acquire()
        neigh_add = neighsync._NEIGH_ADD
        add = True
        k_state = self.NUD_STALE
        k_flags = 0
        operState = self.Intf.GetOperState(lif)
        if operState != "up" or not self.SyncDelayEvent.is_set():
            # SVI not ready, do not add
            add = False
        idx = self.Intf.GetIfIndex(lif)
        if (dst, idx) in self.kernDB:
            k_family = self.kernDB[(dst, idx)][0]
            k_state = self.kernDB[(dst, idx)][2]
            k_flags = self.kernDB[(dst, idx)][3]
            kr_flag = True if (k_family == socket.AF_INET6 and k_flags & self.NTF_ROUTER) else False
            if k_state & (self.NUD_STALE|self.NUD_REACHABLE) and kr_flag == r_flag:
                # do not add if state is STALE/REACHABLE
                add = False
            if k_state & self.NUD_NOARP:
                # do not add if state is NOARP
                add = False
            if k_flags & self.NTF_EXT_LEARNED:
                # offload flag is on, do not add
                add = False
        else:
            # Set EXCL flag to prevent state (OFFLOAD ext) overwrites by clagd
            # when a neigh entry which is not initally in kernel is added by FRR
            # as an external entry
            neigh_add = neighsync._NEIGH_ADD_EXCL
        if add:
            state = k_state if k_state & (self.NUD_STALE|self.NUD_REACHABLE) else self.NUD_STALE
            flag = self.NTF_ROUTER if k_flags & self.NTF_ROUTER else 0
            self.KernWorkQueuePut((neigh_add, dst, family, mac, vlanId, lif, state, flag))
        self.kernDBLock.release()

    def IsPermanentNeighUpdated(self, addr, family, mac, lif, vlanId, state):
        self.permanentNeighDBLock.acquire()
        if state != self.NUD_PERMANENT and (addr, vlanId) in self.permanentNeighDB:
            mac = self.permanentNeighDB[(addr, vlanId)][1]
            iface = self.permanentNeighDB[(addr, vlanId)][2]
            local = self.permanentNeighDB[(addr, vlanId)][3]
            operState = self.Intf.GetOperState(lif)
            if operState == "up" and local ==  False and iface == lif:
                self.log_debug("Re-adding peer SVI address to the permanentNeighDB: %s %s %s" % (lif, addr, mac))
                self.KernWorkQueuePut((neighsync._NEIGH_PERM_ADD, addr, family, mac, vlanId, lif, self.NUD_PERMANENT, self.NTF_EXT_LEARNED))
        self.permanentNeighDBLock.release()

    def UpdateFromKernel(self):
        """
        This is an independent thread of the script. It is responsible for
        processing entries on the NeighWorkQueue, which may result in
        modifications to the neighDB and messages sent to the peer.
        """
        self.log_verbose("Beginning execution of the thread UpdateFromKernel")
        nlmgr = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_ADDR_DEL_ID)

        if self.clagnl.nlm is None:
            self.clagnl.kernelDumpReadyEvent.wait()
        self.clagnl.DumpNeighs()

        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution "
                             "of the thread UpdateFromKernel")
            (opcode, k_params) = self.fromKernWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
                while not queueEmpty and self.keep_going:
                    if opcode == neighsync._WORK_QUEUE_NOP:
                        (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                        continue
                    elif opcode == neighsync._WORK_QUEUE_ADDR_ADD or opcode == neighsync._WORK_QUEUE_ADDR_DEL:
                        self.log_debug("Processing address entry: %d %s"
                                       % (opcode, str(k_params)))
                        if opcode == neighsync._WORK_QUEUE_ADDR_DEL:
                            anycastIp = self.Parser.args.vxlanAnycast
                            if (anycastIp and (self.Intf.GetIfName(k_params[3]) == "lo") and \
                                (self.Parser.args.vxlanAnycastDeleted == False) and \
                                    (k_params[0] == str(anycastIp)) and \
                                    (self.Intf.GetOperStateFromSysFs("lo") != "down")):
                                self.Log.log_error("Anycast Ip Deleted on lo by external app. Re-adding the anycast-ip on lo")
                                self.Intf.AddIpAddress(nlmgr, "lo", str(anycastIp), 32)
                        xmlEntries += self.LocalPermanentNeighUpdate(root, opcode, k_params)
                        (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                        continue
                    elif opcode == neighsync._WORK_QUEUE_DUMP_DONE:
                        neigh_del_list = []
                        self.neighDBLock.acquire()
                        for (dst, vlanId), neigh in self.neighDB.iteritems():
                            if neigh[self.NEIGH_DB_STALE]:
                                neigh_del_list.append((dst, vlanId))
                        self.neighDBLock.release()
                        if neigh_del_list:
                            self.ClearLocalNeighOwnership(neigh_del_list)
                        self.Log.log_warn('neigh resync with kernel complete')
                        (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                        continue

                    self.log_debug("Processing neighbor entry: %d %s"
                                   % (opcode, str(k_params)))
                    (k_dst, k_family, k_mac, k_idx, k_state, k_flags) = k_params

                    k_if  = self.Intf.GetIfName(k_idx)
                    k_vlanId = self.GetVlanIdFromValidEntry(k_dst, k_if)
                    r_flag = True if (k_family == socket.AF_INET6 and k_flags & self.NTF_ROUTER) else False
                    self.UpdateKernelDB(opcode, k_params)

                    if self.Parser.args.peerIp == 'linklocal' and k_if == self.Parser.args.peerIf \
                            and k_state & (self.NUD_STALE|self.NUD_REACHABLE) \
                            and k_dst.startswith("fe80:") and self.Parser.remotePeerIp != k_dst:
                        #linklocal event (use that IP as peerIp)
                        self.Log.log("Using %s as remote Peer IP" % k_dst)
                        self.Parser.remotePeerIp = k_dst
                    if not k_vlanId:
                        (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                        continue

                    neigh_add = None
                    if opcode == neighsync._WORK_QUEUE_ADD and k_state & self.NUD_REACHABLE:
                        # neigh reachable, need to add to neighDB
                        neigh_add = True

                    self.IsPermanentNeighUpdated(k_dst, k_family, k_mac, k_if, k_vlanId, k_state)
                    if neigh_add:
                        xmlEntries += self.SetLocalOwnershipFlag(root, k_dst, k_family, k_mac,
                                                                 k_if, k_vlanId, r_flag)
                    else:
                        #Do not clear local ownership for PROBE state as this is transient
                        if not k_state & self.NUD_PROBE:
                            xmlEntries += self.ClearLocalOwnershipFlag(root, k_dst, k_family, k_mac,
                                                                   k_if, k_vlanId, r_flag)

                    if xmlEntries >= 2000 and self.keep_going:
                        xmlStr = ET.tostring(root)
                        self.sendQueue.put(xmlStr)
                        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                        xmlEntries = 0

                    (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()

                if xmlEntries and self.keep_going:
                    #self.PrintElement(root, 0)
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)
        self.log_verbose("Finished execution of the thread UpdateFromKernel")

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.
        root - The root element of the XML tree
        """
        self.log_verbose("Beginning DecodeXmlInfo")

        for neigh in root.findall('neigh'):
            r_op = neigh.get('op')
            r_dst = neigh.get('dst')
            r_mac = neigh.get('mac')
            r_vlanId = neigh.get('vlanId')
            r_family = neigh.get('family')
            r_flags = neigh.get('flags')
            if r_family == "None":
                r_family = None
            if r_family is not None:
                r_family = int(r_family)
            if r_vlanId == "None":
                r_vlanId = None
            if r_vlanId is not None:
                r_vlanId = int(r_vlanId)
            r_flags = True if r_flags == "True" else False
            l_if = self.Intf.GetLocalIfBasedOnPeerVlanId(r_vlanId)

            if r_op in ["add", "sync"]:
                self.SetPeerOwnershipFlag(r_dst, r_family, r_mac, l_if, r_vlanId, r_flags)
            elif r_op == "delete":
                self.ClearPeerOwnershipFlag(r_dst, r_family, r_mac, l_if, r_vlanId, r_flags)
            elif r_op in ["addr_add", "addr_del"]:
                self.PeerPermanentNeighUpdate(r_op, r_dst, r_family, r_mac, r_vlanId, l_if)

        self.log_verbose("Ending DecodeXmlInfo")

    def DualIfChange(self, ifs, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        pass

    def UpdateToKernel(self):
        """
        This function is an independent thread of the script. It compares the
        neigh data base received from the peer with the neighs
        we have already programmed into the hardware. Any new neighs are added.
        """

        neigh_cmd = {
            neighsync._NEIGH_ADD : (nlmanager.nlpacket.RTM_NEWNEIGH,
                                    nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_REPLACE,
                                    "added to"),
            neighsync._NEIGH_ADD_EXCL : (nlmanager.nlpacket.RTM_NEWNEIGH,
                                    nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_EXCL,
                                    "created and added to"),
            neighsync._NEIGH_PERM_ADD : (nlmanager.nlpacket.RTM_NEWNEIGH,
                                    nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_REPLACE,
                                    "added (permanent) to"),
            neighsync._NEIGH_PERM_DEL : (nlmanager.nlpacket.RTM_DELNEIGH, 0, "removed (permanent) from")
        }

        self.log_verbose("Beginning execution of the thread UpdateToKernel")
        nlm = neigh_nlmanager(self.Log)

        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution of the thread UpdateToKernel")
            (op, dst, family, mac, vlanId, lif, state, flags) = self.toKernWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                params = []
                while not queueEmpty and self.keep_going:
                    (rtm_command, nlm_flags, neigh_str) = neigh_cmd.get(op, (None, None, None))
                    if rtm_command is not None:
                        intf = self.Intf.GetIfIndex(lif)
                        if not intf:
                            try:
                                (op, dst, family, mac, vlanId, lif, state, flags) = self.toKernWorkQueue.get(True, 0.001)
                                intf = self.Intf.GetIfIndex(lif)
                            except Queue.Empty:
                                queueEmpty = True
                            continue
                        params += [(rtm_command, nlm_flags, dst, mac, intf, family, state, flags)]
                        # TODO Move this to verbose
                        self.log_debug("Neighbor %s %s has been %s %s in VLAN %d"
                                       % (dst, mac, neigh_str, lif, vlanId))
                    else:
                        self.Log.log_error("Unknown op code in work queue: %s" % (op,))

                    try:
                        (op, dst, family, mac, vlanId, lif, state, flags) = self.toKernWorkQueue.get(True, 0.001)
                    except Queue.Empty:
                        queueEmpty = True

                nlm.neigh_batch(params)

        self.log_verbose("Finished execution of the thread UpdateToKernel")

class neigh_nlmanager(nlmanager.nlmanager.NetlinkManager):

    def __init__(self, Log):
        super(neigh_nlmanager, self).__init__(util.CLAG_NLM_NEIGH_ID)
        self.Log = Log

    def tx_or_concat_message(self, total_message, neigh):
        """
        Concat several of neighs together before TXing.
        """
        PACKET_CONCAT_SIZE = 16384

        if not total_message:
            total_message = neigh.message
        else:
            total_message += neigh.message

        if len(total_message) >= PACKET_CONCAT_SIZE:
            self.tx_nlpacket_raw(total_message)
            total_message = None

        return total_message

    def neigh_batch(self, neighs):

        total_message = None

        for (rtm_command, nlm_flags, dst, mac, intf, family, state, flags) in neighs:
            debugFlags = self.Log._LOG_DEBUG_NETLINK | self.Log._LOG_DEBUG_NEIGH_SYNC
            debug = (self.Log.getDebug() & debugFlags) == debugFlags
            neigh = nlmanager.nlpacket.Neighbor(rtm_command, debug, self.Log.logger,
                                                self.Log.getHandlerStr() != "syslog")
            neigh.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlm_flags
            neigh.family = family

            if family == socket.AF_INET:
                dst = ipaddr.IPv4Address(dst)
            else:
                dst = ipaddr.IPv6Address(dst)

            neigh.body = struct.pack(neigh.PACK, family, intf, state, flags, 0)
            neigh.add_attribute(neigh.NDA_LLADDR, mac)
            neigh.add_attribute(neigh.NDA_DST, dst)

            neigh.build_message(self.sequence.next(), self.pid)
            total_message = self.tx_or_concat_message(total_message, neigh)

        if total_message:
            self.tx_nlpacket_raw(total_message)
