#!/usr/bin/python

from ifupdown.iface import *
from ifupdown.utils import utils
from ifupdownaddons.modulebase import *
from ifupdownaddons.systemutils import systemUtils
from ifupdown.statemanager import statemanager_api as statemanager
from ipaddr import IPNetwork, IPv4Address, IPv4Network, AddressValueError
from time import asctime
import ifupdown.ifupdownflags as ifupdownflags
import ipaddr
import os
import re
import stat
import json
import socket
import time

class clagd(moduleBase):
    _modinfo = {
        'mhelp' : 'This module generates the clagd defaults file.',
        'attrs' : {
            'clagd-peer-ip' : {
                'help'     : 'The IP address of the clagd peer',
                'validvals' : ['<ipv4>', '<ipv6>', 'linklocal'],
                'required' : True,
                'example'  : ['clagd-peer 10.10.10.2']
            },
            'clagd-sys-mac' : {
                'help'     : 'The system ID of the CLAG pair',
                'validvals' : ['<mac>',],
                'required' : True,
                'example'  : ['clagd-sys-mac 44:38:39:ff:00:00']
            },
            'clagd-priority' : {
                'help'     : 'The priority of this clagd switch',
                'validrange' : ['0', '65535'],
                'example'  : ['clagd-priority 30000']
            },
            'clagd-enable' : {
                'help'     : 'enable clagd',
                'validvals' : ['yes', 'no'],
                'example'  : ['clagd-enable yes']
            },
            'clagd-args' : {
                'help'     : 'Additional command line arguments for clagd',
                'multiline' : True,
                'example'  : [
                    'clagd-args --log /var/log/clagd.log',
                    'clagd-args --verbose --lacpPoll 10',
                    'clagd-args --debug 0x4'
                ]
            },
            'clag-id' : {
                'help' : 'multi-chassis lag id',
                'validrange' : ['0', '65535'],
                'default' : '0',
                'example' : ['clag-id 1']
            },
            'clagd-backup-ip' : {
                'help'     : 'Backup IP address of the clagd peer',
                'validvals' : ['<ipv4-vrf-text>', ],
                'example'  : ['clagd-backup-ip 192.1.1.1',
                              'clagd-backup-ip 192.1.1.1 vrf green']
            }
        }
    }

    _clagd_default_file = "/etc/default/clagd"
    _clagd_default_perm = stat.S_IRUSR | stat.S_IRGRP |stat.S_IROTH
    _clagd_default_template = '''
# Dynamic clagd file is under control of ifupdown2
#     DO NOT EDIT THIS FILE BY HAND -- YOUR CHANGES WILL BE OVERWRITTEN
#     Generated at: %s
CLAGD_ENABLE="%s"
CLAGD_PEER_IF="%s"
CLAGD_SYS_MAC="%s"
CLAGD_PEER_IP="%s"
CLAGD_PRIORITY=%s
CLAGD_BACKUP_VRF="%s"
CLAGD_BACKUP_IP="%s"
CLAGD_VXLAN_ANYCAST_IP="%s"
CLAGD_ARGS="%s"
'''

    def __init__(self, *args, **kargs):
        moduleBase.__init__(self, *args, **kargs)
        if ifupdownflags.flags.PERFMODE:
            self.clagd_cache = {}
            try:
                os.unlink(clagd._clagd_default_file)
            except OSError:
                pass
        else:
            # read and cache the contents of /etc/default/clagd
            self.clagd_cache = self._clagd_params_read()
        # keep clag bond cache for query use
        self.clag_bonds = None
        self.clagd_peerlinks = set()
        self.clagd_peerlink_name = ""
        self.clagd_vxlan_anycast_ip = ""
        self.clagd_vxlan_anycast_ip_deleted = False
        self.clagd_peerlink_enable = False
        self.clagd_restart = False
        self.clagd_old_global_local_ip = ""
        self.clagd_global_local_ip = ""

    def get_dependent_ifacenames(self, ifaceobj, ifacenames_all=None):
        if not self.clagd_old_global_local_ip:
            for ifname, old_obj_list in statemanager.ifaceobjdict.items():
                if old_obj_list:
                    old_obj = old_obj_list[0]
                    if self._is_vxlan_device(old_obj):
                        self._set_old_global_local_ip(old_obj)
                    elif old_obj.name == 'lo':
                        self._set_old_global_local_ip(old_obj)
        if not self.clagd_peerlink_name and self._is_peerlink(ifaceobj):
            self.clagd_peerlink_name = ifaceobj.name
            self.clagd_peerlink_enable = self._is_peerlink_learn_enable(ifaceobj)
        elif self._is_vxlan_device(ifaceobj):
            self._set_global_local_ip(ifaceobj)
        elif ifaceobj.name == 'lo':
            self._set_global_local_ip(ifaceobj)
            # if its "lo" cache anycast ip needed to start clagd. ifreload builds dependency
            # graph twice, first for new interface instance and then for old instance with
            # previous configuration. Hence set clagd_vxlan_anycast_ip only if its not already
            # set and do not set it again if it is being deleted.
            if not self.clagd_vxlan_anycast_ip and not self.clagd_vxlan_anycast_ip_deleted:
                clagd_vxlan_list = ifaceobj.get_attr_value('clagd-vxlan-anycast-ip')
                if clagd_vxlan_list:
                    if len(clagd_vxlan_list) != 1:
                        self.log_warn('%s: multiple clagd-vxlan-anycast-ip lines, using first one'
                                      % (ifaceobj.name,))

                    anycast_addr = clagd_vxlan_list[0]

                    try:
                        anycast_addr_obj = ipaddr.IPv4Network(anycast_addr)

                        if anycast_addr_obj.prefixlen != 32:
                            self.logger.warning("%s: clagd-vxlan-anycast-ip %s: netmask ignored" % (ifaceobj.name, anycast_addr))
                            anycast_addr = "%s/32" % str(anycast_addr_obj.ip)
                        else:
                            anycast_addr = str(anycast_addr_obj.ip)
                    except:
                        self.clagd_vxlan_anycast_ip_deleted = True
                        anycast_addr = ""
                        self.log_error("%s: invalid clagd-vxlan-anycast-ip %s: must be in ipv4 format"
                                          % (ifaceobj.name, anycast_addr), ifaceobj)

                    self.clagd_vxlan_anycast_ip = anycast_addr
                else:
                    self.clagd_vxlan_anycast_ip_deleted = True
        return None

    def _check_peer_ip(self, ifaceobj):
        peerip = ifaceobj.get_attr_value_first('clagd-peer-ip')
        if peerip:
            if (peerip != 'linklocal' and
                not ifaceobj.get_attr_value_first('address')):
                self.log_error('%s: clag-peer-ip should be configured with '
                            'an IP address on the interface' %ifaceobj.name)
                return False
        return True

    def syntax_check(self, ifaceobj, ifaceobj_getfunc):
        if not ifaceobj.link_kind & ifaceLinkKind.BOND:
            if ifaceobj.get_attr_value_first('clag-id'):
                self.logger.error('%s: clag-id can only be applied to a bond interface' % ifaceobj.name)
                return False
        return self._check_peer_ip(ifaceobj)

    def _is_vxlan_device(self, ifaceobj):
        if ifaceobj.get_attr_value_first('vxlan-id'):
            return True
        return False

    def _is_peerlink(self, ifaceobj):
        if ifaceobj.get_attr_value('clagd-peer-ip') and ifaceobj.get_attr_value('clagd-sys-mac'):
            return True
        return False

    def _is_peerlink_learn_enable(self, ifaceobj):
        clagd_args = ifaceobj.get_attr_value('clagd-args')
        if clagd_args:
            for args in clagd_args:
                if "peerlinkLearnEnable" in args:
                    return True
        return False

    def _get_clag_bonds(self):
        clag_bonds = {}
        if self.is_process_running('clagd'):
            try:
                output = utils.exec_commandl(['/usr/bin/clagctl', '-j', 'showclagid'])
                clag_bonds = json.loads(output)
            except Exception, e:
                pass
        return clag_bonds

    def _is_clag_primary(self):
        clag_status = {}
        if self.is_process_running('clagd'):
            try:
                output = utils.exec_commandl(['/usr/bin/clagctl', '-j'])
                clag_status = json.loads(output)
            except Exception, e:
                pass
        return clag_status.get("status", {}).get("ourRole", "unknown") == "primary"

    def _get_clag_id(self, bondname):
        clagId = None
        if self.clag_bonds is None:
            self.clag_bonds = self._get_clag_bonds()
        bond = self.clag_bonds.get(bondname)
        if bond:
            clagId = bond.get('clagId')
        return str(clagId) if clagId else None

    def _check_clag_id(self, bondname, clag_id):
        curr_clag_id = self._get_clag_id(bondname)
        return curr_clag_id == clag_id, curr_clag_id

    def _set_old_global_local_ip(self, ifaceobj):
        old_local_ip = ifaceobj.get_attr_value_first('vxlan-local-tunnelip')
        if old_local_ip and not self.clagd_old_global_local_ip:
            self.clagd_old_global_local_ip = old_local_ip

    def _set_global_local_ip(self, ifaceobj):
        local_ip = ifaceobj.get_attr_value_first('vxlan-local-tunnelip')
        if local_ip and not self.clagd_global_local_ip:
            try:
                self.clagd_global_local_ip = str(IPv4Address(local_ip))
            except AddressValueError:
                try:
                    self.clagd_global_local_ip = str(IPv4Network(local_ip).ip)
                    self.log_warn('%s: vxlan-local-tunnelip %s: netmask ignored'
                                  % (ifaceobj.name, local_ip))
                except:
                    self.log_error("%s: invalid clagd-vxlan-anycast-ip %s: must be in ipv4 format"
                                   % (ifaceobj.name, local_ip))

    def _set_iface_clag_proto_down(self, ifName, protodown):
        old_proto_down = self._is_iface_clag_proto_down(ifName)
        if old_proto_down == protodown:
            return

        # XXX - move to netlink
        stateStr = "on" if protodown else "off"
        cmd = "/bin/ip link set %s protodown %s" %  (ifName, stateStr)
        self.logger.info('clag: setting link %s protodown %s' %(ifName, stateStr))
        utils.exec_command(cmd, stdout=False, stderr=utils.DEVNULL)

    def _set_clag_bond_proto_down(self, ifaceobj, clag_id):
        protodown = 1 if clag_id and clag_id != '0' else 0

        bond_slaves = ifaceobj.get_attr_value_first('bond-slaves')
        if bond_slaves:
            slaves = self.parse_port_list(ifaceobj.name, bond_slaves)
        else:
            slaves = []

        if not slaves:
            return

        for slave in slaves:
            self._set_iface_clag_proto_down(slave, protodown)

    def _set_clag_id(self, ifaceobj, clag_id):
        bondname = ifaceobj.name
        if self.is_process_running('clagd'):
            if not self.clagd_restart:
                (rc, running_clag_id) = self._check_clag_id(bondname, clag_id)
                if not rc:
                    utils.exec_commandl(['/usr/bin/clagctl', 'setclagid', bondname, clag_id])
        else:
            # if clagd is not running we will just protodown the slaves
            self._set_clag_bond_proto_down(ifaceobj, clag_id)

    def _is_iface_clag_proto_down(self, ifName):
        flags = 0
        sys_fs_proto_down = '/sys/class/net/'+ifName+'/proto_down'
        if os.path.isfile(sys_fs_proto_down):
            try:
                for line in open(sys_fs_proto_down):
                    flags = int(line.strip(), 0)
            except IOError:
                pass
        return flags

    def _is_iface_oper_up(self, ifaceobj):
        sys_fs_operstate = '/sys/class/net/'+ifaceobj.name+'/operstate'
        if os.path.isfile(sys_fs_operstate) and self.read_file_oneline(sys_fs_operstate) == 'up':
            return True
        return False

    def _drop_stale_clag_config(self, ifaceobj, post_down):
        # The mandatory clagd attributes are not present or the peerlink is
        # being deleted. If this iface used to be the peer link we will delete the clagd config and stop
        # clagd
        if self.clagd_cache.get('CLAGD_PEER_IF') == ifaceobj.name or\
            (post_down and ifaceobj.upperifaces and
            self.clagd_cache.get('CLAGD_PEER_IF') in ifaceobj.upperifaces):
            utils.exec_command('/bin/systemctl --no-block stop clagd.service')
            try:
                os.unlink(clagd._clagd_default_file)
            except OSError as e:
                if os.path.exists(clagd._clagd_default_file):
                    raise('failed to delete clagd file (%s)' %str(e))
            #Do not change this value 
            for retry in xrange(25):
                if not systemUtils.is_service_running('clagd', '/var/run/clagd.pid'): 
                    break
                else:
                    if retry == 24:
                        self.logger.info("ClagD process exit taking longer than 25s")       
                    else:
                        time.sleep(1)
            self.clagd_peerlink_name = ""
  
    def _post_down(self, ifaceobj, ifaceobj_getfunc=None):
        if ifaceobj.link_kind & ifaceLinkKind.BOND:
            self._set_clag_id(ifaceobj, '0')
        elif ifaceobj.get_attr_value_first('vxlan-id'):
            if self._is_iface_clag_proto_down(ifaceobj.name):
                cmd = '/bin/ip link set %s protodown off' %(ifaceobj.name)
                utils.exec_command(cmd)
        self._drop_stale_clag_config(ifaceobj, post_down=True)

    def _pre_up(self, ifaceobj, ifaceobj_getfunc=None):
        if ifaceobj.link_kind & ifaceLinkKind.BOND:
            attrval = ifaceobj.get_attr_value_first('clag-id')
            if attrval and ifaceobj.upperifaces and self.clagd_peerlink_name in ifaceobj.upperifaces:
                attrval = '0'
                self.log_warn('%s: clag-id may not be assigned to the peerlink. Ignoring...' % 
                              (ifaceobj.name,))
            attrval = attrval if attrval else '0'
            self._set_clag_id(ifaceobj, attrval)
        elif self._is_peerlink(ifaceobj) and os.path.isfile('/lib/clagctl-utils/clagVxlanProtoDown') and \
             not self._is_iface_oper_up(ifaceobj) and not self._is_clag_primary():
            # proto down vxlan devices that are already up only if peer-link
            # is transitioning from down to up and clag role is not primary
            cmd = '/lib/clagctl-utils/clagVxlanProtoDown'
            utils.exec_command(cmd)
        elif ifaceobj.get_attr_value_first('vxlan-id'):
            if self.clagd_peerlink_name and not self._is_iface_oper_up(ifaceobj):
                # proto down vxlan device transitioning from down to up state
                # if peer-link exists
                self.logger.info('clag peerlink exists, protodown vxlan interface')
                if not self._is_iface_clag_proto_down(ifaceobj.name):
                    cmd = '/bin/ip link set %s protodown on' %(ifaceobj.name)
                    utils.exec_command(cmd)
            else:
                if self._is_iface_clag_proto_down(ifaceobj.name):
                    cmd = '/bin/ip link set %s protodown off' %(ifaceobj.name)
                    utils.exec_command(cmd)
        elif ifaceobj.name == 'lo' and self.clagd_global_local_ip and self.is_process_running('clagd'):
            if self.clagd_old_global_local_ip != self.clagd_global_local_ip:
                utils.exec_commandl(['/usr/bin/clagctl', 'setvxlanlocalip', self.clagd_global_local_ip])

    def _config_accept_local(self, ifaceobj, enable):
        ifacename = ('/'.join(ifaceobj.name.split(".")))
        check, val = '1', 0
        if enable:
            check, val = '0', 1
        try:
            if self.read_file_oneline('/proc/sys/net/ipv4/conf/'
                                      '%s/accept_local' %ifaceobj.name) == check:
                self.sysctl_set('net.ipv4.conf.%s.accept_local' %ifacename, val)
        except Exception, e:
            self.log_warn(str(e))

    def _configure_accept_local(self, ifaceobj, vxlan_anycast_ip=None,
                                peerlinks=None, bridge_port=False):
        enable = False
        if vxlan_anycast_ip and peerlinks and not bridge_port:
            enable = True
        self._config_accept_local(ifaceobj, enable)

    def _up(self, ifaceobj, ifaceobj_getfunc=None):
        if not self.clagd_peerlinks:
            self.clagd_peerlinks = self._get_peerlinks(ifaceobj_getfunc)
        if (ifaceobj.link_kind & ifaceLinkKind.BRIDGE) and self.clagd_peerlink_name:
            if ifaceobj.lowerifaces:
                for peerlink in set(ifaceobj.lowerifaces) & self.clagd_peerlinks:
                    utils.exec_command('/sbin/brctl setportmcrouter ' + ifaceobj.name + ' ' + peerlink + ' 2')
                    if not self.clagd_peerlink_enable:
                        utils.exec_command('/sbin/bridge link set dev %s learning off' % peerlink)
            return

        clagd_peer_list = ifaceobj.get_attr_value('clagd-peer-ip')
        clagd_sysmac_list = ifaceobj.get_attr_value('clagd-sys-mac')
        req_args = 0
        if clagd_peer_list:
            req_args += 1
        if clagd_sysmac_list:
            req_args += 1
        if req_args == 2:
            clagd_params = {}
            clagd_params['CLAGD_ENABLE'] = 'yes'
            clagd_enable = ifaceobj.get_attr_value_first('clagd-enable')
            if clagd_enable and clagd_enable in ['yes', 'no']:
               clagd_params['CLAGD_ENABLE'] = clagd_enable

            clagd_params['CLAGD_PEER_IF'] = '%s' %ifaceobj.name
            if len(clagd_peer_list) != 1:
                self.log_warn('%s: multiple clagd-peer-ip lines, using first one'
                              % (ifaceobj.name,))

            clagd_params['CLAGD_PEER_IP'] = clagd_peer_list[0]
            try:
                ipaddr.IPAddress(clagd_peer_list[0])
            except ValueError as e:
                if clagd_peer_list[0] != 'linklocal':
                    self.log_warn('%s: Invalid clagd-peer-ip lines %s, use only IP address or "linklocal".'
                                  % (ifaceobj.name,clagd_peer_list[0]))
                    clagd_params['CLAGD_PEER_IP'] = "INVALID IP ADDRESS (%s)" % (clagd_peer_list[0],)
            else:
                self._check_peer_ip(ifaceobj)

            if len(clagd_sysmac_list) != 1:
                self.log_warn('%s: multiple clagd-sys-mac lines, using first one'
                              % (ifaceobj.name,))
            clagd_params['CLAGD_SYS_MAC'] = clagd_sysmac_list[0]
            clagd_backup = ""
            clagd_backup_vrf = ""
            clagd_backup_list = ifaceobj.get_attr_value('clagd-backup-ip')
            if clagd_backup_list:
                if len(clagd_backup_list) != 1:
                    self.log_warn('%s: multiple clagd-backup-ip lines, using first one'
                                  % (ifaceobj.name,))
                if clagd_backup_list[0].find('vrf') != -1:
                    clagd_backup_info = clagd_backup_list[0].split()
                    if len(clagd_backup_info) != 3:
                        self.log_warn('%s: clagd-backup-ip insufficent params'
                                  % (ifaceobj.name,))
                    else:
                        if clagd_backup_info[1] == 'vrf':
                            clagd_backup_vrf = clagd_backup_info[2]
                            clagd_backup = clagd_backup_info[0]
                        else:
                            self.log_warn('%s: clagd-backup-ip syntax is not correct'
                                      % (ifaceobj.name,))
                else:
                    clagd_backup = clagd_backup_list[0]
            clagd_params['CLAGD_BACKUP_VRF'] = clagd_backup_vrf
            clagd_params['CLAGD_BACKUP_IP'] = clagd_backup
            clagd_vxlan_list = ifaceobj.get_attr_value('clagd-vxlan-anycast-ip')
            if clagd_vxlan_list:
                self.log_warn('%s: clagd-vxlan-anycast-ip allowed only on \'lo\' interface'
                               % (ifaceobj.name))
            clagd_params['CLAGD_VXLAN_ANYCAST_IP'] = self.clagd_vxlan_anycast_ip
            self._configure_accept_local(ifaceobj,
                                         self.clagd_vxlan_anycast_ip,
                                         self.clagd_peerlinks,
                                         (ifaceobj.link_privflags &
                                          ifaceLinkPrivFlags.BRIDGE_PORT))
            clagd_prio_list = ifaceobj.get_attr_value('clagd-priority')
            clagd_priority="32768"
            if clagd_prio_list:
                if len(clagd_prio_list) != 1:
                    self.log_warn('%s: multiple clagd-priority lines, using first one'
                                  % (ifaceobj.name,))
                clagd_priority = clagd_prio_list[0]
            clagd_params['CLAGD_PRIORITY'] = clagd_priority
            clagd_args_list = ifaceobj.get_attr_value('clagd-args')
            if clagd_args_list:
                clagd_params['CLAGD_ARGS'] = " ".join(clagd_args_list).translate(None, "\"'")

            running_clagd_params = self._clagd_params_read()
            if running_clagd_params:
                cmdList = []
                if clagd_params['CLAGD_BACKUP_IP'] != running_clagd_params['CLAGD_BACKUP_IP'] or \
                   clagd_params['CLAGD_BACKUP_VRF'] != running_clagd_params['CLAGD_BACKUP_VRF']:
                    cmd = ['/usr/bin/clagctl', 'setbackupip', clagd_params['CLAGD_BACKUP_IP']]
                    if clagd_params['CLAGD_BACKUP_VRF'] and clagd_params['CLAGD_BACKUP_IP']:
                        cmd.extend(['vrf', clagd_params['CLAGD_BACKUP_VRF']])
                    cmdList.append(cmd)
                if clagd_params['CLAGD_VXLAN_ANYCAST_IP'] != running_clagd_params['CLAGD_VXLAN_ANYCAST_IP']:
                    cmdList.append(['/usr/bin/clagctl', 'setanycastip', clagd_params['CLAGD_VXLAN_ANYCAST_IP']])
                if clagd_params['CLAGD_PRIORITY'] != running_clagd_params['CLAGD_PRIORITY']:
                    cmdList.append(['/usr/bin/clagctl', 'priority', clagd_params['CLAGD_PRIORITY']])
                if cmdList:
                    self._clagd_params_write(ifaceobj, clagd_params)
                    for cmd in cmdList:
                        if self.is_process_running('clagd'):
                            try:
                                output = utils.exec_commandl(cmd)
                            except Exception, e:
                                pass

            if not self._clagd_params_compare(clagd_params, running_clagd_params) or \
               not running_clagd_params:

                self._clagd_params_write(ifaceobj, clagd_params)

                if clagd_params.get('CLAGD_ENABLE') == 'yes':
                    utils.exec_command('/bin/systemctl enable clagd.service')
                    if ifupdownflags.flags.PERFMODE:
                        utils.exec_command('/bin/systemctl --no-block start clagd.service')
                    else:
                        # Start the Clagd service. Let systemd monitor the clag run status
                        utils.exec_command('/bin/systemctl reset-failed clagd.service')
                        utils.exec_command('/bin/systemctl --no-block restart clagd.service')
                        self.clagd_restart = True
                else:
                    utils.exec_command('/bin/systemctl disable clagd.service')
                    utils.exec_command('/bin/systemctl --no-block stop clagd.service')

                # Update the cache to avoid deleting the new config
                self.clagd_cache = clagd_params
        else:
            if req_args == 1:
                self.log_warn('%s: Both the clagd-peer-ip and clagd-sys-mac '
                          'keywords are required. Skipping clagd configuration.'
                          % (ifaceobj.name,))
            self._drop_stale_clag_config(ifaceobj, post_down=False)
            self._configure_accept_local(ifaceobj)

    def _get_peerlinks(self, ifaceobj_getfunc):
        '''
        Return a set of interfaces which may be peerlinks. Take the lower
        interface of the configured mgmt peerlink (i.e. peer6.4000) and then
        get a set of all upper interfaces of that interface.
        '''
        peer_links = set()
        if ifaceobj_getfunc and self.clagd_peerlink_name:
            ifaceobjs = ifaceobj_getfunc(self.clagd_peerlink_name)
            if not ifaceobjs:
                return peer_links
            ifaceobj = ifaceobjs[0]
            if ifaceobj and ifaceobj.lowerifaces:
                peer_link = ifaceobj.lowerifaces[0]
                peer_links = set([peer_link])
                peer_linkobj = ifaceobj_getfunc(peer_link)[0]
                if peer_linkobj and peer_linkobj.upperifaces:
                    peer_links |= set(peer_linkobj.upperifaces)
        return peer_links

    def _clagd_params_compare(self, cparams1, cparams2):
        for p in ['CLAGD_ENABLE', 'CLAGD_PEER_IF', 'CLAGD_SYS_MAC',
                  'CLAGD_PEER_IP', 'CLAGD_ARGS']:
            if cparams1.get(p, '') != cparams2.get(p, ''):
               return False
        return True

    def _clagd_params_read(self):
        clagd_default_vars = {}
        try:
            with open(clagd._clagd_default_file, "r") as f:
                lines = f.readlines()
                for l in lines:
                    match = re.match('^(CLAGD_.*?)=(.*)$', l.strip())
                    if match:
                        g = match.groups()
                        clagd_default_vars[g[0]] = g[1].strip('\"')
        except Exception, e:
            self.logger.debug('error reading current clagd file (%s)' %str(e))
            pass
        return clagd_default_vars

    def _clagd_params_write(self, ifaceobj, clagd_params):
        try:
            if os.path.exists(clagd._clagd_default_file):
                os.chmod(clagd._clagd_default_file, 
                         stat.S_IWUSR | clagd._clagd_default_perm)
        except Exception, e:
            self.log_warn(str(e))
        self.logger.info('writing clagd file %s' %clagd._clagd_default_file)
        try:
            with open(clagd._clagd_default_file, "w") as f:
                f.write(clagd._clagd_default_template %
                    (asctime(), clagd_params.get('CLAGD_ENABLE', 'yes'),
                     clagd_params.get('CLAGD_PEER_IF', ''),
                     clagd_params.get('CLAGD_SYS_MAC', ''),
                     clagd_params.get('CLAGD_PEER_IP', ''), 
                     clagd_params.get('CLAGD_PRIORITY', ''), 
                     clagd_params.get('CLAGD_BACKUP_VRF', ''),
                     clagd_params.get('CLAGD_BACKUP_IP', ''), 
                     clagd_params.get('CLAGD_VXLAN_ANYCAST_IP', ''), 
                     clagd_params.get('CLAGD_ARGS', '')))
        except Exception, e:
            raise('error writing clagd file (%s)' %str(e))
        finally:
            try:
                if os.path.exists(clagd._clagd_default_file):
                   os.chmod(clagd._clagd_default_file,
                            clagd._clagd_default_perm)
            except Exception, e:
                self.log_warn(str(e))

    def split_ip_and_vrf(self, clagd_backup):
        if not clagd_backup:
            return

        ip_vrf_list = clagd_backup.split()
        vrf = None
        ip = None

        try:
            index_vrf = ip_vrf_list.index('vrf')
            vrf = ip_vrf_list[index_vrf + 1]
        except:
            pass

        for ip in ip_vrf_list:
            try:
                ip = IPNetwork(ip)
                break
            except:
                pass

        return ip, vrf

    def _query_check(self, ifaceobj, ifaceobjcurr, ifaceobj_getfunc=None):
        clag_id = ifaceobj.get_attr_value_first('clag-id')
        if clag_id:
            (rc, running_clag_id) = self._check_clag_id(ifaceobj.name, clag_id)
            if rc:
                ifaceobjcurr.update_config_with_status('clag-id',
                                      running_clag_id, 0)
            else:
                ifaceobjcurr.update_config_with_status('clag-id',
                                      running_clag_id, 1)

        clagd_peer = ifaceobj.get_attr_value_first('clagd-peer-ip')
        clagd_sysmac = ifaceobj.get_attr_value_first('clagd-sys-mac')
        clagd_vxlan = ifaceobj.get_attr_value_first('clagd-vxlan-anycast-ip')

        if (clagd_peer and clagd_sysmac) or clagd_vxlan:
            clagd_default_vars = {}

            clagd_default_vars = self._clagd_params_read()
            if not clagd_default_vars:
               ifaceobjcurr.check_n_update_config_with_status_many(ifaceobj,
                                         self.get_mod_attrs(), 1)
               return
            if clagd_peer and clagd_sysmac:
                clagd_backup = ifaceobj.get_attr_value_first('clagd-backup-ip')
                clagd_enable = ifaceobj.get_attr_value_first('clagd-enable')
                clagd_prio = ifaceobj.get_attr_value_first('clagd-priority')
                clagd_args_list = ifaceobj.get_attr_value('clagd-args')
                clagd_args = None
                if clagd_args_list:
                    clagd_args = " ".join(clagd_args_list).translate(None, "\"'")

                running_clagd_enable = clagd_default_vars.get('CLAGD_ENABLE', '')
                if clagd_enable:
                   if clagd_enable == running_clagd_enable:
                      ifaceobjcurr.update_config_with_status('clagd-enable',
                                                    running_clagd_enable, 0)
                   else:
                      ifaceobjcurr.update_config_with_status('clagd-enable',
                                                    running_clagd_enable, 1)

                running_clagd_backup = clagd_default_vars.get('CLAGD_BACKUP_IP', '')
                running_clagd_backup_vrf = clagd_default_vars.get('CLAGD_BACKUP_VRF', '')
                if running_clagd_backup_vrf:
                    running_clagd_backup = 'vrf %s %s' % (running_clagd_backup_vrf, running_clagd_backup)
                if clagd_backup:
                    user_ip, user_vrf  = self.split_ip_and_vrf(clagd_backup)
                    running_ip, running_vrf = self.split_ip_and_vrf(running_clagd_backup)

                    if user_ip == running_ip and user_vrf == running_vrf:
                        ifaceobjcurr.update_config_with_status('clagd-backup-ip',
                                                               running_clagd_backup, 0)
                    else:
                        ifaceobjcurr.update_config_with_status('clagd-backup-ip',
                                                               running_clagd_backup, 1)

                running_clagd_peer = clagd_default_vars.get('CLAGD_PEER_IP',
                                                            '')
                if clagd_peer != running_clagd_peer:
                   ifaceobjcurr.update_config_with_status('clagd-peer-ip',
                                            running_clagd_peer, 1)
                else:
                   ifaceobjcurr.update_config_with_status('clagd-peer-ip',
                                            running_clagd_peer, 0)

                running_clagd_sysmac = clagd_default_vars.get('CLAGD_SYS_MAC',
                                                              '')
                if clagd_sysmac != running_clagd_sysmac:
                   ifaceobjcurr.update_config_with_status('clagd-sys-mac',
                                            running_clagd_sysmac, 1)
                else:
                   ifaceobjcurr.update_config_with_status('clagd-sys-mac',
                                            running_clagd_sysmac, 0)

                if clagd_prio:
                    running_clagd_prio = clagd_default_vars.get('CLAGD_PRIORITY',
                                            '')
                    if clagd_prio != running_clagd_prio:
                       ifaceobjcurr.update_config_with_status('clagd-priority',
                                            running_clagd_prio, 1)
                    else:
                       ifaceobjcurr.update_config_with_status('clagd-priority',
                                            running_clagd_prio, 0)

                if clagd_args:
                    running_clagd_args = clagd_default_vars.get(
                                                   'CLAGD_ARGS', '')
                    if clagd_args != running_clagd_args:
                       ifaceobjcurr.update_config_with_status('clagd-args',
                                            running_clagd_args, 1)
                    else:
                       ifaceobjcurr.update_config_with_status('clagd-args',
                                            running_clagd_args, 0)

            if clagd_vxlan:
                running_clagd_vxlan = clagd_default_vars.get('CLAGD_VXLAN_ANYCAST_IP', '')
                if clagd_vxlan == running_clagd_vxlan:
                   ifaceobjcurr.update_config_with_status('clagd-vxlan-anycast-ip',
                                                 running_clagd_vxlan, 0)
                else:
                   ifaceobjcurr.update_config_with_status('clagd-vxlan-anycast-ip',
                                                 running_clagd_vxlan, 1)
        elif clagd_sysmac and not clagd_peer:
            ifaceobjcurr.update_config_with_status('clagd-sys-mac', '', 1)
        elif clagd_peer and not clagd_sysmac:
            ifaceobjcurr.update_config_with_status('clagd-peer-ip', '', 1)

    def _query_running(self, ifaceobjrunning, ifaceobj_getfunc=None):
        clag_id = self._get_clag_id(ifaceobjrunning.name)
        if clag_id:
            ifaceobjrunning.update_config_dict({'clag-id' : [clag_id]})
        clagd_vars = self._clagd_params_read()
        # if query is on the clag peer we need to add the clagd params
        clagd_peer_if = clagd_vars.get('CLAGD_PEER_IF', '')
        if clagd_peer_if == ifaceobjrunning.name:
            clagd_backup = clagd_vars.get('CLAGD_BACKUP_IP', '')
            clagd_backup_vrf = clagd_vars.get('CLAGD_BACKUP_VRF', '')
            if clagd_backup_vrf:
                clagd_backup = 'vrf %s %s' % (clagd_backup_vrf, clagd_backup)
            if clagd_backup:
                ifaceobjrunning.update_config_dict({'clagd-backup-ip' : [clagd_backup]})
            clagd_peer_ip = clagd_vars.get('CLAGD_PEER_IP', '')
            if clagd_peer_ip:
                ifaceobjrunning.update_config_dict({'clagd-peer-ip' : [clagd_peer_ip]})
            clagd_sys_mac = clagd_vars.get('CLAGD_SYS_MAC', '')
            if clagd_sys_mac:
                ifaceobjrunning.update_config_dict({'clagd-sys-mac' : [clagd_sys_mac]})
            clagd_priority = clagd_vars.get('CLAGD_PRIORITY', '')
            if clagd_priority:
                ifaceobjrunning.update_config_dict({'clagd-priority' : [clagd_priority]})
            clagd_args = clagd_vars.get('CLAGD_ARGS', '')
            if clagd_args:
                ifaceobjrunning.update_config_dict({'clagd-args' : [clagd_args]})
        if ifaceobjrunning.name == 'lo':
            clagd_vxlan = clagd_vars.get('CLAGD_VXLAN_ANYCAST_IP', '')
            if clagd_vxlan:
                ifaceobjrunning.update_config_dict({'clagd-vxlan-anycast-ip' : [clagd_vxlan]})

    _run_ops = {'post-up' : _up,
                'pre-up' : _pre_up,
                'post-down' : _post_down,
                'query-checkcurr' : _query_check,
                'query-running' : _query_running}

    def get_ops(self):
        return self._run_ops.keys()

    def run(self, ifaceobj, operation, query_ifaceobj=None, 
            ifaceobj_getfunc=None, **extra_args):
        op_handler = self._run_ops.get(operation)
        if not op_handler:
           return
        if operation == 'query-checkcurr':
            op_handler(self, ifaceobj, query_ifaceobj,
                       ifaceobj_getfunc=ifaceobj_getfunc)
        else:
            op_handler(self, ifaceobj, ifaceobj_getfunc=ifaceobj_getfunc)
